/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

require([
    "app/common",
    "app/quotas",
    "app/messages",
    "aps/load",
    "aps/ResourceStore",
    "aps/Memory",
    "aps/xhr",
    "dojo/promise/all",
    "dojox/mvc/at",
    "dojox/mvc/getStateful",
    "dojox/mvc/getPlainValue",
    "dijit/registry"
], function (
    common,
    quotas,
    messages,
    load,
    Store,
    Memory,
    xhr,
    all,
    at,
    getStateful,
    getPlainValue,
    registry
) {
    var tenant = aps.context.vars.tenant,
        tenantId = tenant.aps.id,
        account = aps.context.vars.user,
        accountStore = new Store({
            apsType: common.TYPES.USERS,
            target: "/aps/2/resources"
        }),
        quotasStatisticsDefer = xhr.get("/aps/2/resources/" + tenantId + "/quotasStatistics"),
        backupProfileDefer = xhr.get("/aps/2/resources/" + tenantId + "/profile");

    function renderQuota(quotasStatistics, userQuota) {
        var quotaContainer = registry.byId("quotaContainer"),
            controls = quotas.getQuotaControlsForExistingUser(
                account, tenant, quotasStatistics, false, userQuota),
            quotaIsOn = userQuota != common.USER_QUOTA.OFF;

        if (controls.storage && (quotaIsOn || account.userStorageUsage.limit > 0)) {
            controls.storage.set("value", account.userStorageUsage.limit == common.QUOTA_UNLIMITED_VALUE ?
                "" : account.userStorageUsage.limit / common.BYTES_IN_GB);
            quotaContainer.addChild(controls.storage);
        }

        if (controls.workstations && (quotaIsOn || account.UserWorkstationCount.limit > 0)) {
            controls.workstations.set("value", account.UserWorkstationCount.limit == common.QUOTA_UNLIMITED_VALUE ?
                "" : account.UserWorkstationCount.limit);
            quotaContainer.addChild(controls.workstations);
        }
        if (controls.servers && (quotaIsOn || account.userServerCount.limit > 0)) {
            controls.servers.set("value", account.userServerCount.limit == common.QUOTA_UNLIMITED_VALUE ?
                "" : account.userServerCount.limit);
            quotaContainer.addChild(controls.servers);
        }
        if (controls.vms && (quotaIsOn || account.UserVmCount.limit > 0)) {
            controls.vms.set("value", account.UserVmCount.limit == common.QUOTA_UNLIMITED_VALUE ?
                "" : account.UserVmCount.limit);
            quotaContainer.addChild(controls.vms);
        }
        if (controls.mobiles && (quotaIsOn || account.UserMobileDeviceCount.limit > 0)) {
            controls.mobiles.set("value", account.UserMobileDeviceCount.limit == common.QUOTA_UNLIMITED_VALUE ?
                "" : account.UserMobileDeviceCount.limit);
            quotaContainer.addChild(controls.mobiles);
        }
        if (controls.webHostingServers && (quotaIsOn || account.userWebHostingServerCount.limit > 0)) {
            controls.webHostingServers.set("value", account.userWebHostingServerCount.limit == common.QUOTA_UNLIMITED_VALUE ?
                "" : account.userWebHostingServerCount.limit);
            quotaContainer.addChild(controls.webHostingServers);
        }
        if (controls.websites && (quotaIsOn || account.UserWebsiteCount.limit > 0)) {
            controls.websites.set("value", account.UserWebsiteCount.limit == common.QUOTA_UNLIMITED_VALUE ?
                "" : account.UserWebsiteCount.limit);
            quotaContainer.addChild(controls.websites);
        }
    }

    function addQuotaValues(userData) {
        var userStorageUsage = registry.byId("userStorageUsage"),
            userServerCount = registry.byId("userServerCount"),
            userVmCount = registry.byId("userVmCount"),
            userWorkstationCount = registry.byId("userWorkstationCount"),
            userMobileDeviceCount = registry.byId("userMobileDeviceCount"),
            userWebHostingServerCount = registry.byId("userWebHostingServerCount"),
            userWebsiteCount = registry.byId("userWebsiteCount");

        function getLimitValue(limitFieldValue) {
            return limitFieldValue === "" ? common.QUOTA_UNLIMITED_VALUE : limitFieldValue;
        }

        function getStorageLimitValue(limitFieldValue) {
            return limitFieldValue === "" ? common.QUOTA_UNLIMITED_VALUE : limitFieldValue * common.BYTES_IN_GB;
        }

        userData.userStorageUsage = {limit: userStorageUsage ? getStorageLimitValue(userStorageUsage.value) : 0};
        userData.userServerCount = {limit: userServerCount ? getLimitValue(userServerCount.value) : 0};
        userData.UserVmCount = {limit: userVmCount ? getLimitValue(userVmCount.value) : 0};
        userData.UserWorkstationCount = {limit: userWorkstationCount ? getLimitValue(userWorkstationCount.value) : 0};
        userData.UserMobileDeviceCount = {limit: userMobileDeviceCount ? getLimitValue(userMobileDeviceCount.value) : 0};
        userData.userWebHostingServerCount = {limit: userWebHostingServerCount ? getLimitValue(userWebHostingServerCount.value) : 0};
        userData.UserWebsiteCount = {limit: userWebsiteCount ? getLimitValue(userWebsiteCount.value) : 0};
    }

    all({
        profiles: backupProfileDefer,
        quotasStatistics: quotasStatisticsDefer
    }).then(function (response) {
        var backupProfile = response.profiles.length ? response.profiles[0] : null,
            quotasStatistics = response.quotasStatistics,
            userQuota = backupProfile ? backupProfile.userQuota : common.USER_QUOTA.STRICT,
            showQuotas = quotas.canShowQuotas(userQuota, account),
            quotaIsOn = userQuota != common.USER_QUOTA.OFF,
            userModel = getStateful(aps.context.vars.user),
            privilegedCheckBoxState = getStateful({}),
            accountStatus;


        accountStatus = userModel.userstatus == common.USER_STATUS.NOT_ACTIVATED ?
            [["aps/Output", {
                label: _("Status"),
                value: "Not activated"
            }]] :
            [["aps/RadioButton", {
                id: "enableButton",
                label: _('Status'),
                name: "status",
                checked: at(userModel, "servicestatus").transform({
                    format: function (value) {
                        return parseInt(value, 10) === 1;
                    },
                    parse: function (value) {
                        return value ? 1 : 0;
                    }
                }),
                description: _('Enabled')
            }],
            ["aps/RadioButton", {
                id: "disableButton",
                name: "status",
                checked: at(userModel, "servicestatus").transform({
                    format: function (value) {
                        return parseInt(value, 10) !== 1;
                    },
                    parse: function (value) {
                        return value ? 0 : 1;
                    }
                }),
                description: _('Disabled'),
                hint: _('Disabling an account makes backup and recovery operations under this account unavailable. The backed-up data is kept until you delete the account.')
            }]];

        load(["aps/PageContainer", {id: "apsPageContainer"}, [
            ["aps/FieldSet", {
                title: _("Main")
            }, [
                ["aps/Output", {
                    label: _("User name"),
                    value: common.escapeAngularBrackets(userModel.displayName + ' (' + userModel.endUserLogin + ')')
                }],
                ["aps/CheckBox", {
                    id: "privileged",
                    label: _("Administrator privileges"),
                    checked: at(userModel, "privileged").transform({
                        format: function (value) {
                            return parseInt(value, 10) === common.USER_PRIVILEGED.YES;
                        },
                        parse: function (value) {
                            return (value) ? common.USER_PRIVILEGED.YES : common.USER_PRIVILEGED.NO;
                        }
                    }),
                    disabled: at(privilegedCheckBoxState, "disabled")
                }]
            ]],
            ["aps/FieldSet", {
                escapeHTML: false,
                title: _('Account status')
            }, accountStatus
            ],
            ["aps/FieldSet", {
                title: _('Notifications')
            }, [
                ["aps/CheckBox", {
                    id: "quotaSelect",
                    label: _("Enable notifications about quota excess"),
                    checked: at(userModel, "notificationsManagement"),
                    disabled: at(userModel, "servicestatus").transform({
                        format: function (value) {
                            return parseInt(value, 10) !== 1;
                        }
                    }).direction(at.from)
                }],
                ["aps/Select", {
                    id: "notificationSelect",
                    label: _("Backup notifications"),
                    value: at(userModel, "notifyBackup"),
                    store: new Memory({data: common.getBackupNotificationOptions(), idProperty: "value"}),
                    disabled: at(userModel, "servicestatus").transform({
                        format: function (value) {
                            return parseInt(value, 10) !== 1;
                        }
                    }).direction(at.from)
                }]
            ]],
            ["aps/FieldSet", {
                id: "quotaContainer",
                visible: quotaIsOn,
                description: quotas.getQuotaDescription(userQuota),
                title: _('Quotas'),
                disabled: at(userModel, "servicestatus").transform({
                    format: function (value) {
                        return parseInt(value, 10) !== 1;
                    }
                }).direction(at.from)
            }]
        ]]).then(function () {
            if (quotaIsOn) {
                renderQuota(quotasStatistics, userQuota);
            } else if (showQuotas) {
                messages.displayWarning(_("After submitting this form, all user quotas will become unlimited."));
            }

            function setPrivilegedCheckboxState(name, oldValue, newValue) {
                if (typeof newValue == "undefined") {
                    return;
                }
                privilegedCheckBoxState.set("disabled", parseInt(newValue, 10) !== 1);
            }

            userModel.watch("servicestatus", setPrivilegedCheckboxState);
            privilegedCheckBoxState.set("disabled", parseInt(userModel.servicestatus, 10) !== 1);

            aps.app.onSubmit = function () {
                var page = registry.byId("apsPageContainer");
                if (!page.validate()) {
                    aps.apsc.cancelProcessing();
                    return;
                }
                if (quotaIsOn) {
                    addQuotaValues(userModel);
                }
                accountStore.put(getPlainValue(userModel)).then(function () {
                    aps.apsc.gotoView(common.getRelatedViewId('list'));
                }, messages.displayError);
            };
        });
    });

    aps.app.onCancel = function () {
        aps.apsc.gotoView(common.getRelatedViewId('list'));
    };
});
