/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

define([
    "dojo/Deferred",
    "dojo/promise/all",
    "dojox/mvc/at",
    "dojox/mvc/getStateful",
    "aps/xhr",
    "aps/ResourceStore",
    "aps/Memory",
    "aps/FieldSet",
    "aps/Grid",
    "aps/Panel",
    "aps/Output",
    "aps/CheckBox",
    "aps/Container",
    "aps/Spinner",
    "./getStorageOutput",
    "./getResourceOutput",
    "./../../common",
    "./../../uomConverter"
], function (
    Deferred,
    all,
    at,
    getStateful,
    xhr,
    Store,
    Memory,
    FieldSet,
    Grid,
    Panel,
    Output,
    CheckBox,
    Container,
    Spinner,
    getStorageOutput,
    getResourceOutput,
    common,
    uomConverter
) {
    var countersMap = {
        StorageUsage: "userStorageUsage",
        ServerCount: "userServerCount",
        VmCount: "UserVmCount",
        WorkstationCount: "UserWorkstationCount",
        MobileDeviceCount: "UserMobileDeviceCount",
        WebHostingServerCount: "userWebHostingServerCount",
        WebsiteCount: "UserWebsiteCount"
    };

    function convert(value, from, to) {
        if (from === "unit") {
            return value;
        }

        if (value > 0) {
            return uomConverter.convert(value, from, to);
        }
        return value;
    }

    function getBillingResources(tenant) {
        return aps.biz.getResourcesInfo({
            filter: [{apsType: common.TYPES.TENANT}],
            subscriptionId: tenant.aps.subscription
        });
    }

    function findBillingResource(resource, billingResources) {
        var result = null;
        billingResources.some(function (billingResource) {
            if (resource.apsType === billingResource.apsType && resource.title === billingResource.name) {
                result = billingResource;
                return true;
            }
            return false;
        });
        return result;
    }

    function getSubscriptionResources(tenant) {
        return xhr('/aps/2/resources/' + tenant.aps.subscription + '/resources');
    }

    function isBackupAccountResource(resource) {
        var possibleResources = [
            {"apsType": common.TYPES.TENANT, "property": "StorageUsage"},
            {"apsType": common.TYPES.TENANT, "property": "ServerCount"},
            {"apsType": common.TYPES.TENANT, "property": "VmCount"},
            {"apsType": common.TYPES.TENANT, "property": "MobileDeviceCount"},
            {"apsType": common.TYPES.TENANT, "property": "WebHostingServerCount"},
            {"apsType": common.TYPES.TENANT, "property": "WebsiteCount"},
            {"apsType": common.TYPES.TENANT, "property": "WorkstationCount"}
        ];
        return possibleResources.some(function (possibleResource) {
            return resource.apsType === possibleResource.apsType && resource.property === possibleResource.property;
        });
    }

    /**
     * Returns array of global variables
     * @param tenant
     * @returns Deferred
     */
    function getGlobalVariables(tenant) {
        var deferred = new Deferred();
        xhr.get('/aps/2/resources/' + tenant.aps.id + '/global').then(function (response) {
            deferred.resolve(response[0]);
        }, deferred.reject);

        return deferred;
    }

    /**
     * Returns profile
     * @param tenant
     * @returns Deferred
     */
    function getProfile(tenant) {
        var deferred = new Deferred();
        xhr.get('/aps/2/resources/' + tenant.aps.id + '/profile').then(function (response) {
            deferred.resolve(response[0]);
        }, deferred.reject);

        return deferred;
    }

    function getBackupAccountResources(tenant) {
        var deferred = new Deferred();
        getSubscriptionResources(tenant).then(function (subscriptionResources) {
            var availableResources = [];
            subscriptionResources.forEach(function (resource) {
                if (isBackupAccountResource(resource)) {
                    availableResources.push(resource);
                }
            });
            deferred.resolve(availableResources);
        }, deferred.reject);
        return deferred;
    }

    function sumBackupAccountsCounters(resources) {
        var counters = {
            StorageUsage: {usage: 0, allocated: 0},
            ServerCount: {usage: 0, allocated: 0},
            VmCount: {usage: 0, allocated: 0},
            WorkstationCount: {usage: 0, allocated: 0},
            MobileDeviceCount: {usage: 0, allocated: 0},
            WebHostingServerCount: {usage: 0, allocated: 0},
            WebsiteCount: {usage: 0, allocated: 0}
        };

        function addResource(counter, resource) {
            if (!resource) {
                return;
            }

            counter.usage = addUsage(counter.usage, resource.usage);
            counter.allocated = addAllocated(counter.allocated, resource.limit);
        }

        function addUsage(usage, value) {
            if (!value) {
                return usage;
            }
            return (usage + value);
        }

        function addAllocated(allocated, value) {
            if (!value) {
                return allocated;
            }
            if (allocated < 0 || value < 0) {
                return -1;
            }
            return (allocated + value);
        }

        resources.forEach(function (resource) {
            addResource(counters.StorageUsage, resource.userStorageUsage);
            addResource(counters.ServerCount, resource.userServerCount);
            addResource(counters.VmCount, resource.UserVmCount);
            addResource(counters.WorkstationCount, resource.UserWorkstationCount);
            addResource(counters.MobileDeviceCount, resource.UserMobileDeviceCount);
            addResource(counters.WebHostingServerCount, resource.userWebHostingServerCount);
            addResource(counters.WebsiteCount, resource.UserWebsiteCount);
        });
        return counters;
    }

    function getBackupAccounts(tenant, excludeIds) {
        var query = excludeIds && excludeIds.length ? "out(aps.id,(" + excludeIds.join(',') + "))" : null,
            store = new Store({
                target: "/aps/2/resources/" + tenant.aps.id + "/users",
                baseQuery: query
            });
        return store.query();
    }

    function getSubscription(tenant) {
        var deferred = new Deferred();
        if (tenant.subscription) {
            deferred.resolve(tenant.subscription);
        }
        else {
            var store = new Store({target: "/aps/2/resources/" + tenant.aps.id + "/subscription"});
            store.query().then(function (subscriptions) {
                subscriptions && subscriptions.length ? deferred.resolve(subscriptions[0]) : deferred.resolve(null);
            }, deferred.reject);
        }
        return deferred;
    }

    function setDeltas(deltas, delta, apsType, apsProperty) {
        var deltaObj = deltas.find(function (item) {
            return item.apsType === apsType && item.apsProperty === apsProperty;
        });

        if (deltaObj) {
            deltaObj.delta = delta;
            return;
        }

        deltas.push({apsType: apsType, apsProperty: apsProperty, delta: delta});
    }

    function getUserResourcesGrid(tenant, model, divisor) {
        var showUnlimitedColumn = false,
            newUser = !(model.aps && model.aps.id),
            gridState = getStateful({
                unlimitedValue: {
                    StorageUsage: false,
                    ServerCount: false,
                    VmCount: false,
                    WorkstationCount: false,
                    MobileDeviceCount: false,
                    WebHostingServerCount: false,
                    WebsiteCount: false
                },
                spinnerValue: {
                    StorageUsage: 0,
                    ServerCount: 0,
                    VmCount: 0,
                    WorkstationCount: 0,
                    MobileDeviceCount: 0,
                    WebHostingServerCount: 0,
                    WebsiteCount: 0
                },
                counterDisabled: {
                    StorageUsage: false,
                    ServerCount: false,
                    VmCount: false,
                    WorkstationCount: false,
                    MobileDeviceCount: false,
                    WebHostingServerCount: false,
                    WebsiteCount: false
                }
            }),
            deltas = [];

        var deferred = new Deferred();
        all({
            subscription: getSubscription(tenant),
            subscriptionResources: getBackupAccountResources(tenant),
            billingResources: getBillingResources(tenant),
            backupAccounts: getBackupAccounts(tenant),
            global: getGlobalVariables(tenant),
            profile: getProfile(tenant)
        }).then(function (response) {
            var accounts = sumBackupAccountsCounters(response.backupAccounts),
                resources = [];
            response.subscriptionResources.forEach(function (subscriptionResource) {
                var resourceId = subscriptionResource.id,
                    resourceName = subscriptionResource.title,
                    apsType = subscriptionResource.apsType,
                    apsProperty = subscriptionResource.property,
                    resourceUnit = subscriptionResource.unit,
                    tenantLimit = tenant[apsProperty].limit,
                    tenantUsage = tenant[apsProperty].usage ? tenant[apsProperty].usage : 0,
                    allocated = accounts[apsProperty].allocated,
                    userUsage = model[countersMap[apsProperty]] && model[countersMap[apsProperty]].usage ? model[countersMap[apsProperty]].usage : 0,
                    userLimit = model[countersMap[apsProperty]] && model[countersMap[apsProperty]].limit ? model[countersMap[apsProperty]].limit : 0,
                    billingResource = findBillingResource(subscriptionResource, response.billingResources),
                    upSell = !!billingResource && !response.subscription.trial && tenantLimit >= 0;

                if (!tenantLimit && !upSell) {
                    return;
                }

                var available;
                if (tenantLimit < 0) {
                    available = -1;
                    showUnlimitedColumn = true;
                }
                else if (allocated < 0) {
                    available = 0;
                }
                else {
                    available = tenantLimit - allocated;
                    if (available < 0) {
                        available = 0;
                    }
                }

                var unit = resourceUnit === "unit" ? "unit" : "b",
                    spinnerUnit = resourceUnit === "unit" ? "unit" : "gb",
                    spinnerLegend = resourceUnit === "unit" ? "" : _("GB"),
                    resource = {
                        id: resourceId,
                        name: resourceName,

                        unit: unit,
                        spinnerUnit: spinnerUnit,
                        spinnerLegend: spinnerLegend,

                        apsType: apsType,
                        apsProperty: apsProperty,

                        userUsage: userUsage,
                        userLimit: userLimit,

                        tenantLimit: Number(convert(tenantLimit, unit, spinnerUnit).toFixed(0)),
                        tenantUsage: Number(Math.floor(convert(tenantUsage, unit, spinnerUnit)).toFixed(0)),
                        allocated: Number(Math.floor(convert(allocated, unit, spinnerUnit)).toFixed(0)),
                        available: Number(convert(available, unit, spinnerUnit).toFixed(0)),

                        upSell: upSell,
                        upSellMax: billingResource !== null && typeof billingResource.max !== "undefined" ? billingResource.max : null,
                        price: upSell ? billingResource.priceText : null
                    };
                resources.push(resource);
            });
            resources.forEach(function (resource) {
                var spinnerValue, modelValue;
                if (newUser) {
                    if (resource.tenantLimit < 0) {
                        gridState.unlimitedValue[resource.apsProperty] = true;
                        modelValue = -1;
                        spinnerValue = 0;
                    } else {
                        gridState.unlimitedValue[resource.apsProperty] = false;
                        modelValue = spinnerValue = Number(Math.floor(resource.available / divisor));
                    }
                } else {
                    if (resource.userLimit < 0) {
                        if (resource.available == -1) {
                            gridState.unlimitedValue[resource.apsProperty] = true;
                            modelValue = -1;
                        } else {
                            gridState.unlimitedValue[resource.apsProperty] = false;
                            modelValue = 0;
                        }
                        spinnerValue = 0;
                    } else {
                        gridState.unlimitedValue[resource.apsProperty] = false;
                        modelValue = spinnerValue = Number(convert(resource.userLimit, resource.unit, resource.spinnerUnit).toFixed(0));
                    }
                }
                gridState.spinnerValue[resource.apsProperty] = spinnerValue;
                model[countersMap[resource.apsProperty]].set("limit", Number(convert(modelValue, resource.spinnerUnit, resource.unit).toFixed(0)));
            });
            resources.sort(function(a, b) {
                return Number(a.id) - Number(b.id);
            });

            var store = new Memory({data: resources, idProperty: "id"});
            var grid = new Grid({
                noDataText: '',
                showPaging: false,
                pageSizeOptions: "GRID_UNLIMITED",
                store: store,
                columns: [{
                    "name": _("Quota"),
                    "sortable": false,
                    "field": "name",
                    renderCell: function (raw) {
                        var fs = new FieldSet();
                        fs.addChild(new Output({value: _(raw.name)}));
                        if (raw.upSell) {
                            fs.addChild(new Output({value: raw.price, alertType: "info"}));
                        }
                        return fs;
                    }
                }, {
                    "name": _("Current User's Usage"),
                    "sortable": false,
                    "field": "usage",
                    "visible": !newUser,
                    renderCell: function (raw) {
                        if (raw.unit === "unit") {
                            return getResourceOutput({
                                noneLegend: _("None"),
                                limitedLegend: _("${usage} of ${limit}"),
                                zeroUsageLegend: _("${limit} available"),
                                unlimitedLegend: _("${usage}"),
                                usage: raw.userUsage || 0,
                                limit: raw.userLimit || 0
                            });
                        }

                        return getStorageOutput({
                            noneLegend: _("None"),
                            limitedLegend: _("${usage} ${usageUnit} of ${limit} ${limitUnit}"),
                            unlimitedLegend: _("${usage} ${usageUnit}"),
                            usage: raw.userUsage || 0,
                            limit: raw.userLimit || 0
                        });
                    }
                }, {
                    "name": _("Available for Assigning"),
                    "sortable": false,
                    renderCell: function (raw) {
                        if (raw.tenantLimit < 0) {
                            return new Output({
                                value: _("Unlimited")
                            });
                        }

                        if (raw.unit === "unit") {
                            return new Output({
                                content: (_('${available} of ${limit}')),
                                limit: raw.tenantLimit,
                                available: raw.available
                            });
                        }

                        return new Output({
                            content: (_('${available} ${availableUnit} of ${limit} ${limitUnit}')),
                            limit: raw.tenantLimit,
                            limitUnit: _("GB"),
                            available: raw.available,
                            availableUnit: _("GB")
                        });
                    }
                }, {
                    "name": _("Unlimited"),
                    "sortable": false,
                    "align": "center",
                    visible: showUnlimitedColumn,
                    renderCell: function (raw) {
                        return new CheckBox({
                            checked: at(gridState.unlimitedValue, raw.apsProperty).transform({
                                parse: function (value) {
                                    if (value) {
                                        model[countersMap[raw.apsProperty]].set("limit", -1);
                                    }
                                    else {
                                        model[countersMap[raw.apsProperty]].set("limit", Number(convert(Number(gridState.spinnerValue[raw.apsProperty]), raw.spinnerUnit, raw.unit).toFixed(0)));
                                    }
                                    return value;
                                }
                            }),
                            disabled: raw.tenantLimit >= 0 || at(gridState.counterDisabled, raw.apsProperty)
                        });
                    }
                }, {
                    "name": _("New Quota"),
                    "sortable": false,
                    "align": "left",
                    renderCell: function (raw) {
                        var userLimit = Number(convert(raw.userLimit, raw.unit, raw.spinnerUnit).toFixed(0)),
                            available = raw.available + (userLimit < 0 ? 0 : userLimit),
                            maximum = (raw.available < 0) ? null : Number(Math.floor(available / divisor));

                        var spinner = new Spinner({
                            value: at(gridState.spinnerValue, raw.apsProperty).transform({
                                parse: function (value) {
                                    model[countersMap[raw.apsProperty]].set("limit", Number(convert(Number(value), raw.spinnerUnit, raw.unit).toFixed(0)));

                                    if (raw.upSell) {
                                        var incrise = (value * divisor) > available ? ((value * divisor) - available) : 0,
                                            delta = (raw.tenantLimit - raw.tenantUsage) + incrise; // workaround!

                                        setDeltas(deltas, delta, raw.apsType, raw.apsProperty);
                                        aps.biz.requestUsageChange({
                                            deltas: deltas,
                                            operations: [],
                                            subscriptionId: tenant.aps.subscription
                                        });
                                    }
                                    return value;
                                }
                            }),
                            minimum: 0,
                            maximum: raw.upSell ? raw.upSellMax : maximum,
                            step: 1,
                            legend: raw.spinnerLegend,
                            invalidMessage: _("This field can only contain numbers. No spaces or other characters allowed."),
                            extraValidator: function (value) {
                                return Number.isInteger(+value);
                            }
                        });

                        function disableSpinner() {
                            spinner.set("disabled", (gridState.unlimitedValue[raw.apsProperty] || gridState.counterDisabled[raw.apsProperty]));
                        }

                        disableSpinner();
                        gridState.unlimitedValue.watch(raw.apsProperty, disableSpinner);
                        gridState.counterDisabled.watch(raw.apsProperty, disableSpinner);

                        return spinner;
                    }
                }]
            });

            deferred.resolve({
                grid: grid,
                deltas: deltas
            });
        }, deferred.reject);
        return deferred;
    }

    return getUserResourcesGrid;
});
