/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

define([
    "aps/Tile",
    "aps/FieldSet",
    "./getResourceOutput",
    "./getStorageOutput",
    "./../widgets/base/Output",
    "./../../messages",
    "./../utils"
], function (
    Tile,
    FieldSet,
    getResourceOutput,
    getStorageOutput,
    Output,
    messages,
    utils
) {
    var BACKUP_USER_STATUSES = {
        ERROR: "-1",
        INACTIVE: "0",
        ACTIVE: "1",
        IN_PROGRESS: "2",
        NOT_ACTIVATED: "3"
    };

    var BACKUP_SERVICE_STATUSES = {
        NOT_ASSIGNED: "not_assigned",
        PROVISIONING: 'provisioning',
        NOT_ACTIVATED: "not_activated",
        ACTIVATED: "activated",
        DISABLED: "disabled",
        UNPROVISIONING: 'unprovisioning',
        ERROR: "error"
    };

    function getBackupServiceStatus(backupAccount) {
        if (!backupAccount) {
            return BACKUP_SERVICE_STATUSES.NOT_ASSIGNED;
        }
        switch (backupAccount.aps.status) {
            case "aps:proto":
            case "aps:provisioning":
                return BACKUP_SERVICE_STATUSES.PROVISIONING;
            case "aps:unprovisioning":
            case "aps:deleted":
                return BACKUP_SERVICE_STATUSES.UNPROVISIONING;
            case "aps:ready":
                switch (backupAccount.userstatus) {
                    case BACKUP_USER_STATUSES.ERROR:
                        return BACKUP_SERVICE_STATUSES.ERROR;
                    case BACKUP_USER_STATUSES.INACTIVE:
                        return BACKUP_SERVICE_STATUSES.DISABLED;
                    case BACKUP_USER_STATUSES.ACTIVE:
                        return BACKUP_SERVICE_STATUSES.ACTIVATED;
                    case BACKUP_USER_STATUSES.NOT_ACTIVATED:
                        return BACKUP_SERVICE_STATUSES.NOT_ACTIVATED;
                }
        }
        return undefined;
    }

    function createUserResources(backupAccount, backupAccountStatistic, tenant) {
        var fs = new FieldSet({
            gridSize: "md-12 xs-12"
        });

        function hasLimit(tenantCounter, userCounter) {
            if (!tenant.hasOwnProperty(tenantCounter) || !backupAccountStatistic.hasOwnProperty(userCounter)) {
                return false;
            }
            return tenant[tenantCounter].limit !== 0 && backupAccountStatistic[userCounter].limit !== 0;
        }

        fs.addChild(new Output({
            gridSize: "md-12 xs-12",
            label: _("Login"),
            value: backupAccount.endUserLogin
        }));

        if (hasLimit('StorageUsage', 'userStorageUsage')) {
            fs.addChild(getStorageOutput({
                gridSize: "md-4 xs-12",
                label: _('Cloud Storage'),
                noneLegend: _("None"),
                limitedLegend: _("${usage} ${usageUnit} of ${limit} ${limitUnit} used"),
                unlimitedLegend: _("${usage} ${usageUnit} used"),
                usage: backupAccountStatistic.userStorageUsage.usage || 0,
                limit: backupAccountStatistic.userStorageUsage.limit || 0
            }));
        }

        if (hasLimit('WorkstationCount', 'UserWorkstationCount')) {
            fs.addChild(getResourceOutput({
                gridSize: "md-4 xs-12",
                label: _('Workstations'),
                limitedLegend: _("${usage} of ${limit} machines protected"),
                zeroUsageLegend: _("${limit} machines can be protected"),
                unlimitedLegend: _("${usage} machines protected"),
                usage: backupAccountStatistic.UserWorkstationCount.usage || 0,
                limit: backupAccountStatistic.UserWorkstationCount.limit || 0
            }));
        }

        if (hasLimit('ServerCount', 'userServerCount')) {
            fs.addChild(getResourceOutput({
                gridSize: "md-4 xs-12",
                label: _('Servers'),
                limitedLegend: _("${usage} of ${limit} machines protected"),
                zeroUsageLegend: _("${limit} machines can be protected"),
                unlimitedLegend: _("${usage} machines protected"),
                usage: backupAccountStatistic.userServerCount.usage || 0,
                limit: backupAccountStatistic.userServerCount.limit || 0
            }));
        }

        if (hasLimit('VmCount', 'UserVmCount')) {
            fs.addChild(getResourceOutput({
                gridSize: "md-4 xs-12",
                label: _('Virtual Machines'),
                limitedLegend: _("${usage} of ${limit} machines protected"),
                zeroUsageLegend: _("${limit} machines can be protected"),
                unlimitedLegend: _("${usage} machines protected"),
                usage: backupAccountStatistic.UserVmCount.usage || 0,
                limit: backupAccountStatistic.UserVmCount.limit || 0
            }));
        }
        if (hasLimit('MobileDeviceCount', 'UserMobileDeviceCount')) {
            fs.addChild(getResourceOutput({
                gridSize: "md-4 xs-12",
                label: _('Mobile Devices'),
                limitedLegend: _("${usage} of ${limit} devices protected"),
                zeroUsageLegend: _("${limit} devices can be protected"),
                unlimitedLegend: _("${usage} devices protected"),
                usage: backupAccountStatistic.UserMobileDeviceCount.usage || 0,
                limit: backupAccountStatistic.UserMobileDeviceCount.limit || 0
            }));
        }
        if (hasLimit('WebHostingServerCount', 'userWebHostingServerCount')) {
            fs.addChild(getResourceOutput({
                gridSize: "md-4 xs-12",
                label: _('Web Hosting Servers'),
                limitedLegend: _("${usage} of ${limit} servers protected"),
                zeroUsageLegend: _("${limit} servers can be protected"),
                unlimitedLegend: _("${usage} servers protected"),
                usage: backupAccountStatistic.userWebHostingServerCount.usage || 0,
                limit: backupAccountStatistic.userWebHostingServerCount.limit || 0
            }));
        }
        if (hasLimit('WebsiteCount', 'UserWebsiteCount')) {
            fs.addChild(getResourceOutput({
                gridSize: "md-4 xs-12",
                label: _('Websites'),
                limitedLegend: _("${usage} of ${limit} websites protected"),
                zeroUsageLegend: _("${limit} websites can be protected"),
                unlimitedLegend: _("${usage} websites protected"),
                usage: backupAccountStatistic.UserWebsiteCount.usage || 0,
                limit: backupAccountStatistic.UserWebsiteCount.limit || 0
            }));
        }

        return fs;
    }

    function createUserActivation() {
        return new Output({
            escapeHTML: false,
            gridSize: "md-12 xs-12",
            content: utils.ol([
                _("Click Activate button to set password for your backup user."),
                _("Download and install the backup agent on your device. During the installation, specify your backup user login and password."),
                _("Apply a backup plan to the device after the agent installation."),
                _("Log into Backup Console to manage backups.")
            ])
        });
    }

    function getUserResourcesTile(backupAccount, backupAccountStatistic, backupAccountActions, tenant) {
        var status = getBackupServiceStatus(backupAccount),
            buttons = [];

        var tile = new Tile({
            title: _("Backup User"),
            gridSize: "md-8 xs-12"
        });

        switch (status) {
            case BACKUP_SERVICE_STATUSES.NOT_ASSIGNED:
                tile.addChild(new Output({
                    value: _("The service is not yet assigned. Click <b>Assign Service</b> button below to activate it."),
                    escapeHTML: false
                }));
                buttons.push({
                    iconClass: "fa fa-check",
                    title: _("Assign Service"),
                    autoBusy: false,
                    onClick: function () {
                        backupAccountActions.assign().then(null, messages.displayError);
                    }
                });
                break;

            case BACKUP_SERVICE_STATUSES.PROVISIONING:
                tile.addChild(new Output({value: _("Please wait...")}));
                break;

            case BACKUP_SERVICE_STATUSES.NOT_ACTIVATED:
                tile.addChild(createUserActivation());
                tile.set("title", _("Getting Started"));
                buttons.push({
                    iconClass: "fa fa-magic",
                    title: _("Activate"),
                    autoBusy: false,
                    onClick: function () {
                        backupAccountActions.activate().then(null, messages.displayError);
                    }
                });
                break;

            case BACKUP_SERVICE_STATUSES.ACTIVATED:
                tile.addChild(createUserResources(backupAccount, backupAccountStatistic, tenant));

                buttons.push({
                    iconClass: "fa fa-external-link",
                    title: _("Log into Backup Console"),
                    autoBusy: false,
                    onClick: function () {
                        var w = window.open('', '_blank');
                        backupAccountActions.loginToConsole().then(function (uri) {
                            w.location.href = uri;
                        }, function (error) {
                            w.close();
                            messages.displayError(error);
                        });
                    }
                });
                buttons.push({
                    iconClass: "fa fa-key",
                    title: _("Change Password"),
                    type: "danger",
                    autoBusy: false,
                    onClick: function () {
                        backupAccountActions.changePassword().then(null, messages.displayError);
                    }
                });
                break;

            case BACKUP_SERVICE_STATUSES.DISABLED:
                tile.addChild(new Output({value: _("Your account is disabled.")}));
                break;

            case BACKUP_SERVICE_STATUSES.UNPROVISIONING:
                tile.addChild(new Output({value: _("Please wait...")}));
                break;

            case BACKUP_SERVICE_STATUSES.ERROR:
                tile.addChild(new Output({value: _("Oops...")}));
                break;
        }

        tile.set("buttons", buttons);
        return tile;
    }

    return getUserResourcesTile;
});
