/*
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

define([
    'aps/passwdqc/generator',
    './../uomConverter',
    './../common'
], function (
    generator,
    uomConverter,
    common
) {
    /**
     * Returns error string if password is invalid, true if password is ok
     *
     * @param {string} pwd
     * @returns {String|Boolean}
     */
    function doValidatePassword(pwd) {
        if (pwd.match(/\s+/)) {
            return _('Spaces are forbidden characters in passwords');
        }
        if (pwd.length < 7) {
            return _('Password is too short');
        }
        if (!pwd.match(/[a-z]+/))
        {
            return _('Password must contain at least one lowercase character');
        }
        if (!pwd.match(/[A-Z]+/))
        {
            return _('Password must contain at least one uppercase character');
        }
        if (!pwd.match(/[0-9]+/))
        {
            return _('Password must contain at least one number');
        }
        return true;
    }

    return {
        detectOS: function(){
            var os = {};
            if (navigator.platform.indexOf('Win') != -1)
                os = 'win';
            else if (navigator.platform.indexOf('Mac') != -1)
                os = 'mac';
            else if (navigator.platform.indexOf('Linux') != -1)
                os = 'linux';
            else if (navigator.platform.indexOf('iPhone') != -1)
                os = 'ios';
            else if (navigator.platform.indexOf('iPod') != -1)
                os = 'ios';
            else if (navigator.platform.indexOf('iPad') != -1)
                os = 'ios';
            else if (navigator.platform.indexOf('android') != -1)
                os = 'android';
            return os;
        },
        ol: function(items){
            return '<ol><li>' + items.join('</li><li>') + '</li></ol>';
        },
        openURL: function(url) {
            // Todo: Fix for iPhone
            window.open(url, '_blank');
        },
        validatePassword: doValidatePassword,
        passwordGenerator: function() {
            var min = this._getMinimumLength(),
                max = this._getMaximumLength();

            var length = min;
            var attempts = 0;
            var password = generator({ length: length });

            while((this.calcStrength(password) < 5 || doValidatePassword(password) !== true) && (attempts++ <= 50)) {
                length = min + attempts;
                length = length >= max ? max : length;
                password = generator({length: length});
            }
            return password;
        },
        getOsButtonInfo: function() {
            var os = this.detectOS(),
                buttonTitle;

            switch(os){
                case 'win':
                    buttonTitle = _('Download for Windows');
                    break;
                case 'mac':
                    buttonTitle = _('Download for Mac OS X');
                    break;
                case 'linux':
                    buttonTitle = _('Download for Linux');
                    break;
                case 'ios':
                    buttonTitle = _('Download for IOs');
                    break;
                case 'android':
                    buttonTitle = _('Download for Android');
                    break;
                default:
                    buttonTitle = _('Download');
                    break;
            }
            return {
                os: os,
                title: buttonTitle
            };
        },
        createReturnUrl: function(){
            var params = JSON.parse(JSON.stringify(aps.context.params || {}));
            var resourceId = null;
            var viewId = null;
            var nav = null;

            if(aps.navigation){
                nav = aps.navigation.tree.context.slice(-1)[0];
                viewId = nav.viewId;
                resourceId = nav.resourceId;
            } else if(window.parent && window.parent.aps) {
                nav = window.parent.aps.navigation.tree.context.slice(-1)[0];
                viewId = nav.viewId;
                resourceId = nav.resourceId;
            } else {
                viewId = common.APP_ID + '#' + common.getCurrentView();
            }

            return {
                viewId: viewId,
                resourceId: resourceId,
                params: params
            };
        },
        convertStorageUsage: function(value) {
            if (!value) {
                return 0;
            }
            if (value < 0) {
                return common.QUOTA_UNLIMITED_VALUE;
            }
            return Math.ceil(uomConverter.convert(Number(value), 'b', 'gb'));
        },
        convertStorageLimit: function(value) {
            if (!value) {
                return 0;
            }
            if (value < 0) {
                return common.QUOTA_UNLIMITED_VALUE;
            }
            return Number(uomConverter.convert(Number(value), 'b', 'gb').toFixed(0));
        }
    };
});
