/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

define([
    "dojo/_base/declare",
    "dojo/promise/all",
    "aps/xhr",
    "dojo/_base/lang",
    "dojo/Deferred",
    "dojox/mvc/at",
    "dojox/mvc/getStateful",
    "dojox/mvc/getPlainValue",
    "aps/_View",
    "aps/Memory",
    "aps/Container",
    "aps/Output",
    "aps/load",
    "./../components/getUserResourcesGrid",
    "./../../messages",
    "./../../common"
], function (
    declare,
    all,
    xhr,
    lang,
    Deferred,
    at,
    getStateful,
    getPlainValue,
    _View,
    Memory,
    Container,
    Output,
    load,
    getUserResourcesGrid,
    messages,
    common
) {
    return declare(_View, {
        _container: null,
        _tenant: null,
        _model: null,
        _users: null,
        _userQuota: null,
        _deltas: [],

        _submit: function () {
            var self = this;

            var operation = {
                "call": {
                    "resourceId": self._tenant.aps.id,
                    "operation": "setUsersLicensesAfterUpsell",
                    "parameters": {
                        "data": {
                            provisionResource: self._users.map(function (user) {
                                return lang.mixin(getPlainValue(self._model), {user: {aps: {id: user.aps.id}}});
                            })
                        }
                    }
                }
            };

            return aps.biz.requestUsageChange({
                deltas: self._deltas,
                operations: [operation],
                subscriptionId: self._tenant.aps.subscription
            });
        },
        _renderForm: function () {
            var self = this;

            self._container.removeAll();
            self._container.set("title", _("Assign Backup Service for {count} user(s)").replace('{count}', self._users.length));
            if (self._userQuota) {
                self._container.addChild(new Output({
                    value: _("To assign backup service you need to define resource limits for the users you are now creating. You will be able to change them later.")
                }));
            }

            var form = ["aps/Panel", {title: _("Backup user settings")}, [
                ["aps/FieldSet", {gridSize: "md-12 xs-12"}, [
                    ["aps/Select", {
                        id: self.genId('backup_notifications'),
                        gridSize: "md-6 xs-12",
                        label: _("Backup Notifications"),
                        value: at(self._model, "notifyBackup").transform({
                            parse: function (value) {
                                return Number(value);
                            }
                        }),
                        required: true,
                        store: new Memory({data: common.getBackupNotificationOptions(), idProperty: "value"})
                    }],
                    ["aps/Select", {
                        id: self.genId('business_notifications'),
                        gridSize: "md-6 xs-12",
                        label: _("Business Notifications"),
                        description: _("Enable notifications about quota excess for this user."),
                        value: at(self._model, "notificationsManagement").transform({
                            format: function (value) {
                                return Number(value);
                            },
                            parse: function (value) {
                                return !!Number(value);
                            }
                        }),
                        required: true,
                        store: new Memory({data: common.getEnabledDisabledOptions(), idProperty: "value"})
                    }],
                    ["aps/Select", {
                        id: self.genId('administrator_privileges'),
                        gridSize: "md-6 xs-12",
                        label: _("Administrator Privileges"),
                        description: _("An administrator can manage backups of other users and back up Office 365 seats."),
                        value: at(self._model, "privileged").transform({
                            parse: function (value) {
                                return Number(value);
                            },
                            format: function (value) {
                                return String(value);
                            }
                        }),
                        required: true,
                        store: new Memory({idProperty: "value", data: common.getEnabledDisabledOptions()})
                    }]
                ]]
            ]];

            load(form, self._container).then(function () {
                if (self._userQuota) {
                    getUserResourcesGrid(self._tenant, self._model, self._users.length).then(function (resources) {
                        self._container.addChild(resources.grid);
                        self._deltas = resources.deltas;
                        aps.apsc.hideLoading();
                    }, messages.displayError);
                }
                else {
                    aps.apsc.hideLoading();
                }
            }, messages.displayError);
        },
        init: function () {
            this._container = new Container();
            this._container.placeAt(this.domNode);
        },
        onHide: function () {
            messages.removeAll();
        },
        onPrev: function () {
            aps.apsc.prev();
        },
        onNext: function () {
            var self = this;
            if (!self._container.validate()) {
                aps.apsc.cancelProcessing();
                return;
            }
            self._submit().then(function () {
                aps.apsc.next();
            });
        },
        onContext: function () {
            var self = this;
            all({
                resourcesToBind: aps.biz.getResourcesToBind(),
                tenants: xhr.get("/aps/2/resources/?implementing(" + common.TYPES.TENANT + "),eq(aps.subscription," + aps.context.subscriptionId + "),select(profile)")
            }).then(function (response) {
                self._users = response.resourcesToBind.users;
                self._tenant = response.tenants.pop();

                common.getBackupProfile(self._tenant).then(function (profile) {
                    self._userQuota = common.getUserQuota(profile);
                    self._model = getStateful({
                        aps: {
                            type: common.TYPES.USERS,
                            subscription: self._tenant.aps.subscription
                        },
                        tenant: {aps: {id: self._tenant.aps.id}},

                        notifyBackup: common.getBackupNotification(profile),
                        notificationsManagement: common.getManagementNotification(profile),

                        privileged: common.USER_PRIVILEGED.NO,

                        activateDuringCreation: false
                    })
                    ;
                    if (self._userQuota) {
                        self._model.set({
                            userStorageUsage: getStateful({limit: 0}),
                            userServerCount: getStateful({limit: 0}),
                            UserVmCount: getStateful({limit: 0}),
                            UserWorkstationCount: getStateful({limit: 0}),
                            UserMobileDeviceCount: getStateful({limit: 0}),
                            userWebHostingServerCount: getStateful({limit: 0}),
                            UserWebsiteCount: getStateful({limit: 0})
                        });
                    }
                    else {
                        self._model.set({
                            userStorageUsage: getStateful({limit: common.QUOTA_UNLIMITED_VALUE}),
                            userServerCount: getStateful({limit: common.QUOTA_UNLIMITED_VALUE}),
                            UserVmCount: getStateful({limit: common.QUOTA_UNLIMITED_VALUE}),
                            UserWorkstationCount: getStateful({limit: common.QUOTA_UNLIMITED_VALUE}),
                            UserMobileDeviceCount: getStateful({limit: common.QUOTA_UNLIMITED_VALUE}),
                            userWebHostingServerCount: getStateful({limit: common.QUOTA_UNLIMITED_VALUE}),
                            UserWebsiteCount: getStateful({limit: common.QUOTA_UNLIMITED_VALUE})
                        });
                    }

                    self._renderForm();
                }, messages.displayError);
            }, messages.displayError);
        }
    });
});

