/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

define([
    "dojo/_base/declare",
    "dojo/_base/lang",
    "dojo/Deferred",
    "dojo/promise/all",
    "dijit/registry",
    "dojox/mvc/at",
    "dojox/mvc/getStateful",
    "dojox/mvc/getPlainValue",
    "aps/_View",
    "aps/ResourceStore",
    "aps/Container",
    "aps/Memory",
    "aps/load",
    "aps/xhr",
    "./../utils",
    "./../components/getUserResourcesGrid",
    "./../../common",
    "./../../messages"
], function (
    declare,
    lang,
    Deferred,
    all,
    registry,
    at,
    getStateful,
    getPlainValue,
    _View,
    Store,
    Container,
    Memory,
    load,
    xhr,
    utils,
    getUserResourcesGrid,
    common,
    messages
) {
    var counters = [
        "userStorageUsage",
        "userServerCount",
        "UserVmCount",
        "UserWorkstationCount",
        "UserMobileDeviceCount",
        "userWebHostingServerCount",
        "UserWebsiteCount"
    ];

    function getBackupAccount(tenant, user) {
        var deferred = new Deferred();
        xhr.get("/aps/2/resources/" + tenant.aps.id + "/users?eq(user.aps.id," + user.aps.id + ")").then(function (users) {
            users && users.length ? deferred.resolve(users.pop()) : deferred.resolve(null);
        }, deferred.reject);
        return deferred;
    }

    function createModelForNewBackupAccount(tenant, user, profile, needActivation) {
        var userQuota = common.getUserQuota(profile);

        return getStateful({
            aps: {
                type: common.TYPES.USERS,
                subscription: tenant.aps.subscription
            },
            user: [{aps: {id: user.aps.id}}],
            tenant: [{aps: {id: tenant.aps.id}}],

            endUserLogin: user.login,
            endUserPassword: null,
            activateDuringCreation: needActivation,

            notifyBackup: common.getBackupNotification(profile),
            notificationsManagement: common.getManagementNotification(profile),

            privileged: common.USER_PRIVILEGED.NO,

            userStorageUsage: {limit: userQuota ? 0 : common.QUOTA_UNLIMITED_VALUE},
            userServerCount: {limit: userQuota ? 0 : common.QUOTA_UNLIMITED_VALUE},
            UserVmCount: {limit: userQuota ? 0 : common.QUOTA_UNLIMITED_VALUE},
            UserWorkstationCount: {limit: userQuota ? 0 : common.QUOTA_UNLIMITED_VALUE},
            UserMobileDeviceCount: {limit: userQuota ? 0 : common.QUOTA_UNLIMITED_VALUE},
            userWebHostingServerCount: {limit: userQuota ? 0 : common.QUOTA_UNLIMITED_VALUE},
            UserWebsiteCount: {limit: userQuota ? 0 : common.QUOTA_UNLIMITED_VALUE}
        });
    }

    function createModelForExistingBackupAccount(backupAccount, backupAccountStatistic) {
        var account = {
            aps: {
                id: backupAccount.aps.id
            },

            endUserLogin: backupAccount.endUserLogin,

            notifyBackup: backupAccount.notifyBackup,
            notificationsManagement: backupAccount.notificationsManagement,

            privileged: backupAccount.privileged,

            userStorageUsage: backupAccount.userStorageUsage,
            userServerCount: backupAccount.userServerCount,
            UserVmCount: backupAccount.UserVmCount,
            UserWorkstationCount: backupAccount.UserWorkstationCount,
            UserMobileDeviceCount: backupAccount.UserMobileDeviceCount,
            userWebHostingServerCount: backupAccount.userWebHostingServerCount,
            UserWebsiteCount: backupAccount.UserWebsiteCount
        };

        if (backupAccountStatistic) {
            counters.forEach(function (el) {
                if (backupAccount[el] && backupAccountStatistic[el]) {
                    backupAccount[el].usage = backupAccountStatistic[el].usage;
                }
            });
        }
        return getStateful(account);
    }

    return declare(_View, {
        _container: null,

        _user: null,
        _tenant: null,
        _model: null,
        _profile: null,

        _isNewUser: true,
        _needActivation: false,

        _deltas: [],

        _submit: function () {
            var self = this;
            var actions = {};

            if (self._model.aps.id) {
                actions.configureResource = [getPlainValue(self._model)];
            }
            else {
                actions.provisionResource = [getPlainValue(self._model)];
            }

            var operation = {
                'call': {
                    'resourceId': self._tenant.aps.id,
                    'operation': 'setUsersLicensesAfterUpsell',
                    'parameters': {'data': actions}
                }
            };

            return aps.biz.requestUsageChange({
                deltas: self._deltas,
                operations: [operation],
                subscriptionId: self._tenant.aps.subscription
            }).then(function () {
                return aps.biz.commit();
            });
        },
        _renderForm: function () {
            var self = this, form = [];

            self._container.removeAll();

            var fields = [];
            if (self._isNewUser) {
                fields.push(['aps/TextBox', {
                    id: self.genId('login'),
                    label: _('Login'),
                    value: at(self._model, 'endUserLogin'),
                    gridSize: 'md-4 xs-12',
                    required: true,
                    extraValidator: (function () {
                        var result = false,
                            previousValue = '';

                        return function (value) {
                            if (previousValue === value) {
                                return result;
                            }

                            var login = this;
                            xhr.get('/aps/2/resources/' + self._tenant.aps.id + '/loginExists', {
                                query: {login: value, datacenterId: ''},
                                sync: true
                            }).then(function (existing) {
                                if (existing) {
                                    login.set('invalidMessage', _('This login already exists.'));
                                }
                                result = !existing;
                                previousValue = value;
                            }, messages.displayError);

                            return result;
                        };
                    })()
                }]);

                if (self._needActivation) {
                    fields.push(['aps/Password', {
                        id: self.genId('password'),
                        label: _('Password'),
                        value: at(self._model, 'endUserPassword'),
                        gridSize: 'md-4 xs-12',
                        required: true,

                        showResetButton: true,
                        passwordVisible: true,
                        showStrengthIndicator: true,
                        constraints: {
                            minimumLength: 8,
                            maximumLength: 16
                        },
                        generator: utils.passwordGenerator,
                        extraValidator: function (value) {
                            var res = utils.validatePassword(value);
                            if (res === true) {
                                return true;
                            }
                            else {
                                this.set('invalidMessage', res);
                                return false;
                            }
                        }
                    }]);
                }

                form.push(['aps/Panel', {title: _('Set backup user credentials')}, [
                    ['aps/FieldSet', {gridSize: 'md-12 xs-12'}, fields]
                ]]);
            }

            form.push(['aps/Panel', {title: _('Backup user settings')}, [
                ['aps/FieldSet', {gridSize: 'md-12 xs-12'}, [
                    ['aps/Select', {
                        id: self.genId('backup_notifications'),
                        gridSize: 'md-6 xs-12',
                        label: _('Backup Notifications'),
                        value: at(self._model, 'notifyBackup').transform({
                            parse: function (value) {
                                return Number(value);
                            }
                        }),
                        required: true,
                        store: new Memory({data: common.getBackupNotificationOptions(), idProperty: 'value'})
                    }],
                    ['aps/Select', {
                        id: self.genId('business_notifications'),
                        gridSize: 'md-6 xs-12',
                        label: _('Business Notifications'),
                        description: _('Enable notifications about quota excess for this user.'),
                        value: at(self._model, 'notificationsManagement').transform({
                            format: function (value) {
                                return Number(value);
                            },
                            parse: function (value) {
                                return !!Number(value);
                            }
                        }),
                        required: true,
                        store: new Memory({data: common.getEnabledDisabledOptions(), idProperty: 'value'})
                    }],
                    ['aps/Select', {
                        id: self.genId('administrator_privileges'),
                        gridSize: 'md-6 xs-12',
                        label: _('Administrator Privileges'),
                        description: _('An administrator can manage backups of other users and back up Office 365 seats.'),
                        value: at(self._model, 'privileged').transform({
                            parse: function (value) {
                                return Number(value);
                            },
                            format: function (value) {
                                return String(value);
                            }
                        }),
                        required: true,
                        store: new Memory({idProperty: 'value', data: common.getEnabledDisabledOptions()})
                    }]
                ]]
            ]]);

            load(form, self._container).then(function () {
                if (self._needActivation) {
                    var password = registry.byId(self.genId('password'));
                    if (password) {
                        password.generate();
                    }
                }
                if (common.getUserQuota(self._profile)) {
                    getUserResourcesGrid(self._tenant, self._model, 1).then(function (resources) {
                        self._container.addChild(resources.grid);
                        self._deltas = resources.deltas;
                        aps.apsc.hideLoading();
                    }, messages.displayError);
                }
                else {
                    aps.apsc.hideLoading();
                }
            }, messages.displayError);
        },
        init: function () {
            this._container = new Container().placeAt(this.domNode);
        },
        onHide: function () {
            var self = this;
            self._container.removeAll();
        },
        onContext: function (context) {
            var self = this,
                selectedUserId = aps.context.params.userId,
                currentUserId = aps.context.user.aps.id;

            self._isNewUser = !!aps.context.params.isNewUser;
            self._needActivation = self._isNewUser && selectedUserId === currentUserId;

            self._tenant = context.vars.tenant;

            var label = self._isNewUser ? _('Backup User Settings') : _('Change Backup User Settings');
            self.set('label', label);

            all({
                user: xhr.get('/aps/2/resources/' + selectedUserId),
                profile: common.getBackupProfile(self._tenant)
            }).then(function (response) {
                self._user = response.user;
                self._profile = response.profile;
                if (self._isNewUser) {
                    self._model = createModelForNewBackupAccount(self._tenant, self._user, self._profile, self._needActivation);
                    self._renderForm();
                    return;
                }
                getBackupAccount(self._tenant, self._user).then(function (backupAccount) {
                    if (!backupAccount) {
                        messages.removeAll();
                        messages.displayMessage(_('Backup account does not exists for user __displayName__.', self._user), 'error');
                        return;
                    }
                    xhr.get('/aps/2/resources/' + backupAccount.aps.id + '/updateStatistics').then(function (backupAccountStatistic) {
                        self._model = createModelForExistingBackupAccount(backupAccount, backupAccountStatistic);
                        self._renderForm();
                    }, messages.displayError);
                }, messages.displayError);
            }, messages.displayError);
        },
        onSubmit: function () {
            var self = this;
            if (!self.validate()) {
                aps.apsc.cancelProcessing();
                return;
            }
            self._submit().then(function () {
                aps.apsc.prev();
            });
        },
        onCancel: function () {
            aps.apsc.prev();
        }
    });
});

