<?php
/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\Base;

use Acronis\App\Base\ApplicationCounter as Counter;
use Acronis\App\CloudApi\Api;
use Acronis\App\CloudApi\ApiInterface;
use Acronis\App\Services\CountersManager\CountersManagerAwareTrait;
use Acronis\App\Utils\MemoizeTrait;
use Acronis\App\Utils\UomConverter;
use Arr;
use Model_Resource_TenantInterface as TenantStrategyInterface;
use Rest\RestException;
use Service_MspApiClient_ApiInterface as MspApiInterface;
use Str;
use Swagger\Client\Model\OfferingItems\Quota;
use Swagger\Client\Model\Tenants\OfferingItem;
use Swagger\Client\Model\Usages\UsageOutput;

/**
 * todo: ADC APS 3.0. Convert this manager to class "Application" that extends \APS\BaseResource
 * Class ApplicationManager
 *
 * @package Acronis\App\CloudApi
 */
abstract class Application
{
    use MemoizeTrait,
        CountersManagerAwareTrait;

    /** @var string */
    protected $type;
    /** @var string */
    protected $id;
    /** @var \tenant */
    protected $tenant;
    /** @var MspApiInterface */
    private $mspApi;
    /** @var string */
    private $tenantId;
    /** @var TenantStrategyInterface */
    private $tenantStrategy;

    /**
     * Application constructor.
     *
     * @param \tenant $tenant
     * @param MspApiInterface $mspApi
     * @param string|null $tenantId
     * @param TenantStrategyInterface $tenantStrategy
     * @throws RestException
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    public function __construct(
        $tenant,
        MspApiInterface $mspApi,
        $tenantId = null,
        TenantStrategyInterface $tenantStrategy = null
    ) {
        $this->mspApi = $mspApi;
        $this->tenant = $tenant;
        $this->tenantId = $tenantId ?: $tenant->tenantId;
        $this->tenantStrategy = $tenantStrategy;

        // todo: ADC APS 3.0. move to method 'provision'
        $this->type = $this->getApplicationType();
        $this->resolveApplicationId();
    }

    /**
     * todo: ADC APS 3.0. use in method 'configure' for customers and partners
     *
     * @throws RestException
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    public function updateOfferingItems()
    {
        $offeringItems = $this->buildOfferingItemsForApplication();

        if ($this->tenant->type === \tenant::TYPE_RESELLER) {
            $this->enableApplicationOfferingItems($offeringItems);
        } else {
            $this->updateApplicationOfferingItems($offeringItems);
        }
    }

    public function fetchLimitsAndUsages()
    {
        $limits = $this->obtainCountersLimits();
        $usages = $this->obtainCountersUsages();

        return Arr::merge([
            'limit' => $limits,
            'usage' => $usages,
        ], [
            'limit' => 0,
            'usage' => 0,
        ]);
    }

    /**
     * @param string[] $excludedOfferingItems
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    public function deactivateAllTenantOfferingItems(array $excludedOfferingItems = [])
    {
        $tenantId = $this->getTenantId();
        $cloudApi = $this->getCloudApi();

        $tenantOfferingItems = $cloudApi->getTenantApplicationOfferingItems($tenantId, $this->id);

        $excludedOfferingItemsHashtable = array_flip($excludedOfferingItems);

        $offeringItemsToUpdate = [];
        foreach ($tenantOfferingItems as $offeringItem) {
            if ($offeringItem->getStatus() === ApiInterface::OFFERING_ITEM_STATUS_INACTIVE) {
                continue;
            }

            $offeringItemName = $offeringItem->getName();
            if (isset($excludedOfferingItemsHashtable[$offeringItemName])) {
                continue;
            }

            $oi = new OfferingItem();
            $oi->setApplicationId($offeringItem->getApplicationId());
            $oi->setName($offeringItemName);
            $oi->setStatus(ApiInterface::OFFERING_ITEM_STATUS_INACTIVE);
            if ($offeringItem->hasInfraId()) {
                $oi->setInfraId($offeringItem->getInfraId());
            }

            $offeringItemsToUpdate[] = $oi;
        }

        $cloudApi->updateTenantApplicationOfferingItems($tenantId, $offeringItemsToUpdate);
    }

    /**
     * @return TenantStrategyInterface
     */
    protected function getTenantStrategy()
    {
        return $this->tenantStrategy;
    }

    /**
     * @return string
     */
    abstract protected function getApplicationType();

    /**
     * @return Api
     */
    protected function getCloudApi()
    {
        return $this->memoize(function () {
            $mspApi = $this->getMspApi();

            return new Api(
                $mspApi->getUrl(),
                $mspApi->getLogin(),
                $mspApi->getPassword()
            );
        });
    }

    /**
     * @return MspApiInterface
     */
    protected function getMspApi()
    {
        return $this->mspApi;
    }

    /**
     * Returns active offering items only
     *
     * @return OfferingItem[]
     * @throws RestException
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    protected function buildOfferingItemsForApplication()
    {
        $applicationOfferingItems = [];
        $limits = $this->obtainCountersLimits();
        foreach ($limits as $counterName => $limit) {
            // skip counters with limit 0
            if ($limit === 0) {
                continue;
            }

            $counter = $this->getCountersManager()->getCounterByName($counterName);

            $this->checkCounterDependencies($counter, $limits);

            $counterOfferingItems = $this->buildOfferingItemsForCounter($counter, $limit);

            $applicationOfferingItems = array_merge($applicationOfferingItems, $counterOfferingItems);
        }

        return $applicationOfferingItems;
    }

    /**
     * @return array
     * @throws RestException
     */
    protected function obtainCountersLimits()
    {
        $resources = $this->fetchSubscriptionResourcesForApplication();
        $counters = $this->getApplicationCountersForLimits();

        $countersHashtable = array_flip(array_map(function ($counter) {
            /** @var Counter $counter */
            return $counter->getName();
        }, $counters));

        $resources = array_filter($resources, function ($resource) use ($countersHashtable) {
            return Arr::has($countersHashtable, $resource->property);
        });

        return Arr::map($resources, 'property', function ($resource) {
            return $this->getLimitFromSubscriptionResource($resource);
        });
    }

    /**
     * @return array
     * @throws RestException
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    protected function obtainCountersUsages()
    {
        if ($this->tenant->type === \tenant::TYPE_CUSTOMER) {
            return $this->obtainCountersUsagesFromCloud();
        }

        return $this->obtainCountersUsagesFromSubscription();
    }

    /**
     * @return array
     * @throws RestException
     */
    protected function obtainCountersUsagesFromSubscription()
    {
        $resources = $this->fetchSubscriptionResourcesForApplication();
        $counters = $this->getApplicationCountersForUsages();

        $countersHashtable = array_flip(array_map(function ($counter) {
            /** @var Counter $counter */
            return $counter->getName();
        }, $counters));

        $resources = array_filter($resources, function ($resource) use ($countersHashtable) {
            return Arr::has($countersHashtable, $resource->property);
        });

        return Arr::map($resources, 'property', function ($resource) {
            return $this->getUsageFromSubscriptionResource($resource);
        });
    }

    /**
     * @return array
     * @throws RestException
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    protected function obtainCountersUsagesFromCloud()
    {
        $applicationUsages = $this->fetchTenantAggregatedUsages();

        $counters = $this->getApplicationCountersForUsages();

        $countersUsages = [];
        foreach ($counters as $counter) {
            $counterName = $counter->getName();
            $offeringItemName = $counter->getOfferingItemName();
            $offeringItemValue = Arr::get($applicationUsages, $offeringItemName, 0);

            $countersUsages[$counterName] = (int)round($this->getCountersManager()->convertValueFromOfferingItemUnits(
                $counter, $offeringItemValue
            ));
        }

        return $countersUsages;
    }

    /**
     * @param Counter $counter
     * @param int|float|null $limit
     * @return OfferingItem[]
     * @throws RestException
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    protected function buildOfferingItemsForCounter(Counter $counter, $limit)
    {
        if (!$this->checkOfferingItemIsEnabledForRoot($counter->getOfferingItemName())) {
            if ($this->tenant->type === \tenant::TYPE_RESELLER) {
                return [];
            }
            throw new RestException(403, Str::format(
                'Unable to provision resource "%s" if offering item "%s" is not available for tenant "%s".',
                $counter->getName(), $counter->getOfferingItemName(), $this->getRootTenantId()
            ));
        }

        $quotaValue = $counter->hasLimit()
            ? $this->getCountersManager()->convertValueToOfferingItemUnits($counter, $limit)
            : false;

        $offeringItems = [];
        if ($counter->isInfra()) {
            $offeringItemInfras = $this->resolveInfrasForCounter($counter);
            if (empty($offeringItemInfras)) {
                if ($this->tenant->type === \tenant::TYPE_RESELLER) {
                    return [];
                }
                throw new RestException(404, Str::format(
                    'No infrastructure components are available for offering item "%s".',
                    $counter->getOfferingItemName()
                ));
            }
            foreach ($offeringItemInfras as $infraId) {
                $offeringItems[] = $this->createOfferingItem($counter->getOfferingItemName(), $quotaValue, $infraId);
            }
        } else {
            $offeringItems[] = $this->createOfferingItem($counter->getOfferingItemName(), $quotaValue);
        }

        foreach ($counter->getSubOfferingItems() as $offeringItemName) {
            if (!$this->checkOfferingItemIsSupportedForTenantType($offeringItemName)) {
                continue;
            }

            if (!$this->checkOfferingItemIsEnabledForRoot($offeringItemName)) {
                continue;
            }

            $offeringItems[] = $this->createOfferingItem($offeringItemName);
        }

        return $offeringItems;
    }

    /**
     * @return object[]
     * @throws RestException
     */
    protected function fetchSubscriptionResourcesForApplication()
    {
        return $this->memoize(function () {
            $resources = $this->tenant->subscription->resources();

            // get supported counters only
            $counters = $this->getApplicationCounters();
            $countersHashtable = array_flip(array_map(function ($counter) {
                /** @var Counter $counter */
                return $counter->getName();
            }, $counters));

            // get counters only
            $apsType = dirname($this->tenant->aps->type);
            $resources = array_filter($resources, function ($resource) use ($apsType, $countersHashtable) {
                return isset($resource->apsType, $resource->property) &&
                    dirname($resource->apsType) === $apsType &&
                    Arr::has($countersHashtable, $resource->property);
            });

            //check that counters are unique
            $this->checkCountersAreUnique($resources);

            return $resources;
        });
    }

    protected function getLimitFromSubscriptionResource($resource)
    {
        if (!isset($resource->limit)) {
            return Counter::UNLIMITED_VALUE;
        }

        $limit = (int)$resource->limit;
        if ($limit === 0) {
            return 0;
        }

        // Storage are always measured in Kb in subscription resources so convert them to bytes as defined in counters
        if (UomConverter::formatMeasurementName($resource->unit) === UomConverter::KILOBYTES) {
            $limit = UomConverter::convertStorage($limit, UomConverter::KILOBYTES, UomConverter::BYTES);
        }

        return $limit;
    }

    protected function getUsageFromSubscriptionResource($resource)
    {
        if (!isset($resource->usage)) {
            return 0;
        }

        $usage = (int)$resource->usage;
        if ($usage === 0) {
            return 0;
        }

        // Storage are always measured in Kb in subscription resources so convert them to bytes as defined in counters
        if (UomConverter::formatMeasurementName($resource->unit) === UomConverter::KILOBYTES) {
            $usage = UomConverter::convertStorage($usage, UomConverter::KILOBYTES, UomConverter::BYTES);
        }

        return $usage;
    }

    /**
     * @param string $offeringItemName
     * @param int|false $quotaValue
     * @param string|false $infraId
     * @return OfferingItem
     */
    protected function createOfferingItem($offeringItemName, $quotaValue = false, $infraId = false)
    {
        $offeringItem = new OfferingItem();
        $offeringItem->setApplicationId($this->id);
        $offeringItem->setName($offeringItemName);

        if ($infraId !== false) {
            $offeringItem->setInfraId($infraId);
        }

        if ($quotaValue === false) {
            return $offeringItem;
        }

        $quota = new Quota();
        $quota->setValue($quotaValue);
        $quota->setOverage(0);

        $offeringItem->setQuota($quota);

        return $offeringItem;
    }

    /**
     * @param Counter $counter
     * @return mixed
     */
    abstract protected function resolveInfrasForCounter(Counter $counter);

    protected function getTenantId()
    {
        return $this->tenantId;
    }

    /**
     * @return string
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    protected function getRootTenantId()
    {
        return $this->memoize(function () {
            $cloudApi = $this->getCloudApi();
            $me = $cloudApi->getMe();

            return $me->getTenantId();
        });
    }

    protected function getApplicationOfferingItemsNames()
    {
        return $this->memoize(function () {
            $counters = $this->getApplicationCounters();

            $offeringItemsNames = [];
            foreach ($counters as $counter) {
                $offeringItemsNames[] = $counter->getOfferingItemName();
                $offeringItemsNames = array_merge(
                    $offeringItemsNames,
                    $counter->getSubOfferingItems()
                );
            }

            return array_unique($offeringItemsNames);
        });
    }

    /**
     * @return Counter[]
     */
    protected function getApplicationCounters()
    {
        return $this->memoize(function () {
            $counters = $this->getCountersManager()->getApplicationCounters($this->type);

            return array_filter($counters, function ($counter) {
                /** @var Counter $counter */
                $tenantTypes = $counter->getTenantTypes();

                return in_array($this->tenant->type, $tenantTypes);
            });
        });
    }

    /**
     * @return Counter[]
     */
    protected function getApplicationCountersForUsages()
    {
        return $this->memoize(function () {
            $counters = $this->getApplicationCounters();
            $counters = array_filter($counters, function ($counter) {
                /** @var Counter $counter */
                return $counter->hasUsage();
            });

            return $counters;
        });
    }

    /**
     * @return Counter[]
     */
    protected function getApplicationCountersForLimits()
    {
        return $this->memoize(function () {
            $counters = $this->getApplicationCounters();
            $counters = array_filter($counters, function ($counter) {
                /** @var Counter $counter */
                return $counter->isFeature()
                    || $counter->hasLimit();
            });

            return $counters;
        });
    }

    protected function checkOfferingItemIsSupportedForTenantType($offeringItemName)
    {
        $cloudApi = $this->getCloudApi();
        switch ($this->tenant->type) {
            case \tenant::TYPE_CUSTOMER:
                return $cloudApi->isOfferingItemForCustomer($offeringItemName);
            case \tenant::TYPE_PARTNER:
            case \tenant::TYPE_RESELLER:
                return $cloudApi->isOfferingItemForPartner($offeringItemName);
        }

        throw new \RuntimeException(Str::format(
            'Unsupported tenant type "%s".',
            $this->tenant->type
        ));
    }

    /**
     * @param string $offeringItemName
     * @return boolean
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    protected function checkOfferingItemIsEnabledForRoot($offeringItemName)
    {
        $offeringItems = $this->memoize(function () {
            $cloudApi = $this->getCloudApi();
            $me = $cloudApi->getMe();

            return $cloudApi->getTenantApplicationOfferingItems($me->getTenantId(), $this->id);
        });

        $offeringItems = array_filter($offeringItems, function ($offeringItem) use ($offeringItemName) {
            /** @var OfferingItem $offeringItem */
            return $offeringItem->getStatus() === Api::OFFERING_ITEM_STATUS_ACTIVE
                && $offeringItem->getName() === $offeringItemName;
        });

        return count($offeringItems) > 0;
    }

    /**
     * @param string $offeringItemName
     * @return boolean
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    protected function checkOfferingItemIsEnabledForTenant($offeringItemName)
    {
        $offeringItems = $this->getTenantApplicationOfferingItems();

        $offeringItems = array_filter($offeringItems, function ($offeringItem) use ($offeringItemName) {
            /** @var OfferingItem $offeringItem */
            return $offeringItem->getStatus() === Api::OFFERING_ITEM_STATUS_ACTIVE
                && $offeringItem->getName() === $offeringItemName;
        });

        return count($offeringItems) > 0;
    }

    /**
     * @param Counter $counter
     * @param array $limits
     * @throws RestException
     */
    protected function checkCounterDependencies($counter, array $limits)
    {
        // do not check dependencies for resellers
        if ($this->tenant->type === \tenant::TYPE_RESELLER) {
            return;
        }
        foreach ($counter->getDependencies() as $requiredCounterName) {
            $limit = (int)Arr::get($limits, $requiredCounterName, 0);
            if ($limit === 0) {
                throw new RestException(424, Str::format(
                    'Unable to provision resource "%s" without "%s". Please update the service template.',
                    $counter->getName(), $requiredCounterName
                ));
            }
        }
    }

    /**
     * @param bool $force
     * @return \Swagger\Client\Model\OfferingItems\OfferingItemOutput[]
     */
    protected function getTenantApplicationOfferingItems($force = false)
    {
        return $this->memoize(function () {
            $cloudApi = $this->getCloudApi();

            return $cloudApi->getTenantApplicationOfferingItems($this->getTenantId(), $this->id);
        }, '', $force);
    }

    /**
     * @return UsageOutput[]
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    protected function getTenantApplicationUsages()
    {
        return $this->memoize(function () {
            $usages = $this->getCloudApi()->getTenantUsages($this->getTenantId());

            return array_values(array_filter($usages, function ($usage) {
                /** @var UsageOutput $usage */
                return $usage->getApplicationId() === $this->id;
            }));
        });
    }

    /**
     * @param bool $absolute
     * @return array
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    protected function fetchTenantAggregatedUsages($absolute = false)
    {
        $tenantUsages = $this->getTenantApplicationUsages();

        $aggregatedUsages = [];
        foreach ($tenantUsages as $offeringItemUsage) {
            // aggregate all usages by a offering item name
            $offeringItemName = $offeringItemUsage->getName();
            $offeringItemValue = $absolute
                ? $offeringItemUsage->getAbsoluteValue()
                : $offeringItemUsage->getValue();
            if (isset($aggregatedUsages[$offeringItemName])) {
                $aggregatedUsages[$offeringItemName] += $offeringItemValue;
            } else {
                $aggregatedUsages[$offeringItemName] = $offeringItemValue;
            }
        }

        return $aggregatedUsages;
    }

    /**
     * @param OfferingItem[] $offeringItems
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    protected function enableApplicationOfferingItems(array $offeringItems)
    {
        $cloudApi = $this->getCloudApi();
        $cloudApi->enableApplicationOfferingItemsForTenantsChain(
            $this->getRootTenantId(),
            $this->getTenantId(),
            $this->id,
            $offeringItems
        );
    }

    /**
     * @param OfferingItem[] $offeringItemsToEnable
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    protected function updateApplicationOfferingItems(array $offeringItemsToEnable)
    {
        // Disable other offering items
        $offeringItemsToUpdate = array_merge(
            $offeringItemsToEnable,
            $this->resolveOfferingItemsToDisable($offeringItemsToEnable)
        );

        $cloudApi = $this->getCloudApi();
        $cloudApi->updateApplicationOfferingItemsForTenantsChain(
            $this->getRootTenantId(),
            $this->getTenantId(),
            $this->id,
            $offeringItemsToUpdate
        );
    }

    /**
     * @param OfferingItem[] $offeringItemsToEnable
     * @return OfferingItem[]
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    private function resolveOfferingItemsToDisable(array $offeringItemsToEnable)
    {
        $offeringItemsToEnableHashtable = Api::getOfferingItemsHashTable($offeringItemsToEnable);
        $supportedOfferingItemsNamesHashTable = array_fill_keys(
            $this->getApplicationOfferingItemsNames(),
            true
        );

        $offeringItemsToDisable = [];
        $tenantOfferingItems = $this->getTenantApplicationOfferingItems();
        foreach ($tenantOfferingItems as $tenantOfferingItem) {
            $offeringItemHash = Api::getOfferingItemHash($tenantOfferingItem);
            if (Arr::has($offeringItemsToEnableHashtable, $offeringItemHash)) {
                continue;
            }

            // never touch not supported offering items
            if (!Arr::has($supportedOfferingItemsNamesHashTable, $tenantOfferingItem->getName())) {
                continue;
            }

            // don't touch already disabled offering items
            if ($tenantOfferingItem->getStatus() === ApiInterface::OFFERING_ITEM_STATUS_INACTIVE) {
                continue;
            }

            $offeringItem = $this->createOfferingItem(
                $tenantOfferingItem->getName(),
                $tenantOfferingItem->hasQuota() ? 0 : false,
                $tenantOfferingItem->hasInfraId() ? $tenantOfferingItem->getInfraId() : false
            );

            // always disable features otherwise set zero quota
            if ($tenantOfferingItem->getType() === ApiInterface::OFFERING_ITEM_TYPE_FEATURE) {
                $offeringItem->setStatus(ApiInterface::OFFERING_ITEM_STATUS_INACTIVE);
            }

            $offeringItemsToDisable[] = $offeringItem;
        }

        return $offeringItemsToDisable;
    }

    /**
     * @throws RestException
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    private function resolveApplicationId()
    {
        $application = $this->getCloudApi()->getApplicationByType($this->type);
        if (is_null($application)) {
            throw new RestException(404, Str::format(
                'Cannot fetch application "%s".',
                $this->type
            ));
        }
        $this->id = $application->getId();
    }

    /**
     * @param array $resources
     * @throws RestException
     */
    private function checkCountersAreUnique(array $resources)
    {
        $hashtable = [];
        foreach ($resources as $resource) {
            $counterName = $resource->property;
            if (isset($hashtable[$counterName])) {
                $conflictingResource = $hashtable[$counterName];
                throw new RestException(409, Str::format(
                    'Resources "%s" and "%s" have the same application counter "%s". Please remove one of them from the service template.',
                    $conflictingResource->id, $resource->id, $counterName
                )
                );
            }
            $hashtable[$counterName] = $resource;
        }
    }
}