<?php
/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\Base;

use Acronis\App\Base\ApplicationCounter as Counter;
use Acronis\App\CloudApi\Api;
use Acronis\App\CloudApi\ApiInterface;
use Arr;
use Prophecy\Exception\InvalidArgumentException;
use Str;
use Swagger\Client\Model\Infra\Infra;
use Swagger\Client\Model\Infra\InfraCap;
use Swagger\Client\Model\Tenants\OfferingItem;

/**
 * todo: ADC APS 3.0. Convert this manager to class "BackupApplication" that extends \APS\BaseResource
 * Class BackupApplicationManager
 *
 * @package Acronis\App\CloudApi
 */
class BackupApplication extends Application
{
    /* ======= RESOURCE PROPERTIES ======= */

    // todo: ADC APS 3.0. Add resource properties here

    /* ======= RESOURCE COUNTERS ======= */

    // todo: ADC APS 3.0. Add counters here

    /* ======= OTHER PROPERTIES ======= */

    /* ======= RESOURCE METHODS ======= */

    /**
     * @return string
     */
    protected function getApplicationType()
    {
        return ApiInterface::APPLICATION_TYPE_BACKUP;
    }

    /**
     * @return array
     * @throws \Rest\RestException
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    protected function obtainCountersLimits()
    {
        $limits = parent::obtainCountersLimits();

        if ($this->tenant->type === \tenant::TYPE_RESELLER) {
            return $limits;
        }

        // todo: ADC APS 3.0. remove temporary logic
        // business logic to keep the backward capability

        $this->addMailboxInstanceSizeCounter($limits);

        $this->addLocalStorageUsageCounter($limits);

        $this->addO365MailboxesCounter($limits);

        $this->addO365OneDriveCounter($limits);

        $this->addO365SharePointSitesCounter($limits);

        return $limits;
    }

    /**
     * @param ApplicationCounter $counter
     * @return string[]
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    protected function resolveInfrasForCounter(Counter $counter)
    {
        $infraCapability = $counter->getInfraCapability();

        if (is_null($infraCapability)) {
            return [];
        }

        return $this->getInfrasUuidsForCapability($infraCapability);
    }

    /**
     * @param string $infraCapability
     * @return string[]
     */
    protected function getInfrasUuidsForCapability($infraCapability)
    {
        return $this->memoize(function () use ($infraCapability) {
            $offeringItemsNames = $this->getOfferingItemsNamesByCapability($infraCapability);
            $infrasUuids = $this->getTenantInfrasUuidsByOfferingItemsNames($offeringItemsNames);

            if (!empty($infrasUuids)) {
                return $infrasUuids;
            }

            return $this->resolveInfrasUuidsForCapability($infraCapability);
        }, $infraCapability);
    }

    /**
     * @param string $infraCapability
     * @return string[]
     */
    protected function getOfferingItemsNamesByCapability($infraCapability)
    {
        $counters = $this->getApplicationCountersForLimits();
        $counters = array_values(array_filter(
            $counters,
            function ($counter) use ($infraCapability) {
                /** @var Counter $counter */
                return $counter->getInfraCapability() === $infraCapability;
            }
        ));

        return array_map(function ($counter) {
            /** @var Counter $counter */
            return $counter->getOfferingItemName();
        }, $counters);
    }

    /**
     * @param string[] $offeringItemsNames
     * @return string[]
     */
    protected function getTenantInfrasUuidsByOfferingItemsNames(array $offeringItemsNames)
    {
        $offeringItems = $this->getTenantApplicationOfferingItems();
        $offeringItemsNamesHashtable = array_fill_keys($offeringItemsNames, true);
        $offeringItems = array_values(array_filter(
            $offeringItems,
            function ($offeringItem) use ($offeringItemsNamesHashtable) {
                /** @var OfferingItem $offeringItem */
                return isset($offeringItemsNamesHashtable[$offeringItem->getName()])
                    && $offeringItem->getStatus() === Api::OFFERING_ITEM_STATUS_ACTIVE
                    && $offeringItem->hasInfraId();
            }
        ));

        return array_unique(array_map(function ($offeringItem) {
            /** @var OfferingItem $offeringItem */
            return $offeringItem->getInfraId();
        }, $offeringItems));
    }

    /**
     * @param $infraCapability
     * @return string[]
     * @throws \Rest\RestException
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    protected function resolveInfrasUuidsForCapability($infraCapability)
    {
        if ($infraCapability === InfraCap::BACKUP) {
            return $this->resolveBackupStorageInfrasUuids();
        }

        if ($infraCapability === InfraCap::DISASTER_RECOVERY) {
            return $this->resolveDisasterRecoveryInfrasUuids();
        }

        throw new InvalidArgumentException(Str::format(
            'Unsupported infrastructure component capability "%s".',
            $infraCapability
        ));
    }

    /**
     * @param array $limits
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    private function addMailboxInstanceSizeCounter(array &$limits)
    {
        // do nothing if the counter is already specified in the service plan
        if (Arr::has($limits, \tenant::MAILBOX_INSTANCE_SIZE)) {
            return;
        }

        $mailboxCount = Arr::get($limits, \tenant::MAILBOX_COUNT, 0);
        if ($mailboxCount === 0) {
            return;
        }

        if ($this->tenant->type !== \tenant::TYPE_CUSTOMER) {
            if ($this->checkOfferingItemIsEnabledForRoot(ApiInterface::OFFERING_ITEM_MAILBOX_INSTANCE_SIZE)) {
                $limits[\tenant::MAILBOX_INSTANCE_SIZE] = Counter::UNLIMITED_VALUE;
            };

            return;
        }

        $tenant = $this->tenant;
        if ($tenant->profile->isEmpty() || !$tenant->profile->freemiumMailboxStorage) {
            return;
        }

        $storageSize = Arr::get($limits, \tenant::STORAGE_USAGE, 0);
        $mailboxInstanceSize = $tenant->profile->freemiumMailboxStorage;

        $limits[\tenant::MAILBOX_INSTANCE_SIZE] = $this->computeMailboxInstanceStorageSizeForTenant(
            $storageSize, $mailboxCount, $mailboxInstanceSize
        );
    }

    /**
     * @param array $limits
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    private function addO365MailboxesCounter(array &$limits)
    {
        // do nothing if the counter is already specified in the service plan
        if (Arr::has($limits, \tenant::O365_MAILBOXES)) {
            return;
        }

        // do nothing if there are no Office 365 seats
        if (Arr::get($limits, \tenant::MAILBOX_COUNT, 0) === 0) {
            return;
        }

        // do nothing if the offering item is not enabled for root
        if (!$this->checkOfferingItemIsEnabledForRoot(ApiInterface::OFFERING_ITEM_O365_MAILBOXES)) {
            return;
        };

        $limits[\tenant::O365_MAILBOXES] = Counter::UNLIMITED_VALUE;
    }

    /**
     * @param array $limits
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    private function addO365OneDriveCounter(array &$limits)
    {
        // do nothing if the counter is already specified in the service plan
        if (Arr::has($limits, \tenant::O365_ONE_DRIVE)) {
            return;
        }

        // do nothing if there are no Office 365 seats
        if (Arr::get($limits, \tenant::MAILBOX_COUNT, 0) === 0) {
            return;
        }

        // do nothing if the offering item is not enabled for root
        if (!$this->checkOfferingItemIsEnabledForRoot(ApiInterface::OFFERING_ITEM_O365_ONEDRIVE)) {
            return;
        };

        $limits[\tenant::O365_ONE_DRIVE] = Counter::UNLIMITED_VALUE;
    }

    /**
     * @param array $limits
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    private function addO365SharePointSitesCounter(array &$limits)
    {
        // do nothing if the counter is already specified in the service plan
        if (Arr::has($limits, \tenant::O365_SHARE_POINT_SITES_COUNT)) {
            return;
        }

        // do nothing if there are no Office 365 seats
        if (Arr::get($limits, \tenant::MAILBOX_COUNT, 0) === 0) {
            return;
        }

        // do nothing if the offering item is not enabled for root
        if (!$this->checkOfferingItemIsEnabledForRoot(ApiInterface::OFFERING_ITEM_O365_SHAREPOINT_SITES)) {
            return;
        };

        $limits[\tenant::O365_SHARE_POINT_SITES_COUNT] = Counter::UNLIMITED_VALUE;
    }

    /**
     * @param array $limits
     */
    private function addLocalStorageUsageCounter(array &$limits)
    {
        // do nothing if the counter is already specified in the service plan
        if (Arr::has($limits, \tenant::LOCAL_STORAGE_USAGE)) {
            return;
        }

        $profile = $this->tenant->profile;
        if ($profile->isEmpty() || !$profile->storageLocation) {
            return;
        }

        $limits[\tenant::LOCAL_STORAGE_USAGE] = Counter::UNLIMITED_VALUE;
    }

    /**
     * @param int|null $storageSize
     * @param int|null $mailboxCount
     * @param int|null $mailboxInstanceSize
     * @return int|null
     */
    private function computeMailboxInstanceStorageSizeForTenant($storageSize, $mailboxCount, $mailboxInstanceSize)
    {
        if ($mailboxCount === Counter::UNLIMITED_VALUE ||
            $storageSize === Counter::UNLIMITED_VALUE ||
            $mailboxInstanceSize === Counter::UNLIMITED_VALUE
        ) {
            return Counter::UNLIMITED_VALUE;
        }

        if ($storageSize > $mailboxCount * $mailboxInstanceSize) {
            return Counter::UNLIMITED_VALUE;
        }

        return $mailboxInstanceSize;
    }

    /**
     * @return string[]
     * @throws \Rest\RestException
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    private function resolveBackupStorageInfrasUuids()
    {
        $rootTenantId = $this->getRootTenantId();
        $leafTenantId = $this->getTenantId();

        $storages = $this->fetchBackupStorageUuidsForTenantsChain($rootTenantId, $leafTenantId);
        if ($this->tenant->type !== \tenant::TYPE_CUSTOMER) {
            return array_values($storages);
        }

        $tenantStrategy = $this->getTenantStrategy();
        $storageId = $tenantStrategy->resolveBackupStorageId();

        $storageUuid = Arr::get($storages, $storageId);
        if (is_null($storageUuid)) {
            throw new \Rest\RestException(404, Str::format(
                'There is no backup storage with ID %s.',
                $storageId
            ));
        }

        return [$storageUuid];
    }

    /**
     * @param $rootTenantId
     * @param $leafTenantId
     * @return string[]
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    private function fetchBackupStorageUuidsForTenantsChain($rootTenantId, $leafTenantId)
    {
        $cloudApi = $this->getCloudApi();
        $tenantPath = $cloudApi->resolveTenantPath($rootTenantId, $leafTenantId);

        $mspApi = $this->getMspApi();
        $storages = [];
        foreach ($tenantPath as $tenantId) {
            $storages = array_merge($storages, $mspApi->getStorages($tenantId));
        }

        return Arr::map(
            $storages,
            function ($storage) {
                /** @var \Service_MspApiClient_Response_Struct_Storage $storage */
                return $storage->getId();
            },
            function ($storage) {
                /** @var \Service_MspApiClient_Response_Struct_Storage $storage */
                return $storage->getUuid();
            },
            false
        );
    }

    /**
     * @return string[]
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    private function resolveDisasterRecoveryInfrasUuids()
    {
        $rootTenantId = $this->getRootTenantId();
        $leafTenantId = $this->getTenantId();
        $cloudApi = $this->getCloudApi();
        $availableInfras = $cloudApi->fetchInfrasForTenantsChain($rootTenantId, $leafTenantId);

        if ($this->tenant->type === \tenant::TYPE_RESELLER) {
            $disasterRecoveryInfras = array_filter($availableInfras, function($infra) {
                /** @var Infra $infra */
                return in_array(InfraCap::DISASTER_RECOVERY, $infra->getCapabilities());
            });
            return array_map(function($infra) {
                /** @var Infra $infra */
                return $infra->getId();
            }, $disasterRecoveryInfras);
        }

        $backupStorageInfrasUuids = $this->getInfrasUuidsForCapability(InfraCap::BACKUP);
        $backupStorageInfras = array_filter($availableInfras, function ($infra) use ($backupStorageInfrasUuids) {
            /** @var Infra $infra */
            return in_array($infra->getId(), $backupStorageInfrasUuids);
        });
        if (empty($backupStorageInfras)) {
            return [];
        }

        $locationsUuids = array_map(function ($infra) {
            /** @var Infra $infra */
            return $infra->getLocationId();
        }, $backupStorageInfras);

        $disasterRecoveryInfras = array_filter($availableInfras, function ($infra) use ($locationsUuids) {
            /** @var Infra $infra */
            return in_array($infra->getLocationId(), $locationsUuids)
                && in_array(InfraCap::DISASTER_RECOVERY, $infra->getCapabilities());
        });
        if (empty($disasterRecoveryInfras)) {
            return [];
        }

        $disasterRecoveryInfrasUuids = array_map(function ($infra) {
            /** @var Infra $infra */
            return $infra->getId();
        }, $disasterRecoveryInfras);

        return array_values($disasterRecoveryInfrasUuids);
    }
}