<?php
/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\CloudApi;

use Acronis\App\Utils\CacheableTrait;
use Swagger\Client\ApiException;
use Swagger\Client\HttpException;
use Swagger\Client\HttpResponseInterface;
use Swagger\Client\Model\Idp\Token;

abstract class AuthorizedApi extends AwareApi
{
    use CacheableTrait;

    /**
     * @return string
     */
    abstract public function getLogin();

    /**
     * @return string
     */
    abstract public function getPassword();

    /**
     * @param callable $fn
     * @return mixed
     */
    protected function call(callable $fn)
    {
        /** @var HttpResponseInterface $response */
        $response = $fn();

        if (is_object($response) && $response instanceof HttpResponseInterface) {
            return $response->getResponseBody();
        }

        return $response;
    }

    /**
     * @param callable $fn
     * @return mixed
     * @throws HttpException
     * @throws ApiException
     */
    protected function authorizedCall(callable $fn)
    {
        if (!$this->isAuthorized()) {
            $this->authorize();
        }
        try {
            return $this->call($fn);
        } catch (HttpException $e) {
            if ($e->getCode() != 401) {
                throw $e;
            }
            $this->authorize(true);
        }

        return $this->call($fn);
    }

    /**
     * @param bool $force
     * @return string
     */
    protected function obtainAccessToken($force = false)
    {
        return $this->fromCache(function () {
            /** @var Token $idpToken */
            $idpToken = $this->call(function () {
                return $this->getIdpApi()->postIdpToken(
                    'password', null, $this->getLogin(), $this->getPassword()
                );
            });

            return $idpToken->getAccessToken();
        }, $this->getAccessTokenTtl(), true, $this->getCredentialsHash(), $force);
    }

    abstract protected function getCredentialsHash();

    private function isAuthorized()
    {
        $token = $this->getApiClient()
            ->getConfig()
            ->getAccessToken();

        return !empty($token);
    }

    private function authorize($force = false)
    {
        $this->getApiClient()
            ->getConfig()
            ->setAccessToken($this->obtainAccessToken($force));
    }
}