<?php

/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */
class Model_ApplicationInitializationManager
{
    const ATTRIBUTE_VAR_VALUE = 'var_value';
    const ATTRIBUTE_VAR_NAME = 'var_name';
    const ATTRIBUTE_APPLICATION_RESOURCE_ID = 'application_resource_id';
    const ATTRIBUTE_RESOURCE_ID = 'resource_id';
    const ATTRIBUTE_APP_ID = 'app_id';
    const ATTRIBUTE_SERVICE_ID = 'service_id';

    const APS_STATUS_READY = 'aps:ready';
    const APS_CLASS_NAME_RESOURCE = 'rc.saas.resource';

    const LOG_RESOURCE_TYPE_CREATED = 'Resource type "%s" has been created with ID %s.';
    const LOG_SCREEN_STORE_CREATED = 'Screen store with name "%s" created ';
    const LOG_RESOURCE_CATEGORY_CREATED = 'Resource category created with name "%s".';
    const LOG_SERVICE_PLAN_CREATED =  'Service plan created with name "%s".';
    const DEFAULT_APPLICATION_TYPE_ID = 1;
    const FIELD_PARAMETER_ID = 'paramId';
    const FIELD_VALUE = 'value';
    const DEFAULT_DATACENTER_NAME = 'Acronis Backup Cloud';

    /**
     * @var Service_OpenApi_ApiInterface
     */
    private $openApi;

    /**
     * @var \APS\ControllerProxy
     */
    private $apsc;


    /**
     * Model_ResourceManager constructor.
     * @param Service_OpenApi_ApiInterface $openApi
     * @param $apsc
     */
    public function __construct(Service_OpenApi_ApiInterface $openApi, $apsc)
    {
        $this->openApi = $openApi;
        $this->apsc = $apsc;
    }

    /**
     * @param $applicationResourceId
     * @param $resourcesConfig
     * @param $configuration
     */
    public function initializeInstance($applicationResourceId, $resourcesConfig, $configuration)
    {
        $types = array();
        $this->configurationProcessing($configuration, $types);
        $applicationId = $this->openApi->apsGetApplicationId(APSResource::PACKAGE_ID);

        $applicationTypeId = $this->openApi->addApplicationInstance(
            $resourcesConfig->application,
            $applicationResourceId,
            $applicationId
        );

        Log::getInstance()->event(
            self::LOG_RESOURCE_TYPE_CREATED,
            $resourcesConfig->application,
            $applicationTypeId
        );

        foreach ($resourcesConfig->services as $service) {
            $this->createService($applicationId, $service, $types);
            $this->createCountersForService($service, $applicationId, $types);
        }

        foreach ($resourcesConfig->profiles as $profile) {
            $newProfileId = $this->createBackupProfile($profile->resource, $types);
            Log::getInstance()->event(
                self::LOG_RESOURCE_TYPE_CREATED,
                $profile->name,
                $newProfileId
            );
            $types[$profile->id] = $newProfileId;
        }

        foreach ($resourcesConfig->templates as $template) {
            $types[$template->id] = $this->createServiceTemplate(
                $applicationTypeId,
                $template,
                $types
            );
        }


        $this->changedUnitMeasurements($resourcesConfig, $types);

        $resourceCategories = $this->createResourceCategories($resourcesConfig);

        foreach ($resourcesConfig->storeScreens as $storeScreen) {
            $this->createStoreScreen($storeScreen);
        }

        foreach ($resourcesConfig->purchaseScenarios as $purchaseScenario) {
            $this->openApi->addPurchaseScenario(
                $purchaseScenario->id,
                $purchaseScenario->name,
                $purchaseScenario->orderNum,
                ADDITIONAL_FILE_PATH . $purchaseScenario->imageFile,
                $purchaseScenario->storeScreens
            );
        }

        $this->createServicePlans($resourcesConfig, $types, $resourceCategories);

        Log::getInstance()->event('Application "%s" has been successfully initialized.', $resourcesConfig->application);
    }

    /**
     * @param $resource
     * @param $types
     * @return mixed
     */
    private function createBackupProfile($resource, $types)
    {
        $newProfile = APS\TypeLibrary::newResourceByTypeId($resource->aps->type);
        $newProfile->name = $resource->name;
        $newProfile->backupNotification = $resource->backupNotification;
        $newProfile->managementNotification = $resource->managementNotification;
        $newProfile->storages = $types['storages'];
        $newProfile->userQuota = $resource->userQuota;
        $newProfile->storageLocation = $resource->storageLocation;
        $newProfile->storagePolicy = $resource->storagePolicy;

        $profile = $this->apsc->provisionResource($newProfile);

        while ($profile->aps->status != self::APS_STATUS_READY) {
            sleep(2);
            $profile = $this->apsc->getResource($profile->aps->id);
        }

        return $profile->resourceTypeId;
    }


    /**
     * @param $resourceConfig
     * @param $resourceMap
     * @param $resourceCategories
     */
    public function createServicePlans($resourceConfig, &$resourceMap, $resourceCategories)
    {
        foreach ($resourceConfig->plans as $plan) {
            $this->createServicePlan($resourceMap, $resourceCategories, $plan);
        }

        foreach ($resourceConfig->planUpgrade as $planUpgrade) {
            $this->openApi->addPlanUpgrade(
                $resourceMap[$planUpgrade->from],
                $resourceMap[$planUpgrade->to]
            );

        }
    }

    /**
     * @param $applicationId
     * @param $service
     * @param $types
     */
    protected function createService($applicationId, $service, &$types)
    {
        $newServiceTypeId = $this->openApi->addServiceType(
            $applicationId,
            $service->name,
            $service->type,
            $service->autoprovide
        );

        $types[$service->id] = $newServiceTypeId;
        Log::getInstance()->event(
            self::LOG_RESOURCE_TYPE_CREATED,
            $service->name,
            $newServiceTypeId
        );
    }

    /**
     * @param $applicationTypeId
     * @param $template
     * @param $types
     * @return int
     */
    protected function createServiceTemplate($applicationTypeId, $template, &$types)
    {
        $resources = array(
            $applicationTypeId => self::DEFAULT_APPLICATION_TYPE_ID
        );
        foreach ($template->resources as $id => $resource) {
            $resources[$types[$id]] = $resource;
        }

        $serviceTemplateId = $this->openApi->addActivatedServiceTemplate(
            $template->name,
            $resources,
            $template->category
        );
        Log::getInstance()->event(
            'Service template "%s" has been created with ID %s."',
            $template->name,
            $serviceTemplateId
        );
        return $serviceTemplateId;
    }

    /**
     * @param $service
     * @param $applicationId
     * @param $types
     */
    protected function createCountersForService($service, $applicationId, &$types)
    {
        foreach ($service->counters as $counter) {
            $attr = array(
                array(
                    self::ATTRIBUTE_VAR_NAME => self::ATTRIBUTE_RESOURCE_ID,
                    self::ATTRIBUTE_VAR_VALUE => $counter->type
                ),
                array(
                    self::ATTRIBUTE_VAR_NAME => self::ATTRIBUTE_SERVICE_ID,
                    self::ATTRIBUTE_VAR_VALUE => $service->type
                ),
                array(
                    self::ATTRIBUTE_VAR_NAME => self::ATTRIBUTE_APP_ID,
                    self::ATTRIBUTE_VAR_VALUE => (string)$applicationId,
                )
            );
            $reClassName = $counter->unit ? self::APS_CLASS_NAME_RESOURCE . '.' . $counter->unit : self::APS_CLASS_NAME_RESOURCE;
            $newCounterId = $this->openApi->addResourceType($reClassName, $counter->name, null, null, $attr);
            Log::getInstance()->event(
                self::LOG_RESOURCE_TYPE_CREATED,
                $counter->name,
                $newCounterId
            );
            $types[$counter->id] = $newCounterId;
        }
    }

    /**
     * @param $storeScreen
     */
    protected function createStoreScreen($storeScreen)
    {
        $parameters = array();
        foreach ($storeScreen->storeScreenParams as $storeScreenParameter) {
            $parameters[] = array(
                self::FIELD_PARAMETER_ID => $storeScreenParameter->paramId,
                self::FIELD_VALUE => $storeScreenParameter->valueStr
            );
        }


        $saleCategories = array();
        foreach ($storeScreen->salesCategories as $salesCategory) {
            $newCategory = $salesCategory;
            $newCategory->imageFile = ADDITIONAL_FILE_PATH . $newCategory->imageFile;
            $saleCategories[] = $newCategory;
        }

        $this->openApi->addStoreScreen(
            $storeScreen->id,
            $storeScreen->storeTemplateId,
            $storeScreen->title,
            $parameters,
            $saleCategories
        );
        Log::getInstance()->event(
            self::LOG_SCREEN_STORE_CREATED,
            $storeScreen->name
        );
    }

    /**
     * @param $resourceMap
     * @param $resourceCategory
     * @param $plan
     */
    protected function createServicePlan(&$resourceMap, $resourceCategories, $plan)
    {
        $resourceRates = array();
        foreach ($plan->resourceRates as $resourceRate) {
            $resourceRate->resourceId = $resourceMap[$resourceRate->resourceId];
            $resourceRate->resourceCategory = $resourceCategories[$resourceRate->resourceCategory];
            $resourceRates[] = $resourceRate;
        }
        $resourceMap[$plan->id] = $this->openApi->addServicePlan(
            $plan->name,
            $resourceMap[$plan->serviceTemplateId],
            $plan->shortDescription,
            $plan->longDescription,
            $plan->salesCategory,
            ADDITIONAL_FILE_PATH . $plan->imageFile,
            $plan->published,
            $plan->notificationTemplate,
            $plan->billPeriod,
            $plan->billPeriodType,
            $plan->chargingType,
            $plan->autoRenewType,
            $plan->renewOrderInterval,
            $resourceRates,
            $plan->planPeriods,
            $plan->listingPriority
        );
        Log::getInstance()->event(
            self::LOG_SERVICE_PLAN_CREATED,
            $plan->name
        );
    }

    /**
     * @param $resourcesConfig
     * @param $types
     */
    protected function changedUnitMeasurements($resourcesConfig, $types)
    {
        foreach ($resourcesConfig->measurementUnit as $measurementUnit) {
            $this->openApi->changeMeasurementUnit(
                $types[$measurementUnit->resourceID],
                $measurementUnit->measurementUnit
            );
        }
    }

    /**
     * @param $resourcesConfig
     * @return array
     */
    protected function createResourceCategories($resourcesConfig)
    {
        $resourceCategories = array();
        foreach ($resourcesConfig->resourceCategories as $category) {
            $this->openApi->addResourceCategory($category->name);
            $resourceCategories[$category->id] = $category->name;

            Log::getInstance()->event(
                self::LOG_RESOURCE_CATEGORY_CREATED,
                $category->name
            );
        }
        return $resourceCategories;
    }

    /**
     * @param $configuration
     * @throws \Rest\RestException
     */
    protected function configurationProcessing($configuration, &$types)
    {
        $datacenter = APS\TypeLibrary::newResourceByTypeId(\APS\TypeLibrary::getTypeByServiceId('datacenter'));
        $datacenter->url = $configuration->datacenter->url;
        $datacenter->login = $configuration->datacenter->login;
        $datacenter->password = $configuration->datacenter->password;
        $datacenter->name = self::DEFAULT_DATACENTER_NAME;
        /** @var datacenter $response */
        $response = $this->apsc->provisionResource($datacenter);

        $storages = $response->getStorages();

        if (empty($storages)) {
            return;
        }

        $storage = array_pop($storages);
        $newStorage = new BackupProfileStorage();
        $newStorage->datacenterId = $response->id;
        $newStorage->storageId = $storage->id;
        $types['storages'][] = $newStorage;

    }
}
