<?php
/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

class Model_ConnectionChecker_OpenApiHostConnection implements Model_ConnectionChecker_CheckerInterface
{
    const HTTPS = "https";
    const HTTP = "http";
    const OPEN_API_URL_TEMPLATE = '{proto}://{host}';
    const HOST = "host";
    const PORT = "port";

    /** @var string */
    private $_host;
    /** @var int */
    private $_port;
    /** @var Model_ConnectionChecker_CurlConnector */
    private  $_curlConnector;
    /** @var  Service_ConfigEditor */
    private $_configEditor;
    /** @var int */
    private $_errorCode = 0;
    /** @var string */
    private $_errorMessage = '';

    /**
     * @param string $host
     * @param int $port
     * @param Model_ConnectionChecker_CurlConnectorInterface $curlConnector
     * @param Service_ConfigEditor $configEditor
     */
    public function __construct(
        $host,
        $port,
        Model_ConnectionChecker_CurlConnectorInterface $curlConnector = NULL,
        Service_ConfigEditor $configEditor = NULL
    )
    {
        $this->_host = $host;
        $this->_port = $port;
        $this->_curlConnector = is_null($curlConnector) ? new Model_ConnectionChecker_CurlConnector() : $curlConnector;
        $this->_configEditor = is_null($configEditor) ? new Service_ConfigEditor() : $configEditor;
    }

    /**
     * @return bool
     */
    public function check()
    {
        $this->_errorCode = 0;
        $this->_errorMessage = '';

        $openApiProtocol = defined('OPEN_API_PROTOCOL') ? OPEN_API_PROTOCOL : self::HTTP;
        $openApiUrl = $this->_formatOpenApiUri($openApiProtocol, $this->_host);
        try
        {
            $this->_curlConnector->tryConnectToHost($openApiUrl, $this->_port);
            return true;
        }
        catch (Exception $e)
        {
            Log::getInstance()->warning($e->getMessage());
        }

        $newProtocol = ($openApiProtocol === self::HTTPS) ? self::HTTP : self::HTTPS;
        $openApiUrl = $this->_formatOpenApiUri($newProtocol, $this->_host);
        try 
        {
            $this->_curlConnector->tryConnectToHost($openApiUrl, $this->_port);
            $this->_configEditor->updateOpenApiProtocol($newProtocol);
            return true;
        }
        catch (Exception $e)
        {
            $this->_errorCode = $e->getCode();
            $this->_errorMessage = $e->getMessage();
            return false;
        }
    }

    /**
     * @return array
    */
    public function getCheckDescription()
    {
        return (object) array(
            self::MESSAGE => 'Verifying connection to host __host__, port __port__',
            self::KEYS    => (object) array(
                self::HOST => $this->_host,
                self::PORT => $this->_port
            )
        );
    }

    /**
     * @return array
     */
    public function getCheckError()
    {
        return (object) array(
            self::CODE    => $this->_errorCode,
            self::MESSAGE => $this->_errorMessage
        );
    }

    /**
     * @param string $httpProtocol
     * @param string $host
     * @return string
     */
    private function _formatOpenApiUri($httpProtocol, $host)
    {
        return strtr(
            self::OPEN_API_URL_TEMPLATE,
            array(
                '{proto}' => $httpProtocol,
                '{host}' => $host,
            )
        );
    }
}