<?php
/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

class Model_MspGroup_OpenApiRetriever implements Model_MspGroup_OpenApiRetrieverInterface
{
    const EXCEPTION_CANNOT_CREATE_GROUP_MESSAGE = "Cannot create group for reseller \"%s\"";

    /**
     * @var Model_ResellerMapHelperInterface
     */
    private $_resellerMapHelper;
    /** @var Service_MspApiClient_ApiInterface */
    private $_mspApi;
    /** @var Service_OpenApi_Api */
    private $_openApi;
    /** @var array */
    private $_parentStack = array();

    /**
     * @param  Model_ResellerMapHelperInterface $resellerMapHelper
     * @param Service_MspApiClient_ApiInterface $mspApi
     * @param Service_OpenApi_ApiInterface $openApi
     */
    public function __construct(
        Model_ResellerMapHelperInterface $resellerMapHelper,
        Service_MspApiClient_ApiInterface $mspApi,
        Service_OpenApi_ApiInterface $openApi
    )
    {
        $this->_resellerMapHelper = $resellerMapHelper;
        $this->_mspApi = $mspApi;
        $this->_openApi = $openApi;
    }

    /**
     *
     * @param int $poaAccountId
     * @return int
     */
    public function getGroupIdByPoaAccountId($poaAccountId)
    {
        $accountInfo = $this->_openApi->getAccountInfo($poaAccountId);
        $parentMspGroupId = $this->_getParentGroupId($accountInfo);

        while (count($this->_parentStack))
        {
            $accountInfo = array_pop($this->_parentStack);
            $parentMspGroupId = $this->_createGroupForReseller($accountInfo, $parentMspGroupId);
            $this->_resellerMapHelper->addResellerIdToMap($accountInfo->getAccountId(), $parentMspGroupId);
        }

        $this->_resellerMapHelper->updateResellerMap();
        return $parentMspGroupId;
    }

    /**
     * @param Service_OpenApi_Response_AccountInfo $accountInfo
     * @return int | null
     */
    private function _getGroupIdForReseller($accountInfo)
    {
        $accountId = $accountInfo->getAccountId();
        $groupId = $this->_resellerMapHelper->getResellerIdFromMap($accountId);
        if (!is_null($groupId))
        {
            return $groupId;
        }

        $name = Model_MspGroupManager::formatResellerGroupName($accountInfo->getCompanyName(), $accountId);
        $groupInfo = $this->_searchGroupByName($name);
        if (is_null($groupInfo))
        {
            return null;
        }

        $groupId = $groupInfo->getId();
        $this->_resellerMapHelper->addResellerIdToMap($accountId, $groupId);
        return $groupId;
    }

    /**
     * @param Service_OpenApi_Response_AccountInfoInterface $parentAccountInfo
     * @return int
     */
    private function _getParentGroupId($parentAccountInfo)
    {
        $this->_parentStack = array();
        do
        {
            $parentAccountId = $parentAccountInfo->getParentAccountId();
            $parentAccountInfo = $this->_openApi->getAccountInfo($parentAccountId);
            if ($parentAccountInfo->getAccountType() == Service_OpenApi_ApiInterface::ACCOUNT_TYPE_PROVIDER) {
                return $this->_mspApi->getAccountGroupId();
            }
            $mspGroupId = $this->_getGroupIdForReseller($parentAccountInfo);
        }
        while (is_null($mspGroupId) && array_push($this->_parentStack, $parentAccountInfo));
        return $mspGroupId;
    }

    /**
     * @param $companyName
     * @return NULL | Service_MspApiClient_Response_Struct_SearchItem
     */
    private function _searchGroupByName($companyName)
    {
        $groups = $this->_mspApi->search($companyName);
        if (!is_array($groups)) {
            return NULL;
        }
        foreach ($groups as $group) {
            if ($group->getName() === $companyName) {
                return $group;
            }
        }
        return NULL;
    }

    /**
     * @param Service_OpenApi_Response_AccountInfoInterface $accountInfo
     * @param int $parentGroupId
     * @return int
     * @throws Exception
     */
    private function _createGroupForReseller($accountInfo, $parentGroupId)
    {
        $apsc = \APS\Request::getController();
        $apsApi = new Service_APSApi_Api($apsc);//@todo: move to the constructor
        $tenant = $this->getTenantByAccountId($apsApi, $accountInfo);

        if (!is_null($tenant)) {
            $mspGroupmanager = new Model_MspGroupManager($this->_mspApi, $apsApi);
            return $mspGroupmanager->createGroupForReseller($tenant, $parentGroupId);
        }
        return $this->createGroupForResellerByAccountInfo($accountInfo, $parentGroupId);
    }

    private function createGroupForResellerByAccountInfo($accountInfo, $parentGroupId)
    {
        $groupInfo = new Service_MspApiClient_Request_Struct_CreateGroup();

        $groupName = Model_MspGroupManager::formatResellerGroupName(
            $accountInfo->getCompanyName(),
            $accountInfo->getAccountId()
        );
        $groupInfo->setName($groupName);
        $groupInfo->setKind(Service_MspApiClient_Request_Struct_CreateGroupInterface::RESELLER_GROUP_KIND_ID);

        $pricing = $groupInfo->getPricing();
        // for reseller we create only production EUC
        $pricing->setMode(Service_MspApiClient_Request_Struct_GroupPricing::PRODUCTION_MODE);
        $pricing->setEnabled(Service_MspApiClient_Request_Struct_GroupPricing::ENABLED_OFF);

        $groupInfo->setCustomerId(Arr::encode(array(
            Model_MspGroupManager::OSA_ACCOUNT_ID => $accountInfo->getAccountId()
        )));

        $accountInfoContact = $accountInfo->getContact();
        $contact = $groupInfo->getContact();
        $contact->setEmail($accountInfoContact->getEmail());
        $contact->setFirstName($accountInfoContact->getGivenName());
        $contact->setLastName($accountInfoContact->getFamilyName());
        $contact->setPhone($accountInfo->getPhone());

        $accountInfoAddress = $accountInfo->getAddress();
        $contact->setCountry(strtoupper($accountInfoAddress->getCountry()));
        $contact->setZipCode($accountInfoAddress->getZipCode());
        $contact->setState($accountInfoAddress->getState());
        $contact->setCity($accountInfoAddress->getCity());
        $contact->setAddress1($accountInfoAddress->getAddress1());
        $contact->setAddress2($accountInfoAddress->getAddress2());

        $creationResult = $this->_mspApi->createGroup($groupInfo, $parentGroupId);
        if (!$creationResult) {
            throw new Exception(sprintf(self::EXCEPTION_CANNOT_CREATE_GROUP_MESSAGE, $groupName));
        }

        $this->enableCloudToCloudBackup($creationResult->getId(), $creationResult->getVersion());

        return $creationResult->getId();
    }

    /**
     * @param Service_APSApi_ApiInterface $apsApi
     * @param Service_OpenApi_Response_AccountInfoInterface $accountInfo
     * @return mixed
     */
    private function getTenantByAccountId($apsApi, $accountInfo)
    {
        $tenants = $apsApi->getResellersByAccountId($accountInfo->getAccountId());
        $tenant = count($tenants)>0 ? reset($tenants) : null;
        return $tenant;
    }

    /**
     * @param $groupId
     * @param $version
     * @return int
     */
    private function enableCloudToCloudBackup($groupId, $version)
    {
        try {
            $cloudToCloudGroupInfo = new Service_MspApiClient_Request_Struct_UpdateGroup();
            $privileges = $cloudToCloudGroupInfo->getPrivileges();
            $privileges->setCloudToCloudBackup(1);
            return $this->_mspApi->updateGroup($groupId, $cloudToCloudGroupInfo, $version);
        } catch (Exception $e) {
            // Skip any errors
            Log::getInstance()->warning($e->getMessage());
        }
        return $version;
    }
}
