<?php
/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

use Acronis\App\Base\BackupApplication;

class Model_MspGroupManager implements Model_MspGroupManagerInterface
{
    const LOG_EVENT_METHOD_WAS_CALLED = 'Method "%s" was called.';
    const LOG_EVENT_MIGRATE_GROUP = 'Migrate group with ID :group_id was updated. Details: group_old_name=":group_old_name", group_new_name=":group_new_name", group_new_version=:group_new_version, old_parent_id=:old_parent_id, new_parent_id=:new_parent_id';
    const LOG_EVENT_NORMALIZE_GROUP_NAME_AFTER_MIGRATE = 'Name of group with ID :group_id was normalized after migration. Details: group_old_name=":group_old_name", group_new_name=":group_new_name", group_new_version=:group_new_version';

    const RESELLER_GROUP_NAME_TEMPLATE = '{companyName} (Reseller {accountId})';
    const CUSTOMER_GROUP_NAME_TEMPLATE = '{companyName} /{suffix}';
    const CUSTOMER_GROUP_NAME_PATTERN = '/^%s(?: \/(\d+))?$/i';
    const PATTERN_DELIMITER = '/';

    const EXCEPTION_CANNOT_CREATE_GROUP_MESSAGE = "Cannot create group for account \"%s\"";
    const WARNING_INTERNAL_TAG_WAS_NOT_UPDATED = 'Internal tag was not updated';

    const HASH = '#';

    const UNLIMITED_QUOTA_VALUE = -1;

    const OSA_ACCOUNT_ID = 'osa_account_id';
    const OSA_SUBSCRIPTION_ID = 'osa_subscription_id';
    const CI_APS_ABC_VERSION = 'ci-aps-abc-version';
    const TIER_LEVEL = 'aps-abc-tier';
    const START_PROD_DATE = 'aps-abc-start-prod-date';

    /** @var Service_MspApiClient_ApiInterface */
    private $_mspApi;
    /** @var Service_APSApi_ApiInterface */
    private $_apsApi;

    /**
     * Model_MspGroupManager constructor.
     *
     * @param Service_MspApiClient_ApiInterface $mspApi
     * @param Service_APSApi_ApiInterface $apsApi
     */
    public function __construct(
        Service_MspApiClient_ApiInterface $mspApi,
        Service_APSApi_ApiInterface $apsApi
    ) {
        $this->_mspApi = $mspApi;
        $this->_apsApi = $apsApi;
    }

    /**
     * @param tenant $tenant
     * @param int $parentGroupId
     * @param int $datacenterId
     * @return int
     * @throws Exception
     */
    public function createGroupForReseller($tenant, $parentGroupId, $datacenterId = null)
    {
        Log::getInstance()->event(sprintf(self::LOG_EVENT_METHOD_WAS_CALLED, __METHOD__));

        $groupName = self::formatResellerGroupName(
            $tenant->account->companyName,
            $tenant->account->id
        );

        $groupId = $this->findGroupByName($parentGroupId, $groupName);
        if (!is_null($groupId)) {
            $this->_apsApi->checkGroupCanBeAssignedOrEx($groupId, $datacenterId);
            $this->updateGroupForReseller($groupId, $tenant);

            return $groupId;
        }

        $groupInfo = new Service_MspApiClient_Request_Struct_CreateGroup();
        $groupInfo->setName($groupName);
        $groupInfo->setKind(Service_MspApiClient_Request_Struct_CreateGroupInterface::RESELLER_GROUP_KIND_ID);

        $pricing = $groupInfo->getPricing();
        $pricing->setMode(Service_MspApiClient_Request_Struct_GroupPricing::PRODUCTION_MODE);
        $pricing->setEnabled(Service_MspApiClient_Request_Struct_GroupPricing::ENABLED_OFF);

        $groupInfo->setCustomerId(Arr::encode([
            self::OSA_ACCOUNT_ID => $tenant->account->id,
            self::OSA_SUBSCRIPTION_ID => $tenant->subscription->subscriptionId,
        ]));

        $this->initGroupInfoContact($tenant, $groupInfo);

        $createResult = $this->createGroup($parentGroupId, $groupInfo);
        $version = $createResult->getVersion();

        if (!empty($tenant->startProductionTime) || !$tenant->minimalCommitmentTier->isEmpty()) {
            $updateGroupInfo = new Service_MspApiClient_Request_Struct_UpdateGroup();
            $minimalCommitmentTier = $tenant->minimalCommitmentTier;
            $internalTag = [];
            if (!$tenant->minimalCommitmentTier->isEmpty()) {
                $internalTag[self::TIER_LEVEL] = $minimalCommitmentTier->level;
            }
            if (!empty($tenant->startProductionTime)) {
                $internalTag[self::START_PROD_DATE] = $tenant->startProductionTime;
            }
            $updateGroupInfo->setInternalTag(Arr::encode($internalTag));
            $this->updateGroup($createResult->getId(), $updateGroupInfo, $version);
        }

        $applicationManager = $this->createBackupApplicationManager($tenant, $createResult->getUuid());

        try {
            // it's a w/a to disable all offering items after creating via APIv1
            $applicationManager->deactivateAllTenantOfferingItems();
        } catch (Exception $e) {
            Log::getInstance()->warning(
                'Unable to sanitize offering items for tenant "%s". Error: "%s".',
                $createResult->getUuid(), $e->getMessage()
            );
        }

        $applicationManager->updateOfferingItems();

        return $createResult->getId();
    }

    /**
     * @param int $groupId
     * @param tenant $tenant
     * @return int
     * @throws Exception
     */
    public function updateGroupForReseller($groupId, $tenant)
    {
        Log::getInstance()->event(sprintf(self::LOG_EVENT_METHOD_WAS_CALLED, __METHOD__));

        $groupInfo = $this->_mspApi->getGroupById($groupId);

        $updateGroupInfo = new Service_MspApiClient_Request_Struct_UpdateGroup();

        $groupName = self::formatResellerGroupName(
            $tenant->account->companyName,
            $tenant->account->id
        );
        $updateGroupInfo->setName($groupName);

        $pricing = $updateGroupInfo->getPricing();
        $pricing->setMode(Service_MspApiClient_Request_Struct_GroupPricing::PRODUCTION_MODE);
        $pricing->setEnabled(Service_MspApiClient_Request_Struct_GroupPricing::ENABLED_OFF);

        $customerIdParts = Arr::decode($groupInfo->getCustomerId());
        $customerIdParts[self::OSA_ACCOUNT_ID] = $tenant->account->id;
        $customerIdParts[self::OSA_SUBSCRIPTION_ID] = $tenant->subscription->subscriptionId;
        $updateGroupInfo->setCustomerId(Arr::encode($customerIdParts));

        $minimalCommitmentTier = $tenant->minimalCommitmentTier;
        $internalTag = Arr::decode($groupInfo->getInternalTag());
        if (!$minimalCommitmentTier->isEmpty()) {
            $internalTag[self::TIER_LEVEL] = $minimalCommitmentTier->level;
        }
        if (!empty($tenant->startProductionTime)) {
            $internalTag[self::START_PROD_DATE] = $tenant->startProductionTime;
        }
        $updateGroupInfo->setInternalTag(Arr::encode($internalTag));

        $this->initGroupInfoContact($tenant, $updateGroupInfo);

        $version = $this->updateGroup($groupId, $updateGroupInfo, $groupInfo->getVersion());

        $applicationManager = $this->createBackupApplicationManager($tenant, $groupInfo->getUuid());
        $applicationManager->updateOfferingItems();

        return $version;
    }

    /**
     * @param tenant $tenant
     * @param int $parentGroupId
     * @param int $groupKind
     * @param string $language
     * @return int
     * @throws Exception
     * @throws Service_BaseClient_HTTPException
     */
    public function createGroupForCustomer(
        $tenant,
        $parentGroupId,
        $groupKind = Service_MspApiClient_Request_Struct_CreateGroup::EUC_GROUP_KIND_ID,
        $language = null
    ) {
        Log::getInstance()->event(sprintf(self::LOG_EVENT_METHOD_WAS_CALLED, __METHOD__));

        $companyName = $tenant->account->companyName;
        $groupSuffix = $this->getLastGroupNameSuffix($companyName, $parentGroupId);
        $groupName = is_null($groupSuffix) ? $companyName : self::formatCustomerGroupName($companyName, ++$groupSuffix);

        $groupInfo = new Service_MspApiClient_Request_Struct_CreateGroup();
        $groupInfo->setName($groupName);
        $groupInfo->setKind($groupKind);

        $pricing = $groupInfo->getPricing();
        $mode = $groupKind !== Service_MspApiClient_Request_Struct_CreateGroupInterface::EUC_GROUP_KIND_ID ?
            Service_MspApiClient_Request_Struct_GroupPricing::PRODUCTION_MODE :
            $this->getGroupMode($tenant);
        $pricing->setMode($mode);
        $pricing->setEnabled(Service_MspApiClient_Request_Struct_GroupPricing::ENABLED_OFF);

        $groupInfo->setCustomerId(Arr::encode([
            self::OSA_ACCOUNT_ID => $tenant->account->id,
            self::OSA_SUBSCRIPTION_ID => $tenant->subscription->subscriptionId,
        ]));

        $this->initGroupInfoContact($tenant, $groupInfo);

        if (!is_null($language)) {
            $groupInfo->setLanguage($language);
        }
        $createResult = $this->createGroup($parentGroupId, $groupInfo);

        // it's a w/a to disable all offering items after creating via APIv1 enabled by default
        try {
            $applicationManager = $this->createBackupApplicationManager($tenant, $createResult->getUuid());
            $applicationManager->deactivateAllTenantOfferingItems();
        } catch (Exception $e) {
            Log::getInstance()->warning(
                'Unable to sanitize offering items for tenant "%s". Error: "%s".',
                $createResult->getUuid(), $e->getMessage()
            );
        }

        return $createResult->getId();
    }

    /**
     * @param tenant $tenant
     * @param $tenantStrategy
     * @throws \Rest\RestException
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    public function updatePrivilegesForCustomers($tenant, $tenantStrategy)
    {
        $applicationManager = $this->createBackupApplicationManager($tenant, null, $tenantStrategy);
        $applicationManager->updateOfferingItems();
    }

    /**
     * @param int $groupId
     * @param tenant $tenant
     * @return int
     * @throws Exception
     * @throws Service_BaseClient_HTTPException
     */
    public function updateGroupForCustomer($groupId, $tenant)
    {
        Log::getInstance()->event(sprintf(self::LOG_EVENT_METHOD_WAS_CALLED, __METHOD__));

        $groupInfo = $this->_mspApi->getGroupById($groupId);

        $companyName = $tenant->account->companyName;

        $groupName = $this->fetchNameForGroup($companyName, $groupInfo);

        $updateGroupInfo = new Service_MspApiClient_Request_Struct_UpdateGroup();
        $updateGroupInfo->setName($groupName);

        $pricing = $updateGroupInfo->getPricing();
        $mode = $groupInfo->getKind() !== Service_MspApiClient_Request_Struct_CreateGroupInterface::EUC_GROUP_KIND_ID ?
            Service_MspApiClient_Request_Struct_GroupPricing::PRODUCTION_MODE :
            $this->getGroupMode($tenant);

        $currentMode = $groupInfo->getPricing()->getMode();
        if ($currentMode !== $mode) {
            if ($currentMode === Service_MspApiClient_Request_Struct_GroupPricing::PRODUCTION_MODE) {
                // never change pricing mode from production mode
                Log::getInstance()->warning(
                    'The current pricing mode of the tenant is "%s" while the pricing mode of the subscription is "%s".',
                    $currentMode, $mode
                );
            } else {
                $pricing->setMode($mode);
            }
        }

        $pricing->setEnabled(Service_MspApiClient_Request_Struct_GroupPricing::ENABLED_OFF);

        $customerIdParts = Arr::decode($groupInfo->getCustomerId());
        $customerIdParts[self::OSA_ACCOUNT_ID] = $tenant->account->id;
        $customerIdParts[self::OSA_SUBSCRIPTION_ID] = $tenant->subscription->subscriptionId;
        $updateGroupInfo->setCustomerId(Arr::encode($customerIdParts));

        $this->initGroupInfoContact($tenant, $updateGroupInfo);

        $version = $this->updateGroup($groupId, $updateGroupInfo, $groupInfo->getVersion());

        return $version;
    }

    /**
     * @param int $groupId
     * @param tenant $tenant
     * @param int $newParentGroupId
     * @return int
     * @throws Exception
     * @throws Service_BaseClient_HTTPException
     */
    public function migrateGroupForCustomer($groupId, $tenant, $newParentGroupId)
    {
        Log::getInstance()->event(sprintf(self::LOG_EVENT_METHOD_WAS_CALLED, __METHOD__));

        $groupInfo = $this->_mspApi->getGroupById($groupId);

        if ($groupInfo->getParentId() == $newParentGroupId) {
            return $groupInfo->getVersion();
        }

        $companyName = $tenant->account->companyName;

        // Actually, we split moving group in 2 step (2 http requests)

        // 1) Set name that will be unique for old and new parents and move group (first http request)

        $updateGroupInfo = new Service_MspApiClient_Request_Struct_UpdateGroup();
        $tmpGroupName = $companyName . " ({$groupId})";
        $updateGroupInfo->setName($tmpGroupName);
        $updateGroupInfo->setParentId($newParentGroupId);
        $groupNewVersion = $this->updateGroup($groupId, $updateGroupInfo, $groupInfo->getVersion());

        Log::getInstance()->event(
            self::LOG_EVENT_MIGRATE_GROUP,
            [
                ':group_id' => $groupInfo->getId(),
                ':group_old_name' => $groupInfo->getName(),
                ':group_new_name' => $tmpGroupName,
                ':old_parent_id' => $groupInfo->getParentId(),
                ':new_parent_id' => $newParentGroupId,
                ':group_new_version' => $groupNewVersion,
            ]
        );

        // 2) Rename moved group deleting unique part (second http request) and enable group if it status is inactive

        $groupSuffix = $this->getLastGroupNameSuffix($companyName, $newParentGroupId);
        $finalGroupName = is_null($groupSuffix) ? $companyName : self::formatCustomerGroupName($companyName,
            ++$groupSuffix);

        $updateGroupInfo = new Service_MspApiClient_Request_Struct_UpdateGroup();
        $updateGroupInfo->setName($finalGroupName);
        if ($groupInfo->getStatus() == Model_TenantManager::STATUS_INACTIVE) {
            $updateGroupInfo->setStatus(Model_TenantManager::STATUS_ACTIVE);
        }
        $groupNewVersion = $this->updateGroup($groupId, $updateGroupInfo, $groupNewVersion);

        Log::getInstance()->event(
            self::LOG_EVENT_NORMALIZE_GROUP_NAME_AFTER_MIGRATE,
            [
                ':group_id' => $groupInfo->getId(),
                ':group_old_name' => $tmpGroupName,
                ':group_new_name' => $finalGroupName,
            ]
        );

        return $groupNewVersion;
    }

    /**
     * @param string $companyName
     * @param Service_MspApiClient_Response_Struct_GroupInfo $groupInfo
     * @return string
     */
    public function fetchNameForGroup($companyName, Service_MspApiClient_Response_Struct_GroupInfo $groupInfo)
    {
        $groupSuffix = $this->getGroupNameSuffix($companyName, $groupInfo->getName());
        if (is_null($groupSuffix)) {
            $groupName = $companyName;

            return $groupName;
        } elseif ($groupSuffix === false) {
            $groupSuffix = $this->getLastGroupNameSuffix($companyName, $groupInfo->getParentId());
            $groupName = is_null($groupSuffix) ? $companyName : self::formatCustomerGroupName($companyName,
                ++$groupSuffix);

            return $groupName;
        } else {
            $groupName = self::formatCustomerGroupName($companyName, $groupSuffix);

            return $groupName;
        }
    }

    /**
     * @param string $companyName
     * @param int $id
     * @return string
     */
    static public function formatResellerGroupName($companyName, $id)
    {
        return strtr(self::RESELLER_GROUP_NAME_TEMPLATE, [
            '{companyName}' => $companyName,
            '{accountId}' => $id,
        ]);
    }

    /**
     * @param tenant $tenant
     * @param string|null $tenantId
     * @param $tenantStrategy
     * @return BackupApplication
     * @throws \Rest\RestException
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    protected function createBackupApplicationManager(
        $tenant,
        $tenantId = null,
        $tenantStrategy = null
    ) {
        return new BackupApplication($tenant, $this->_mspApi, $tenantId, $tenantStrategy);
    }

    /**
     * @param int $parentGroupId
     * @param string $companyName
     * @return int
     * @throws Exception
     */
    private function findGroupByName($parentGroupId, $companyName)
    {
        /**
         * @var Service_MspApiClient_Response_Struct_GroupItem[] $groups
         */
        $groups = $this->_mspApi->getGroups($parentGroupId);
        foreach ($groups as $group) {
            if ($group->getName() === $companyName) {
                return $group->getId();
            }
        }

        return null;
    }

    /**
     * @param tenant $tenant
     * @param Service_MspApiClient_Request_Struct_GroupInterface $groupInfo
     */
    private function initGroupInfoContact($tenant, Service_MspApiClient_Request_Struct_GroupInterface $groupInfo)
    {
        $admin = $this->_apsApi->getAdmin($tenant);
        if (is_null($admin)) {
            return;
        }

        $contact = $groupInfo->getContact();
        $contact->setEmail($admin->email);
        $contact->setFirstName($admin->givenName);
        $contact->setLastName($admin->familyName);
        $contact->setPhone(str_replace(self::HASH, '', $admin->telWork));

        $addressPostal = $admin->addressPostal;
        if (is_null($addressPostal)) {
            return;
        }
        $contact->setCountry(strtoupper($addressPostal->countryName));
        $contact->setZipCode($addressPostal->postalCode);
        $contact->setState($addressPostal->region);
        $contact->setCity($addressPostal->locality);
        $contact->setAddress1($addressPostal->streetAddress);
        $contact->setAddress2($addressPostal->extendedAddress);
    }

    /**
     * @param int $parentGroupId
     * @param Service_MspApiClient_Request_Struct_CreateGroupInterface $groupInfo
     * @return Service_MspApiClient_Response_Struct_CreateGroup
     * @throws Exception
     */
    private function createGroup($parentGroupId, Service_MspApiClient_Request_Struct_CreateGroupInterface $groupInfo)
    {
        $creationResult = $this->_mspApi->createGroup($groupInfo, $parentGroupId);
        if (!$creationResult) {
            throw new Exception(sprintf(self::EXCEPTION_CANNOT_CREATE_GROUP_MESSAGE, $groupInfo->getName()));
        }

        return $creationResult;
    }

    /**
     * @param $groupId
     * @param $updateGroupInfo
     * @param $version
     * @return int
     */
    private function updateGroup($groupId, $updateGroupInfo, $version)
    {
        $tenantManager = new Model_TenantManager($this->_mspApi);

        return $tenantManager->updateGroup($groupId, $updateGroupInfo, $version);
    }

    /**
     * @param string $companyName
     * @param int $parentGroupId
     * @return int|null
     */
    private function getLastGroupNameSuffix($companyName, $parentGroupId)
    {
        $children = $this->_mspApi->getGroups($parentGroupId);
        $groupSuffix = null;
        foreach ($children as $child) {
            $childSuffix = $this->getGroupNameSuffix($companyName, $child->getName());
            if ($childSuffix === false) {
                continue;
            }

            if (is_null($childSuffix)) {
                $childSuffix = 0;
            }

            $groupSuffix = is_null($groupSuffix) ? $childSuffix : max($groupSuffix, $childSuffix);
        }

        return $groupSuffix;
    }

    /**
     * @param string $companyName
     * @param int $suffix
     * @return string
     */
    static public function formatCustomerGroupName($companyName, $suffix)
    {
        return strtr(self::CUSTOMER_GROUP_NAME_TEMPLATE, [
            '{companyName}' => $companyName,
            '{suffix}' => $suffix,
        ]);
    }

    /**
     * @param tenant $tenant
     * @return int
     */
    private function getGroupMode($tenant)
    {
        return $tenant->subscription->trial
            ? Service_MspApiClient_Request_Struct_GroupPricing::TRIAL_MODE
            : Service_MspApiClient_Request_Struct_GroupPricing::PRODUCTION_MODE;
    }

    /**
     * @param string $companyName
     * @param string $groupName
     * @return bool|int|null NULL - no suffix, int - group suffix, false - names don't match
     */
    private function getGroupNameSuffix($companyName, $groupName)
    {
        $pattern = sprintf(self::CUSTOMER_GROUP_NAME_PATTERN, preg_quote($companyName, self::PATTERN_DELIMITER));
        if (preg_match($pattern, $groupName, $matches)) {
            return isset($matches[1]) ? intval($matches[1]) : null;
        }

        return false;
    }

    public function updatePackageVersionInfo($groupId)
    {
        $mspApi = $this->_mspApi;
        $group = $mspApi->getGroupById($groupId);
        $buildInfo = new BuildInfo();
        $apsPackageVersion = $buildInfo->getPackageVersion();
        try {
            $updateGroupInfo = new Service_MspApiClient_Request_Struct_UpdateGroup();

            $internalTag = Arr::decode($group->getInternalTag());
            if (isset($internalTag[self::CI_APS_ABC_VERSION])
                && $internalTag[self::CI_APS_ABC_VERSION] == $apsPackageVersion
            ) {
                return;
            }
            $internalTag[self::CI_APS_ABC_VERSION] = $apsPackageVersion;
            $updateGroupInfo->setInternalTag(Arr::encode($internalTag));

            $mspApi->updateGroup($group->getId(), $updateGroupInfo, $group->getVersion());
        } catch (Exception $e) {
            if ($e->getCode() != 403) {
                throw $e;
            }
            Log::getInstance()->warning(self::WARNING_INTERNAL_TAG_WAS_NOT_UPDATED);
        }
    }
}
