<?php
/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

/**
 * See APS documentation about sending notification
 * @link https://doc.apsstandard.org/2.2/spec/ui/integration-ccp/notifications/
 */
class Model_ScreenNotification_Transport implements Model_ScreenNotification_TransportInterface
{
    const NOTIFICATOR_TYPE      = 'http://www.parallels.com/pa/pa-core-services/notification-manager';
    const MASK_RQL_IMPLEMENTING = 'implementing(%s)';
    const MASK_PATH_RESOURCES   = '/aps/2/resources/%s%s';
    const MASK_PATH_NOTIFICATIONS = "/notifications";
    const MASK_PATH_NOTIFICATIONS_WITH_ID = "/notifications/%s";
    const EXCEPTION_MSG_RECEIVE_NOTIFICATOR = "Could not receive notification manager";
    const EXCEPTION_MSG_EXECUTE_OPERATION = "Could not execute operation %s %s";

    /**
     * @var \APS\ControllerProxy
     */
    private $_apsc;

    /**
     * @var \APS\ResourceBase
     */
    private $_notificator;

    /**
     * @param \APS\ControllerProxy $apsc
     * @throws Exception
     */
    public function __construct(\APS\ControllerProxy $apsc = null)
    {
        /**
         * See APS documentation about sending notification  "Sequence of Operations"
         */
        if (is_null($apsc))
        {
            $apsc = \APS\Request::getController();
        }

        $this->_apsc = clone $apsc;
        $this->_apsc->resetSession();
    }

    private function _getNotificator()
    {
        if (!isset($this->_notificator))
        {
            $resources = $this->_apsc->getResources(sprintf(self::MASK_RQL_IMPLEMENTING, self::NOTIFICATOR_TYPE));
            if (isset($resources[0]))
            {
                $this->_notificator = $resources[0];
            }
            else
            {
                throw new Exception(self::EXCEPTION_MSG_RECEIVE_NOTIFICATOR);
            }
        }
        return $this->_notificator;
    }

    /**
     * @param string $verb
     * @param string $path
     * @param mixed $input
     *
     * @throws Exception
     *
     * @return mixed
     */
    private function _request($verb, $path, $input = null)
    {
        $notificator =  $this->_getNotificator();
        $notificatorId = $notificator->aps->id;

        $path = sprintf(self::MASK_PATH_RESOURCES, $notificatorId, $path);

        if (isset($input))
        {
            $input = json_encode($input);
        }

        try
        {
            $response = $this->_apsc->getIo()->sendRequest($verb, $path, $input);
        }
        catch(Exception $e)
        {
            throw new Exception(sprintf(self::EXCEPTION_MSG_EXECUTE_OPERATION, $verb, $path), 0, $e);
        }

        if (!empty($response))
        {
            $response = json_decode($response, true);
        }

        return $response;
    }

    /**
     * Send a notification
     *
     * @param Model_ScreenNotification_Struct_MessageInterface $message
     *
     * @return Model_ScreenNotification_Struct_MessageInterface
     * @throws Exception
     */
    public function send(Model_ScreenNotification_Struct_MessageInterface $message)
    {
        $request = Model_ScreenNotification_Struct_MessageConverter::toArray($message);
        $response = $this->_request(\APS\Proto::POST, self::MASK_PATH_NOTIFICATIONS, $request);
        return Model_ScreenNotification_Struct_MessageConverter::fromArray($response);
    }

    /**
     * Update a notification
     *
     * @param string $notificationId
     * @param Model_ScreenNotification_Struct_MessageInterface $message
     *
     * @return Model_ScreenNotification_Struct_MessageInterface
     * @throws Exception
     */
    public function edit($notificationId, Model_ScreenNotification_Struct_MessageInterface $message)
    {
        $request = Model_ScreenNotification_Struct_MessageConverter::toArray($message);
        $response = $this->_request(\APS\Proto::PUT, sprintf(self::MASK_PATH_NOTIFICATIONS_WITH_ID, $notificationId), $request);
        return Model_ScreenNotification_Struct_MessageConverter::fromArray($response);
    }

    /**
     * Delete a notification
     *
     * @param string $notificationId
     *
     * @return string
     * @throws Exception
     */
    public function delete($notificationId)
    {
        return $this->_request(\APS\Proto::DELETE, sprintf(self::MASK_PATH_NOTIFICATIONS_WITH_ID, $notificationId));
    }
}