<?php

/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */
class Model_UsageReportManager
{
    const EXCEPTION_REPORT_IS_NOT_READY = 'Report is not ready.';
    const EXCEPTION_REPORT_WAS_NOT_REQUESTED = 'Report was not requested before.';
    const ACRONIS_STORAGE_NAME_POSTFIX = 'acronis.com';
    const REPORT_FILENAME = '%s_%s-%s';

    const WHEN_OFFSET_END_DATE = '+2 hours';

    const LOG_MESSAGE_REQUEST_REPORT = 'Request to generate report for group %s with report attributes: period %s-%s, when %s';
    const LOG_MESSAGE_GET_REPORT = 'Get report for group %s with report attributes: period %s-%s, level %s';

    const DEFAULT_TIMEZONE = 'UTC';

    /**
     * @var Service_MspApiClient_Api
     */
    private $mspApi;
    private $datacenter;

    public function __construct(Service_MspApiClient_ApiInterface $mspApi, $datacenter)
    {
        $this->mspApi = $mspApi;
        $this->datacenter = $datacenter;
    }

    /**
     * @param int $groupId
     * @return mixed|null
     */
    public function fetchCurrentReportForGroup($groupId)
    {
        $timezone = new DateTimeZone(self::DEFAULT_TIMEZONE);

        $startDate = new DateTime('now', $timezone);
        $startDate->setTime(0, 0, 0);
        $endDate = new DateTime('now', $timezone);
        $endDate->setTime(23, 59, 59);

        $startDate->modify('-2 day');
        $endDate->modify('-2 day');
        $this->removeReport($startDate, $endDate);

        $startDate->modify('+1 day');
        $endDate->modify('+1 day');
        $this->generateReportForRootGroup($startDate, $endDate);

        return $this->getUsageReportForGroupFromFile($groupId, $startDate, $endDate);
    }

    /**
     * @param DateTime $startDate
     * @param DateTime $endDate
     */
    private function removeReport($startDate, $endDate)
    {
        $filePath = $this->getReportDatFilePath($startDate, $endDate);
        if (file_exists($filePath)) {
            unlink($filePath);
        }
    }

    /**
     * @param DateTime $startDate
     * @param DateTime $endDate
     * @return string
     */
    private function getReportDatFilePath($startDate, $endDate)
    {
        $reportName = Str::format(self::REPORT_FILENAME . '.dat', $this->datacenter->id, $startDate->format('Ymd'), $endDate->format('Ymd'));
        return DIR_TMP . $reportName;
    }

    /**
     * @param DateTime $startDate
     * @param DateTime $endDate
     * @return string
     */
    private function getReportJsonFilePath($startDate, $endDate)
    {
        $reportName = Str::format(self::REPORT_FILENAME . '.json', $this->datacenter->id, $startDate->format('Ymd'), $endDate->format('Ymd'));
        return DIR_TMP . $reportName;
    }

    /**
     * @param object $reportItem
     * @return bool
     */
    private function isAcronisStorage($reportItem) {
        return is_null($reportItem->getStorageId()) ||
            strstr($reportItem->getStorageName(), self::ACRONIS_STORAGE_NAME_POSTFIX) !== false;
    }

    /**
     * @param DateTime $beginDate
     * @param DateTime $endDate
     */
    private function generateReportForRootGroup($beginDate, $endDate)
    {
        $filePath = $this->getReportDatFilePath($beginDate, $endDate);
        if (!file_exists($filePath)) {
            $rootGroupId = $this->mspApi->getGroupById('self')->getId();
            $report = $this->getUsageReport($beginDate, $endDate, $rootGroupId);
            if (is_null($report)) {
                return;
            }
            $aggregateReport = $this->aggregateReport($report);
            file_put_contents($filePath, serialize($aggregateReport));
        }
    }

    /**
     * Getting report of usage for group
     * @param DateTime $beginDate
     * @param DateTime $endDate
     * @param int $groupId
     * @return mixed
     * @throws Exception
     */
    private function getUsageReport($beginDate, $endDate, $groupId)
    {
        $result = null;
        $reportAttributes = $this->createUsageReportStruct($beginDate, $endDate);
        try {
            Log::getInstance()->event(Str::format(
                self::LOG_MESSAGE_GET_REPORT,
                $groupId,
                $beginDate->format('Y-m-d H:i:s'),
                $endDate->format('Y-m-d H:i:s'),
                $reportAttributes->getLevel()
            ));
            $result = $this->mspApi->getUsageReport($reportAttributes, $groupId);
        } catch (Exception $e) {
            switch ($e->getCode()) {
                case 202:
                    Log::getInstance()->event(self::EXCEPTION_REPORT_IS_NOT_READY);
                    break;
                case 404:
                    Log::getInstance()->event(self::EXCEPTION_REPORT_WAS_NOT_REQUESTED);
                    $this->requestUsageReport($groupId, $beginDate, $endDate);
                    break;
                default:
                    throw $e;
            }
        }
        if (empty($result)) {
            return null;
        }
        $filePath = $this->getReportJsonFilePath($beginDate, $endDate);
        file_put_contents($filePath, json_encode($result->toArray()));
        return $this->convertReport($result);
    }

    /**
     * @param DateTime $startDate
     * @param DateTime $endDate
     * @param int $level
     * @return Service_MspApiClient_Request_Struct_UsageReport
     */
    private function createUsageReportStruct(
        DateTime $startDate,
        DateTime $endDate,
        $level = Service_MspApiClient_Request_Struct_UsageReport::LEVEL_UP_TO_EUC

    ) {
        $usageReport = new Service_MspApiClient_Request_Struct_UsageReport();
        $usageReport->setBeginDate($startDate->getTimestamp());
        $usageReport->setEndDate($endDate->getTimestamp());
        $usageReport->setLevel($level);
        return $usageReport;
    }

    /**
     * @param int $groupId
     * @param DateTime $startDate
     * @param DateTime $endDate
     * @return mixed
     */
    private function requestUsageReport($groupId, $startDate, $endDate)
    {
        $when = clone $endDate;
        $when->modify(self::WHEN_OFFSET_END_DATE);
        Log::getInstance()->event(Str::format(
            self::LOG_MESSAGE_REQUEST_REPORT,
            $groupId,
            $startDate->format('Y-m-d H:i:s'),
            $endDate->format('Y-m-d H:i:s'),
            $when->format('Y-m-d H:i:s')
        ));
        $requestUsageReport = $this->createRequestUsageReportStruct($startDate, $endDate, $when);
        return $this->mspApi->requestUsageReport($requestUsageReport, $groupId);
    }

    /**
     * @param DateTime $startDate
     * @param DateTime $endDate
     * @param DateTime|null $when
     * @param int $kind
     * @param string $resultAction
     * @param int $level
     * @return Service_MspApiClient_Request_Struct_RequestUsageReport
     */
    private function createRequestUsageReportStruct(
        DateTime $startDate,
        DateTime $endDate,
        $when = null,
        $kind = Service_MspApiClient_Request_Struct_RequestUsageReport::KIND_DAILY,
        $resultAction = Service_MspApiClient_Request_Struct_RequestUsageReport::RESULT_ACTION_SAVE,
        $level = Service_MspApiClient_Request_Struct_RequestUsageReport::LEVEL_UP_TO_EUC
    ) {
        $when = is_null($when) ? Service_MspApiClient_Request_Struct_RequestUsageReport::WHEN_ASAP : $when->getTimestamp();

        $requestUsageReport = new Service_MspApiClient_Request_Struct_RequestUsageReport();
        $requestUsageReport->setBeginDate($startDate->getTimestamp());
        $requestUsageReport->setEndDate($endDate->getTimestamp());
        $requestUsageReport->setWhen($when);
        $requestUsageReport->setKind($kind);
        $requestUsageReport->setResultAction($resultAction);
        $requestUsageReport->setLevel($level);
        return $requestUsageReport;
    }

    /**
     * @param Service_MspApiClient_Response_Struct_UsageReport $input
     * @return array
     */
    private function convertReport($input)
    {
        $result = array();

        /** @var Service_MspApiClient_Response_Struct_UsageReportBodyItem $row */
        foreach ($input->getTenants() as $row) {
            $result[$row->getGroupId()][] = $row;
        }
        return $result;
    }

    /**
     * @param array $report
     * @return mixed
     */
    private function aggregateReport($report)
    {
        foreach ($report as $groupId => $groupReport) {
            $groupReportStruct = new StdClass();
            $groupReportStruct->acronisStorage = 0;
            $groupReportStruct->partnerStorage = 0;
            $groupReportStruct->vms = 0;
            $groupReportStruct->servers = 0;
            $groupReportStruct->workstations = 0;
            $groupReportStruct->mailboxes = 0;
            $groupReportStruct->mobiles = 0;
            $groupReportStruct->websites = 0;
            $groupReportStruct->totalStorage = 0;
            $groupReportStruct->acronisWorkstationStorage = 0;
            $groupReportStruct->partnerWorkstationStorage = 0;

            /** @var Service_MspApiClient_Response_Struct_UsageReportBodyItem $reportItem */
            foreach ($groupReport as $reportItem) {
                $productionStorage = $reportItem->getProductionStorage();
                if ($this->isAcronisStorage($reportItem)) {
                    $groupReportStruct->acronisStorage += $productionStorage;
                    $groupReportStruct->acronisStorage -= $reportItem->getProductionStorageOrphaned();
                    $groupReportStruct->acronisWorkstationStorage += $reportItem->getProductionWorkstationsStorage();
                } else {
                    $groupReportStruct->partnerStorage += $productionStorage;
                    $groupReportStruct->partnerStorage -= $reportItem->getProductionStorageOrphaned();
                    $groupReportStruct->partnerWorkstationStorage += $reportItem->getProductionWorkstationsStorage();
                }
                $groupReportStruct->totalStorage += $productionStorage;
                $groupReportStruct->vms += $reportItem->getProductionVms();
                $groupReportStruct->servers += $reportItem->getProductionServers();
                $groupReportStruct->workstations += $reportItem->getProductionWorkstations();
                $groupReportStruct->mailboxes += $reportItem->getProductionMailboxes();
                $groupReportStruct->mobiles += $reportItem->getProductionMobiles();
                $groupReportStruct->websites += $reportItem->getProductionWebsites();
            }
            $report[$groupId] = $groupReportStruct;
        }
        return $report;
    }

    /**
     * @param int $groupId
     * @param DateTime $startDate
     * @param DateTime $endDate
     * @return mixed|null
     */
    private function getUsageReportForGroupFromFile($groupId, $startDate, $endDate)
    {
        $report = $this->getReportFromFile($startDate, $endDate);
        return isset($report[$groupId]) ? $report[$groupId] : null;
    }

    /**
     * @param DateTime $startDate
     * @param DateTime $endDate
     * @return mixed|null
     */
    private function getReportFromFile($startDate, $endDate)
    {
        $report = null;
        $reportPath = $this->getReportDatFilePath($startDate, $endDate);
        if (file_exists($reportPath)) {
            $reportSource = file_get_contents($reportPath);
            $report = unserialize($reportSource);
            return $report;
        }

        return $report;
    }
}
