<?php
/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

class Service_File_Local implements Service_FileInterface
{
    const APP_FILE_GROUP = 'apache';
    const APP_FILE_USER = 'apache';
    const APP_FILE_MODE = 0644;

    const MODE_BINARY_READ = 'rb';

    private $_pointer;
    private $_path;

    /**
     * @param string $path
     */
    public function __construct($path = '')
    {
        if (!empty($path))
        {
            $this->_path = $path;
        }
    }

    public function __destruct()
    {
        if (!is_null($this->_pointer))
        {
            $this->close();
        }
    }

    /**
     * @return string
     */
    public function getPath()
    {
        return $this->_path;
    }

    /**
     * @return string
     */
    public function getName()
    {
        return basename($this->_path);
    }

    /**
     * @return bool
     */
    public function isExists()
    {
        return $this->_clearstatcache() && file_exists($this->_path);
    }

    /**
     * @return bool
     */
    public function isReadable()
    {
        return $this->_clearstatcache() && is_readable($this->_path);
    }

    /**
     * @return bool
     */
    public function isWritable()
    {
        return $this->_clearstatcache() && is_writable($this->_path);
    }

    /**
     * @param string $mode
     * @return bool|resource
     */
    public function open($mode)
    {
        if ($this->isReadable())
        {
            $this->_pointer = fopen($this->_path, $mode);
            return $this->_pointer;
        }
        return false;
    }

    /**
     * @return bool
     */
    public function close()
    {
        if (is_resource($this->_pointer))
        {
            $return = fclose($this->_pointer);
            $this->_pointer = NULL;
            return $return;
        }
        return false;
    }

    /**
     * @return bool
     */
    public function delete()
    {
        if (file_exists($this->_path))
        {
            return unlink($this->_path);
        }
        return false;
    }

    /**
     * @param      $dataToWrite
     * @param NULL $length
     * @return bool false|int Number of bytes have been written
     */
    public function write($dataToWrite, $length = NULL)
    {
        if (!is_null($this->_pointer))
        {
            if (is_null($length))
            {
                return fwrite($this->_pointer, (string)$dataToWrite);
            }
            else
            {
                return fwrite($this->_pointer, (string)$dataToWrite, $length);
            }
        }
        return false;
    }

    /**
     * @param array $dataToWrite
     * @param string $delimiter
     * @param string $enclosure
     * @return bool false|int Number of bytes have been written
     */
    public function fputcsv(array $dataToWrite, $delimiter = ',', $enclosure = '"')
    {
        if (!is_null($this->_pointer))
        {
            return fputcsv($this->_pointer, $dataToWrite, $delimiter, $enclosure);
        }
        return false;
    }

    /**
     * @param string $group
     * @return bool
     */
    public function chgrp($group)
    {
        try
        {
            return chgrp($this->_path, $group);
        }
        catch (Exception $e)
        {
            Log::getInstance()->error($e->getMessage());
            return false;
        }
    }

    /**
     * @param integer $mode octal number that represents file rights
     * @return bool
     */
    public function chmod($mode)
    {
        try
        {
            return chmod($this->_path, $mode);
        }
        catch (Exception $e)
        {
            Log::getInstance()->error($e->getMessage());
            return false;
        }
    }

    /**
     * @param string $owner
     * @return bool
     */
    public function chown($owner)
    {
        try
        {
            return chown($this->_path, $owner);
        }
        catch (Exception $e)
        {
            Log::getInstance()->error($e->getMessage());
            return false;
        }
    }

    /**
     * @param int $createTime
     * @param int $accessTime
     * @return void
     */
    public function create($createTime = NULL, $accessTime = NULL)
    {
        $currentTime = time();
        if (is_null($createTime))
        {
            $createTime = $currentTime;
        }
        if (is_null($accessTime))
        {
            $accessTime = $currentTime;
        }
        touch($this->_path, $createTime, $accessTime);
    }

    /**
     * @param string $content
     * @param int $flags
     * @param resource $context
     * @return int
     */
    public function putContents($content, $flags = NULL, $context = NULL)
    {
        return file_put_contents($this->_path, $content, $flags, $context);
    }

    /**
     * @param int $flags
     * @param resource $context
     * @param int $offset
     * @param int $maxlen
     * @return string
     * @throws \Exception
     */
    public function getContents($flags = NULL, $context = NULL, $offset = NULL, $maxlen = NULL)
    {
        if (!is_readable($this->_path)) {
            return false;
        }
        if (is_null($maxlen))
        {
            return file_get_contents($this->_path, $flags, $context, $offset);
        }
        return file_get_contents($this->_path, $flags, $context, $offset, $maxlen);
        }

    /**
     * @param string $destinationPath
     * @param resource $context
     * @return bool
     */
    public function copy($destinationPath, $context = NULL)
    {
        if (is_null($context))
        {
            return copy($this->getPath(), $destinationPath);
        }
        return copy($this->getPath(), $destinationPath, $context);
    }

    /**
     * @param int $operation
     * @return bool
     */
    public function lock($operation)
    {
        if (is_null($this->_pointer))
        {
            $this->open('a+');
        }
        return flock($this->_pointer, $operation);
    }

    /**
     * @return string
     */
    public function mime()
    {
        $finfo = finfo_open(FILEINFO_MIME_TYPE);
        $type = finfo_file($finfo, $this->_path);
        finfo_close($finfo);
        return $type;
    }

    /**
     * @link http://php.net/manual/en/function.clearstatcache.php
     * @return bool true
     */
    private function _clearstatcache()
    {
        clearstatcache();
        return true;
    }

    /**
     * @return int Unix timestamp
     */
    public function lastAccessTime()
    {
        $this->_clearstatcache();
        return fileatime($this->_path);
    }

    /**
     * @return int Unix timestamp
     */
    public function createdOn()
    {
        $this->_clearstatcache();
        return filectime($this->_path);
    }

    /**
     * @return int
     */
    public function getSize()
    {
        $this->_clearstatcache();
        return filesize($this->_path);
    }
}
