<?php
/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

class Service_OpenApi_Client
{
    const DEFAULT_PORT = 8440;
    const URL_TEMPLATE = '{proto}://{host}:{port}';
    const CONTENT_TYPE_TEXT_XML = "Content-Type: text/xml";
    const POST = 'POST';
    const RESPONSE = 'response';
    const LOG_OPEN_API_RESPONSE = 'OpenApi response from request (%s) with parameters (%s):\r\n %s';
    const LOG_OPEN_API_REQUEST = 'OpenApi request (%s) with parameters (%s)';
    const RESULT = 'result';
    const HTTPS = 'https';
    const HTTP = 'http';

    /** @var string  */
    private $_apiHost;
    /** @var Service_OpenApi_HttpTransport  */
    private $_transport;
    /** @var Service_OpenApi_XmlRpc  */
    private $_xmlRpc;
    /** @var  Service_ConfigEditor */
    private $_configEditor;
    /** @var  Service_OpenApi_Validator */
    private $_validator;

    /**
     * @param string $apiHost
     * @param Service_OpenApi_IHttpTransport $httpTransport
     * @param Service_OpenApi_XmlRpc $xmlRpc
     * @param Service_ConfigEditor $configEditor
     * @param Service_OpenApi_Validator $validator
     */
    public function __construct(
        $apiHost,
        Service_OpenApi_IHttpTransport $httpTransport = NULL,
        Service_OpenApi_XmlRpc $xmlRpc = NULL,
        Service_ConfigEditor $configEditor = NULL,
        Service_OpenApi_Validator $validator = NULL
    )
    {
        $this->_apiHost = $apiHost;
        $this->_transport = is_null($httpTransport) ? new Service_OpenApi_HttpTransport() : $httpTransport;
        $this->_xmlRpc = is_null($xmlRpc) ? new Service_OpenApi_XmlRpc() : $xmlRpc;
        $this->_configEditor = is_null($configEditor) ? new Service_ConfigEditor() : $configEditor;
        $this->_validator = is_null($validator) ? new Service_OpenApi_Validator() : $validator;
    }

    /**
     * @param string $method
     * @param array $arguments
     * @return array
     * @throws Service_OpenApi_Validator_BaseException
     */
    public function call($method, array $arguments)
    {
        Log::getInstance()->event(sprintf(self::LOG_OPEN_API_REQUEST, $method, print_r($arguments, true)));

        $openApiProtocol = defined('OPEN_API_PROTOCOL') ? OPEN_API_PROTOCOL : self::HTTP;
        $request = $this->_xmlRpc->encodeRequest($method, $arguments);
        try
        {
            $response = $this->_transport->sendRequest(self::POST, $this->_formatApiUri($openApiProtocol), $request);
        }
        catch (Service_OpenApi_HttpTransport_ConnectionException $e)
        {
            Log::getInstance()->warning($e->getMessage());
            $newProtocol = ($openApiProtocol === self::HTTPS) ? self::HTTP : self::HTTPS;
            $response = $this->_transport->sendRequest(self::POST, $this->_formatApiUri($newProtocol), $request);
            $this->_configEditor->updateOpenApiProtocol($newProtocol);
        }
        $decodedResponse = $this->_xmlRpc->decode($response->getBody());

        Log::getInstance()->event(sprintf(
            self::LOG_OPEN_API_RESPONSE,
            $method,
            print_r($arguments, true),
            print_r($decodedResponse, true)
        ));

        $this->_validator->validateResponse($decodedResponse, $response);
        return isset($decodedResponse[self::RESULT]) ? $decodedResponse[self::RESULT] : null;
    }

    /**
     * @param string $httpProtocol
     * @return string
     */
    private function _formatApiUri($httpProtocol = self::HTTPS)
    {
        return strtr(
            self::URL_TEMPLATE,
            array(
                '{proto}' => $httpProtocol,
                '{host}' => $this->_apiHost,
                '{port}' => self::DEFAULT_PORT
            )
        );
    }
}