<?php

/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */
class Service_PbaApi_HttpTransport implements Service_PbaApi_HttpTransportInterface
{
    const EXCEPTION_IO_MESSAGE = 'Can\'t establish TCP/IP connection';
    const TIMEOUT = 300;
    const CONNECT_TIMEOUT = 10;
    const RETURN_TRANSFER = true;
    const SSL_VERIFY_HOST = false;
    const SSL_VERIFY_PEER = false;

    const LOG_MESSAGE_REQUEST_FORMAT = "Request %s to %s request body: (%s)";
    const LOG_MESSAGE_RESPONSE_FORMAT = "Response %s from %s code %s resp_body: (%s)";

    const MARKED_PASSWORD_REPLACE = '******';
    const FILTER_PARAMETER_PASSWORD = 'Password';

    /** @var CurlInterface */
    private $_curl;

    /**
     * @param CurlInterface $curl
     */
    public function __construct(CurlInterface $curl = NULL)
    {
        $this->_curl = is_null($curl) ? new Curl() : $curl;

        $this->_curl->setOption(CURLOPT_TIMEOUT, self::TIMEOUT);
        $this->_curl->setOption(CURLOPT_CONNECTTIMEOUT, self::CONNECT_TIMEOUT);
        $this->_curl->setOption(CURLOPT_RETURNTRANSFER, self::RETURN_TRANSFER);
        $this->_curl->setOption(CURLOPT_SSL_VERIFYHOST, self::SSL_VERIFY_HOST);
        $this->_curl->setOption(CURLOPT_SSL_VERIFYPEER, self::SSL_VERIFY_PEER);
        $this->_curl->setOption(CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);
    }

    /**
     * @param Service_PbaApi_HttpTransport_RequestInterface $request
     * @return Service_PbaApi_HttpTransport_ResponseInterface
     * @throws Service_PbaApi_HttpTransport_IOException
     */
    public function sendRequest(Service_PbaApi_HttpTransport_RequestInterface $request)
    {
        $this->_curl->setOption(CURLOPT_CUSTOMREQUEST, $request->getMethod());
        $this->_curl->setOption(CURLOPT_URL, $request->getUrl());
        $this->_curl->setOption(CURLOPT_POSTFIELDS, $request->getBody());
        $this->_curl->setOption(CURLOPT_HTTPHEADER, $request->getHeaders());

        Log::getInstance()->event(self::LOG_MESSAGE_REQUEST_FORMAT, $request->getMethod(), $request->getUrl(), $request->getBody());
        $responseBody = $this->_curl->execute();
        $errorNo = $this->_curl->getErrorNo();
        if ($errorNo) {
            $errorMessage = $this->_curl->getError();
            Log::getInstance()->error($errorNo . '-' . $errorMessage);
            throw new Service_PbaApi_HttpTransport_IOException($errorMessage, $errorNo);
        }
        $httpCode = $this->_curl->getInfo(CURLINFO_HTTP_CODE);
        $logResponseBody = $this->filterPassword($request, $responseBody);
        Log::getInstance()->event(self::LOG_MESSAGE_RESPONSE_FORMAT, $request->getMethod(), $request->getUrl(), $httpCode, $logResponseBody);
        return new Service_PbaApi_HttpTransport_Response($httpCode, $responseBody);
    }

    /**
     * @param Service_PbaApi_HttpTransport_RequestInterface $request
     * @param string $responseBody
     * @return string
     */
    private function filterPassword(Service_PbaApi_HttpTransport_RequestInterface $request, $responseBody)
    {
        $body = xmlrpc_decode($request->getBody());
        $requestBody = reset($body);
        if (!isset($requestBody[Service_PbaApi_Client::XML_RPC_PARAM_METHOD])
            || $requestBody[Service_PbaApi_Client::XML_RPC_PARAM_METHOD] != Service_PbaApi_Api_Request_GetSubscriptionParameterValue::METHOD
        ) {
            return $responseBody;
        }
        if (!isset($requestBody[Service_PbaApi_Client::XML_RPC_PARAM_PARAMS])
            || !isset($requestBody[Service_PbaApi_Client::XML_RPC_PARAM_PARAMS][1])
            || $requestBody[Service_PbaApi_Client::XML_RPC_PARAM_PARAMS][1] != self::FILTER_PARAMETER_PASSWORD
        ) {
            return $responseBody;
        }
        $responseBodyArray = xmlrpc_decode($responseBody);
        $responseBodyArray[Service_PbaApi_Client::XML_RPC_FIELD_RESULT][0][0] = self::MARKED_PASSWORD_REPLACE;

        return xmlrpc_encode($responseBodyArray);
    }
}