<?php
/**
 * @Copyright © 2002-2018 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\UsageReport;

use Acronis\App\CloudApi\Api;
use Acronis\App\CloudApi\ApiInterface;
use Acronis\App\Database\Model\ReportEntry;
use Acronis\App\Database\Model\ReportEntryHelper;
use Acronis\App\Services\Cache\CacheAwareTrait;
use Acronis\App\Services\Cache\CacheInterface;
use Acronis\App\Services\Config\ConfigSchema;
use Acronis\App\Services\Logger\AppLoggerAwareTrait;
use Acronis\App\Services\Redis\RedisAwareTrait;
use Acronis\App\UsageReport\Aggregation\AggregationStrategy;
use Acronis\App\UsageReport\Aggregation\Aggregator;
use Acronis\App\UsageReport\Exception\ManagerException;
use Acronis\App\Utils\MemoizeTrait;
use APS\ControllerProxy;
use APS\Proto;
use Arr;
use DirectoryIterator;
use Json;
use RecursiveDirectoryIterator;
use RecursiveIteratorIterator;
use Rest\RestException;
use RqlBuilder;
use RuntimeException;
use Service_BaseClient_HTTPException;
use Service_MspApiClient_MspApiFactoryTrait;
use Str;
use Swagger\Client\Model\Reports\Stored\StoredReportParamsItems;
use Swagger\Client\Model\Reports\Stored\StoredReportParamsReportFormat;
use Swagger\Client\Model\Reports\Stored\StoredReportParamsStatus;
use Time;

/**
 * Usage Report Manager, providing basic operations to process usage report , and method
 * getCounters() to get stored in cache counters for subscription
 * @package Acronis\App\UsageReport
 */
class Manager implements ManagerInterface
{
    use CacheAwareTrait;
    use RedisAwareTrait;
    use AppLoggerAwareTrait;
    use MemoizeTrait;
    use Service_MspApiClient_MspApiFactoryTrait;

    const STORAGE_TYPE_REDIS = 'redis';

    const TENANTS_BATCH_SIZE = 100;

    const DOWNLOADED_CSV_REPORT_FILENAME_PATTERN = '%s.csv.gz';

    /** @var string[] */
    private $config;

    /** @var Aggregator */
    private $aggregator;

    /** @var ControllerProxy */
    private $apsc;

    /**
     * Manager constructor.
     * @param array $config
     */
    public function __construct($config)
    {
        $this->config = $config;
    }

    /**
     * @param ControllerProxy $apsc
     */
    public function setApsController($apsc)
    {
        $this->apsc = $apsc;
    }

    /**
     * @param string $subscriptionId
     * @param string $date
     * @param string $datacenterApsId
     * @return mixed
     * @throws ManagerException
     */
    public function getUsagesForSubscription($subscriptionId, $date, $datacenterApsId)
    {
        $storage = $this->createStorage($datacenterApsId);
        $key = (string)new CompositeKey($subscriptionId, $date);

        return $storage->get($key);
    }

    /**
     * @param ReportEntry $reportEntry
     * @param $apsInstanceId
     * @return Aggregator
     * @throws ManagerException
     * @throws \Service_BaseClient_HTTPException
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     * todo remove $apsInstanceId after adding tenantId to reseller's map
     */
    public function getAggregator($reportEntry, $apsInstanceId)
    {
        if (is_null($this->aggregator)) {
            $this->aggregator = $this->createAggregatorInstance($reportEntry->datacenter_id, $apsInstanceId);
        }

        return $this->aggregator;
    }

    /**
     * @param ReportEntry $reportEntry
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     * @throws ManagerException
     */
    public function orderReport($reportEntry)
    {
        $api = $this->createCloudApiForDatacenter($reportEntry->datacenter_id);
        $datacenterTenant = $api->getMe();
        $datacenterTenantId = $datacenterTenant->getTenantId();

        // Remove this method, create method for ordering current report after fixing RAML or lib
        $reportData = $api->orderCurrentUsageReportForAccounts($datacenterTenantId);

        $reportEntry->report_id = $reportData->getId();
        $reportEntry->status = ReportEntry::REPORT_STATUS_ORDERED;
        $reportEntry->save();
    }

    /**
     * @param ReportEntry $reportEntry
     * @throws ManagerException
     */
    public function checkForStoredReport($reportEntry)
    {
        $api = $this->createCloudApiForDatacenter($reportEntry->datacenter_id);
        $storedReport = $this->getStoredReport($reportEntry, $api);

        if ($storedReport->getStatus() === StoredReportParamsStatus::SAVED) {
            $reportEntry->status = ReportEntry::REPORT_STATUS_READY;
            $reportEntry->stored_report_id = $storedReport->getId();
            $reportEntry->save();
        }

        if ($storedReport->getStatus() === StoredReportParamsStatus::FAILED) {
            $reportEntry->status = ReportEntry::REPORT_STATUS_ERROR;
            $reportEntry->save();
        }
    }

    /**
     * @param ReportEntry $reportEntry
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     * @throws ManagerException
     */
    public function downloadReport($reportEntry)
    {
        $api = $this->createCloudApiForDatacenter($reportEntry->datacenter_id);
        $downloadPath = $this->resolveReportDownloadPath($reportEntry);
        if (!file_exists($downloadPath)) {
            mkdir($downloadPath);
        }

        $destinationFilePath = $this->generateReportPath($reportEntry);

        $api->downloadStoredUsageReportArchive($reportEntry->report_id, $reportEntry->stored_report_id,
            $destinationFilePath);

        $reportEntry->status = ReportEntry::REPORT_STATUS_DOWNLOADED;
        $reportEntry->file_path = $destinationFilePath;
        $reportEntry->save();
    }

    /**
     * Removes all saved reports and {date} folders from /tmp/reports/{date}/{id}.csv.gz
     */
    public function cleanReportDownloadPath()
    {
        $downloadPath = $this->getReportsBasePath();
        $this->getLog()->notice('Cleaning up reports download path recursively.');
        $this->removeFilesRecursively($downloadPath);
        $this->getLog()->notice('Done cleaning up reports download path.');
    }

    /**
     * After this ttl all downloaded report files should be deleted. -1 means infinite ttl.
     * @return int
     */
    public function getReportDownloadedFilesTtlInDays()
    {
        return Arr::get($this->config, ConfigSchema::NODE_USAGE_REPORT_DOWNLOADED_FILES_TTL_IN_DAYS, -1);
    }

    public function eraseReports()
    {
        $ttl = $this->getReportDownloadedFilesTtlInDays();
        $reportHelper = new ReportEntryHelper();

        if ($ttl === -1) {
            return;
        }

        $dateTill = Time::offsetDate(new \DateTime(), -$ttl);
        $reportsToErase = $reportHelper->getReportsTillDate($dateTill, false);

        foreach ($reportsToErase as $reportEntry) {
            $this->eraseReport($reportEntry);
        }

        $this->removeEmptyFoldersInReportsDownloadsPath();
    }

    /**
     * @param Aggregator $aggregator
     * @param ReportEntry $reportEntry
     * @throws \Exception
     * @throws RuntimeException
     */
    public function aggregateReport($aggregator, $reportEntry)
    {
        $date = $reportEntry->date;
        $reportIterator = $this->createReportIterator($reportEntry->file_path);

        $aggregator->aggregateReportUsage($reportIterator, $date);
        $reportEntry->status = ReportEntry::REPORT_STATUS_AGGREGATED;
        $reportEntry->save();
    }

    /**
     * @return string
     */
    public function getRetryTimeout()
    {
        return $this->config[ConfigSchema::NODE_USAGE_REPORT_RETRY_TIMEOUT];
    }

    /**
     * @return string
     */
    public function getRetriesLimit()
    {
        return $this->config[ConfigSchema::NODE_USAGE_REPORT_RETRIES_LIMIT];
    }

    /**
     * @param string $filePath
     * @return CsvReportIterator
     * @throws RuntimeException
     */
    protected function createReportIterator($filePath)
    {
        return new CsvReportIterator($filePath, new CsvFormatParser());
    }

    /**
     * @param ReportEntry $reportEntry
     */
    private function eraseReport($reportEntry)
    {
        $this->getLog()->notice(
            'Erasing report entry #%s.',
            $reportEntry->id
        );

        if ($reportEntry->status >= ReportEntry::REPORT_STATUS_DOWNLOADED) {
            $filePath = $reportEntry->file_path;
            if (file_exists($filePath)) {
                unlink($filePath);
                $this->getLog()->notice(
                    'Removed report #%s downloaded file "%s".',
                    $reportEntry->id, $filePath
                );
            } else {
                $this->getLog()->warning(
                    'Report downloaded file "%s" does\'nt exists.',
                    $filePath
                );
            }
        }

        $reportEntry->file_path = null;
        $reportEntry->status = ReportEntry::REPORT_STATUS_ERASED;
        $reportEntry->save();

        $this->getLog()->notice(
            'Report entry #%s marked as "erased".',
            $reportEntry->id
        );
    }

    private function removeFilesRecursively($path)
    {
        $files = new RecursiveIteratorIterator(
            new RecursiveDirectoryIterator($path, RecursiveDirectoryIterator::SKIP_DOTS),
            RecursiveIteratorIterator::CHILD_FIRST
        );

        foreach ($files as $fileInfo) {
            $path = $fileInfo->getRealPath();
            $fileInfo->isDir() ? rmdir($path) : unlink($path);
        }
    }

    private function removeEmptyFoldersInReportsDownloadsPath()
    {
        $downloadPath = $this->getReportsBasePath();
        foreach (new DirectoryIterator($downloadPath) as $dirInfo) {
            if (!$dirInfo->isDir() || $dirInfo->isDot()) {
                continue;
            }

            $dirPath = $dirInfo->getRealPath();
            $isDirEmpty = !(new \FilesystemIterator($dirPath))->valid();

            if ($isDirEmpty) {
                rmdir($dirPath);
                $this->getLog()->notice(
                    'Removed empty report files folder "%s".',
                    $dirPath
                );
            }
        }
    }

    /**
     * @param string $datacenterApsId
     * @param $apsInstanceId
     * @return Aggregator
     * @throws ManagerException
     * todo remove $apsInstanceId after adding tenantId to reseller's map
     * @throws \Service_BaseClient_HTTPException
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    private function createAggregatorInstance($datacenterApsId, $apsInstanceId)
    {
        $cache = $this->createCache($datacenterApsId);
        $storage = $this->createStorage($datacenterApsId);
        $cloudApi = $this->createCloudApiForDatacenter($datacenterApsId);
        $tenants = $this->fetchTenantsMap($datacenterApsId, $apsInstanceId);
        $strategy = new AggregationStrategy($cloudApi, $this->getApsc());

        return new Aggregator($cache, $storage, $strategy, $tenants);
    }

    /**
     * @param ReportEntry $reportEntry
     * @return string
     */
    private function resolveReportDownloadPath($reportEntry)
    {
        return implode(DS, [
            $this->getReportsBasePath(),
            $reportEntry->date,
        ]);
    }

    /**
     * @param ReportEntry $reportEntry
     * @return string
     */
    private function generateReportPath($reportEntry)
    {
        return $this->resolveReportDownloadPath($reportEntry) . DS .
            Str::format(
                static::DOWNLOADED_CSV_REPORT_FILENAME_PATTERN,
                $reportEntry->id
            );
    }

    /**
     * @param $datacenterApsId
     * @param $apsInstanceId
     * @return \tenant[]
     * @throws ManagerException
     * @throws \Service_BaseClient_HTTPException
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     * todo remove $apsInstanceId after adding tenantId to reseller's map
     */
    private function fetchTenantsMap($datacenterApsId, $apsInstanceId)
    {
        $this->getLog()->notice('Fetching tenants mapping.');

        $tenantsMap = [];
        $path = $this->getApsc()->getIo()->resourcePath();

        $datacenter = $this->fetchDatacenter($datacenterApsId);
        if (is_null($datacenter)) {
            throw new RuntimeException(Str::format(
                'Unable to fetch datacenter with ID "%s".',
                $datacenterApsId
            ));
        }
        $accountToTenantMap = $this->buildResellersAccountToTenantMap($datacenter, $apsInstanceId);

        $fetchTenantsResources = function ($offset, $limit) use ($datacenterApsId, $path) {
            $rqlFilter = RqlBuilder::createInstance()
                ->implementing(\APSResource::CLOUD_BACKUP_TENANT_TYPE)
                ->eq('aps.status', 'aps:ready')
                ->limit($offset, $limit)
                ->select(['tenantId', 'subscriptionId', 'type', 'accountId', 'datacenter'])
                ->getQuery();
            $response = $this->getApsc()->getIo()->sendRequest(Proto::GET, $path . '?' . $rqlFilter);

            return Json::decode($response);
        };

        $sanitizeTenantsMap = function (&$tenantsMap) {
            /** @var \tenant[] $tenantsMap */
            array_walk($tenantsMap, function (&$tenantResource, $key) {
                unset ($tenantResource->aps);
                unset ($tenantResource->accountId);
                unset ($tenantResource->datacenter);
            }, $tenantsMap);
        };

        for ($offset = 0; true; $offset += static::TENANTS_BATCH_SIZE) {
            $tenants = $fetchTenantsResources($offset, static::TENANTS_BATCH_SIZE);
            if (!count($tenants)) {
                break;
            }

            /** @var \tenant $tenant */
            foreach ($tenants as $tenant) {
                if ($tenant->type === \tenant::TYPE_RESELLER) {
                    if (!isset($accountToTenantMap[$tenant->accountId])) {
                        $this->getLog()->warning(
                            'There is no reseller relation for account "%s".',
                            $tenant->accountId
                        );
                        continue;
                    }

                    $tenant->tenantId = $accountToTenantMap[$tenant->accountId];
                } else if ($tenant->datacenter->aps->id != $datacenterApsId) {
                    continue;
                }

                $tenantsMap[$tenant->tenantId] = $tenant;
            }
        }

        $sanitizeTenantsMap($tenantsMap);

        $this->getLog()->notice('Finished fetching tenants mapping.');

        return $tenantsMap;
    }

    /**
     * @param \datacenter $datacenter
     * @param $apsInstanceId
     * @return object[]
     * @throws ManagerException
     * @throws \Service_BaseClient_HTTPException
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     * todo remove $apsInstanceId after adding tenantId to reseller's map
     */
    private function buildResellersAccountToTenantMap($datacenter, $apsInstanceId)
    {
        $this->getLog()->notice('Building resellers account to tenant mapping.');

        $accountToTenantMap = [];
        $global = $this->getGlobal();
        $api = $this->createMspApiInstance($datacenter->url, $datacenter->login, $datacenter->password, $apsInstanceId);

        /** @var \ResellerRelation[] $resellerRelation */
        $resellerRelation = $global->resellerRelations ?: [];

        foreach ($resellerRelation as $relation) {
            if ($relation->datacenterId !== $datacenter->id) {
                continue;
            }

            try {
                $accountToTenantMap[$relation->accountId] = $this->resolveTenantId($relation->groupId, $api);
            } catch (Service_BaseClient_HTTPException $e) {
                if ($e->getCode() != 404) {
                    throw $e;
                }

                $this->getLog()->warning(
                    'Unable to resolve tenant UUID, group ID "%s".',
                    $relation->groupId
                );
            }
        }

        $this->getLog()->notice('Successfully built account to tenant mapping.');

        return $accountToTenantMap;
    }

    /**
     * @return \globals|null
     * @throws ManagerException
     */
    private function getGlobal()
    {
        $rqlFilter = RqlBuilder::createInstance()
            ->implementing(\APSResource::CLOUD_BACKUP_GLOBALS_TYPE)
            ->getQuery();
        /** @var \globals[]|null $globals */
        $globals = $this->getApsc()->getResources($rqlFilter);

        if (empty($globals)) {
            throw new RuntimeException('Unable to fetch reseller relation information.');
        }

        return reset($globals);
    }

    /**
     * @param $groupId
     * @param \Service_MspApiClient_ApiInterface $api
     * @return string
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     * todo remove method and its usage after adding tenantId to reseller's map
     */
    private function resolveTenantId($groupId, $api)
    {
        $group = $api->getGroupById($groupId);

        return $group->getUuid();
    }

    /**
     * @return ControllerProxy
     * @throws ManagerException
     */
    private function getApsc()
    {
        if (is_null($this->apsc)) {
            throw new ManagerException('APS Controller must be set before using Usage Report Manager.');
        }

        return $this->apsc;
    }

    /**
     * @return CacheInterface
     * @throws ManagerException
     */
    private function createCache($datacenterApsId)
    {
        $keyPrefix = $this->getPrefix() . $datacenterApsId . '_';

        return new ReportCache($keyPrefix);
    }

    /**
     * @return ReportStorageInterface
     * @param string $datacenterApsId
     * @throws ManagerException
     */
    private function createStorage($datacenterApsId)
    {
        $storageType = $this->config[ConfigSchema::NODE_USAGE_REPORT_STORAGE];

        if ($storageType !== static::STORAGE_TYPE_REDIS) {
            throw new ManagerException(Str::format(
                'Unknown storage type "%s".',
                $storageType
            ));
        }

        $keyPrefix = $this->getPrefix() . $datacenterApsId . '_';
        $ttl = Arr::get($this->config, ConfigSchema::NODE_USAGE_REPORT_TTL);

        return new RedisStorage($this->getRedis(), $keyPrefix, $ttl);
    }

    /**
     * @param ReportEntry $reportEntry
     * @param ApiInterface $api
     * @throws ManagerException
     * @return StoredReportParamsItems
     */
    private function getStoredReport($reportEntry, $api)
    {
        $storedReports = $api->getStoredUsageReports($reportEntry->report_id);
        $suitableReports = array_filter($storedReports, function ($storedReport) {
            /** @var StoredReportParamsItems $storedReport */
            return $storedReport->getReportFormat() === StoredReportParamsReportFormat::CSV_V2_0;
        });

        if (!count($suitableReports)) {
            throw new ManagerException(Str::format(
                'There are no stored reports of format "%s" for report #%s.',
                StoredReportParamsReportFormat::CSV_V2_0, $reportEntry->id
            ));
        }

        return $suitableReports[0];
    }

    /**
     * @param string $datacenterApsId
     * @return mixed|null
     * @throws ManagerException
     * @throws \Exception
     */
    private function fetchDatacenter($datacenterApsId, $forceFetch = false)
    {
        return $this->memoize(function () use ($datacenterApsId) {
            $apsc = $this->getApsc();
            $query = RqlBuilder::createInstance()
                ->implementing(\APSResource::CLOUD_BACKUP_DATACENTER_TYPE)
                ->eq('aps.id', $datacenterApsId)
                ->getQuery();

            try {
                $datacenters = $apsc->getResources($query);
            } catch (RestException $e) {
                $this->getLog()->error($e->getMessage());
                throw new ManagerException(Str::format(
                    'Unable to fetch datacenter "%s". Please check command argument "apsInstanceId".',
                    $datacenterApsId
                ));
            }

            return count($datacenters) ? reset($datacenters) : null;
        }, $datacenterApsId, $forceFetch);
    }

    /**
     * @param string $datacenterApsId
     * @return Api
     * @throws ManagerException
     */
    private function createCloudApiForDatacenter($datacenterApsId)
    {
        $datacenter = $this->fetchDatacenter($datacenterApsId);

        return new Api($datacenter->url, $datacenter->login, $datacenter->password);
    }

    /**
     * @return string
     */
    private function getReportsBasePath()
    {
        return $this->config[ConfigSchema::NODE_USAGE_REPORT_DOWNLOAD_PATH];
    }

    /**
     * @return string
     */
    private function getPrefix()
    {
        return Arr::get($this->config, ConfigSchema::NODE_USAGE_REPORT_PREFIX, '');
    }
}