<?php
/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\Utils;

use Acronis\App\Services\Cache\CacheFactory;
use Acronis\App\Services\Cache\CacheInterface;
use Acronis\App\Services\Locator;
use Acronis\App\Services\Logger\AppLoggerAdapter;
use Acronis\App\Services\Logger\AppLoggerFactory;

/**
 * Trait CacheableTrait
 * Use CacheableTrait in cases you want to fetch/build some object identified by $objectKey only one time and then
 * return it from cache every time when possible
 * $this->fromCache(function() use ($anyData) { return $this->veryExpensiveFetchRequest($objectKey); }, objectKey, [force]);
 * If you want to be able to fallback to fetching/building object every time, pass last parameter as true
 * @package Acronis\App\Utils
 */
trait CacheableTrait
{
    /**
     * @param callable $method
     * @param int|null $ttl
     * @param bool $refreshTtl
     * @param string $objectId
     * @param bool $force
     * @return mixed
     */
    protected function fromCache(callable $method, $ttl = null, $refreshTtl = false, $objectId = '', $force = false)
    {
        $callerName = Fn::getCallerName();
        $objectId = $this->buildCacheableObjectKey($callerName, $objectId);

        try {
            if (!$force && $this->hasCacheableStoredObject($objectId)) {
                return $this->getCacheableStoredObject($objectId, $ttl, $refreshTtl);
            }
        } catch (\Exception $e) {
            /** @var AppLoggerAdapter $log */
            $log = Locator::getInstance()->get(AppLoggerFactory::NAME);
            $log->error(
                'Unable to fetch object with id "%s" from cache. Error: %s',
                $objectId, $e->getMessage()
            );
            return $method();
        }

        $object = $method();
        $this->storeCacheableObject($objectId, $object, $ttl);

        return $object;
    }

    /**
     * @param int|null $ttl
     * @param string $objectId
     * @param mixed $value
     */
    private function storeCacheableObject($objectId, $value, $ttl)
    {
        $this->getCacheInstance()->set($objectId, $value, $ttl);
    }

    /**
     * @param string $objectId
     * @param int|null
     * @param bool $refreshTtl
     * @return mixed
     */
    private function getCacheableStoredObject($objectId, $ttl, $refreshTtl)
    {
        $result = $this->getCacheInstance()->get($objectId);

        if ($refreshTtl && $this->getCacheInstance()->has($objectId)) {
            $this->getCacheInstance()->refresh($objectId, $ttl);
        }

        return $result;
    }

    /**
     * @param string $objectId
     * @return bool
     */
    private function hasCacheableStoredObject($objectId)
    {
        return $this->getCacheInstance()->has($objectId);
    }

    /**
     * @param string $method
     * @param string $objectId
     * @return string
     */
    private function buildCacheableObjectKey($method, $objectId)
    {
        return implode('|', [$method, $objectId]);
    }

    /**
     * @return CacheInterface
     */
    private function getCacheInstance()
    {
        return Locator::getInstance()->get(CacheFactory::NAME);
    }
}