<?php
/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

class Session_Store implements Session_StoreInterface
{
    private static $_instance;
    private $_started = false;

    const SESSION_EXPIRE_KEY = 'expire';
    const SESSION_DATA_KEY = 'data';
    const SESSION_TTL_KEY = 'ttl';
    const SESSION_DEFAULT_TTL = 90;

    const PARAM_COOKIE_SECURE = 'session.cookie_secure';
    const PARAM_COOKIE_HTTPONLY = 'session.cookie_httponly';
    const PARAM_COOKIE_ENABLED = 'session.use_cookies';
    const PARAM_MAX_LIFETIME = 'session.gc_maxlifetime';

    /**
     * init session ini parameters
     */
    private function __construct()
    {
        // Set secure HTTP-only cookie
        ini_set(self::PARAM_COOKIE_ENABLED, 0);
        ini_set(self::PARAM_COOKIE_HTTPONLY, 0);
        ini_set(self::PARAM_COOKIE_SECURE, 0);
        ini_set(self::PARAM_MAX_LIFETIME, 0);

        //set custom session directory
        session_save_path(SESSION_SAVE_PATH);
        register_shutdown_function('session_write_close');
    }

    /**
     * Singleton implementation
     *
     * @return Session_Store object
     */
    public static function getInstance()
    {
        if(is_null(self::$_instance))
        {
            self::$_instance = new Session_Store();
        }
        return self::$_instance;
    }

    /**
     * Set session id
     *
     * @param $id
     * @return string
     */
    public function setSessionId($id)
    {
        return session_id($id);
    }

    /**
     * start session if it's not started
     *
     * @return true
     */
    public function startSession()
    {
        if(!$this->_started)
        {
            $this->_started = true;
            session_start();
        }
        return true;
    }

    /**
     * stop session if it's started
     * @return boolean
     */
    public function stopSession()
    {
        if($this->_started)
        {
            $this->_started = false;
            session_write_close();
        }
        return true;
    }

    /**
     * increse for session_lifetime current expire value for key
     * @param string $key
     */
    public function increaseExpireTime($key)
    {
        $ttl = isset($_SESSION[$key][self::SESSION_TTL_KEY]) ? intval($_SESSION[$key][self::SESSION_TTL_KEY]) : self::SESSION_DEFAULT_TTL;
        $_SESSION[$key][self::SESSION_EXPIRE_KEY] = time() + $ttl;
    }

    /**
     * Sets value to session by key and create expire time for this value
     *
     * @param string $key
     * @param mixed $value
     * @param int|null $ttl
     */
    public function setValue($key,$value,$ttl=NULL)
    {
        $_SESSION[$key][self::SESSION_DATA_KEY] = $value;
        $_SESSION[$key][self::SESSION_TTL_KEY] = $ttl;
        $this->increaseExpireTime($key);
    }

    /**
     * Return value from session by key and update expire time
     *
     * @param string $key
     * @return mixed
     */
    public function getValue($key)
    {
        if(isset($_SESSION[$key]) && !self::isExpired($key))
        {
            $this->increaseExpireTime($key);
            return $_SESSION[$key][self::SESSION_DATA_KEY];
        }
        return null;
    }

    /**
     * Check that session value by key is not expired
     *
     * @param string $key
     * @return bool
     */
    public function isExpired($key)
    {
        return ($_SESSION[$key][self::SESSION_EXPIRE_KEY] < time());
    }

    /**
     * Delete value from session by key
     *
     * @param string $key
     */
    public function unsetValue($key)
    {
        $data = $this->getValue($key);
        unset($_SESSION[$key]);
        return $data;
    }
}
