<?php

/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */
class BillingHelper
{
    const PROPERTY_CONTROLLER_URI = "controller_uri";
    const PROPERTY_APS_TOKEN = "aps_token";

    const APS_UNLIMITED_VALUE = -1;
    const APS_APPLICATIONS_PATH = '/aps/2/applications';
    const APS_TOKEN_HEADER = 'APS-Token: ';

    const BILLING_ORDER_NAME = 'Minimum commitment';

    const ORDER_TYPE_BO = 'BO';
    const ORDER_TYPE_CH = 'CH';

    const ORDER_STATUS_NEW = 'NW';
    const ORDER_STATUS_OPEN = 'OP';

    const PROVIDER_ACCOUNT_ID = 1;
    const PROVIDER_SUBSCRIPTION_ID = 1;

    const DETAIL_TYPE_PLAN_RECURRING = 110;
    const DETAIL_TYPE_RESOURCE_RECURRING = 130;
    const DETAIL_TYPE_RESOURCE_OVERUSAGE = 140;

    const ACCOUNT_PROCESS_TYPE_ONBOARDING = 'onboarding';
    const ACCOUNT_PROCESS_TYPE_CANCELLATION = 'cancellation';

    /**
     * @var ApiFactory
     */
    private $apiFactory;

    /**
     * @var Service_OpenApi_Api
     */
    private $openApi;

    /**
     * @var Service_PbaApi_Api
     */
    private $pbaApi;

    /**
     * @var \APS\ControllerProxy
     */
    private $apsController;

    /**
     * @var \APS\ResourceBase
     */
    private $globals;

    /**
     * @var string
     */
    private $endpointHost;

    /**
     * @var int
     */
    private $applicationId;

    /**
     * @var DateTime
     */
    private $startPeriodDate;

    /**
     * @var DateTime
     */
    private $endPeriodDate;

    /**
     * @var Service_PbaApi_Api_Response_Struct_Account[]
     */
    private $accountsInfo;


    /**
     * BillingHelper constructor.
     * @param ApiFactory $apiFactory
     * @param int $applicationId
     */
    public function __construct($apiFactory, $applicationId)
    {
        $this->apiFactory = $apiFactory;
        $this->applicationId = $applicationId;

        $this->openApi = $this->createOpenApiInstance();
        $this->pbaApi = $this->createPbaApiInstance();
    }

    public function process()
    {
        Log::getInstance()->notice('Processing subscriptions.');

        $applicationInstances = $this->openApi->getApplicationInstances($this->applicationId);
        $apsApplicationInstances = $this->fetchApsApplicationInstances();
        foreach ($applicationInstances as $applicationInstance) {
            $appInstanceId = $applicationInstance['application_instance_id'];
            $appResourceId = $applicationInstance['application_resource_id'];
            $apsInstanceId = $this->findApsInstanceIdByResourceId($apsApplicationInstances, $appResourceId);
            try {
                $this->processApplicationInstance($appInstanceId, $apsInstanceId, $appResourceId);
            } catch (Exception $e) {
                Log::getInstance()->error($e->getMessage());
            }
        }
    }

    /**
     * @param $appInstanceId
     * @param $apsInstanceId
     * @param $appResourceId
     */
    private function processApplicationInstance($appInstanceId, $apsInstanceId, $appResourceId)
    {
        Log::getInstance()->notice('Processing application instance "%s".', $appInstanceId);

        $this->apsController = $this->createApsControllerInstance($appInstanceId, $apsInstanceId);
        $this->globals = $this->apsController->getResource($appResourceId);
        $this->endpointHost = $this->fetchEndPointHost();

        $this->processAccounts();
        $this->processSubscriptions();
    }

    private function processAccounts()
    {
        $accounts = $this->getAccountsFromProcessList();
        if (empty($accounts)) {
            Log::getInstance()->notice('There are no accounts to process.');
            return;
        }

        Log::getInstance()->notice('Processing accounts.');
        foreach ($accounts as $record) {
            Log::getInstance()->notice(str_repeat('=', 80));
            try {
                $this->processAccount($record);
            } catch (Exception $e) {
                Log::getInstance()->error($e->getMessage());
            }
        }
        Log::getInstance()->notice(str_repeat('=', 80));
    }

    /**
     * @param $record
     * @throws Exception
     */
    private function processAccount($record)
    {
        $accountId = $record->accountId;
        $switchTime = $record->date;
        $type = $record->type;
        $subscriptions = $record->subscriptions;

        Log::getInstance()->notice(
            'Processing the account %s to do %s for date %s.',
            $accountId, $type, date('Y-m-d', $switchTime)
        );

        if (!in_array($type, array(
            BillingHelper::ACCOUNT_PROCESS_TYPE_ONBOARDING,
            BillingHelper::ACCOUNT_PROCESS_TYPE_CANCELLATION
        ))
        ) {
            throw new Exception(Str::format(
                'Unsupported action type "%s" for account %s.',
                $type, $accountId
            ));
        }

        if (empty($subscriptions)) {
            Log::getInstance()->notice('Account %s does not have customers\' subscriptions.', $accountId);
            $this->removeAccountFromProcessList($record);
            return;
        }

        if (!$this->needToProcessAccount($switchTime)) {
            Log::getInstance()->notice(
                'Skipping the account %s. The switch date is %s.',
                $accountId, date('Y-m-d', $switchTime)
            );
            return;
        }

        Log::getInstance()->notice(
            'Fetching deliver lines for the account %s by date %s.',
            $accountId, date('Y-m-d', $switchTime)
        );
        $deliverLines = $this->getDeliverLinesForDate($accountId, $switchTime);
        if (empty($deliverLines)) {
            Log::getInstance()->notice(
                'There are no suitable deliver lines for the account %s by date %s.',
                $accountId, date('Y-m-d', $switchTime)
            );
            $this->removeAccountFromProcessList($record);
            return;
        }

        Log::getInstance()->notice('Filling order details from the deliver lines.');
        $accountInfo = $this->fetchAccountInfo($accountId);
        $orderDetails = $this->makeOrderDetailsFromDeliverLines($accountInfo->getVendorAccountID(), $deliverLines, $type, $subscriptions);
        if (empty($orderDetails)) {
            Log::getInstance()->notice('There are no order details to place.');
            $this->removeAccountFromProcessList($record);
            return;
        }

        $this->pbaApi->startTransaction();
        try {
            $orderInfo = $this->placeAdjustingOrderForAccount($accountId, $orderDetails);
            Log::getInstance()->notice('%s %s', $orderInfo->getOrderNumber(), $orderInfo->getComments());
            $this->removeAccountFromProcessList($record);
        } catch (Exception $e) {
            $this->pbaApi->rollbackTransaction();
            throw  $e;
        }
        $this->pbaApi->commitTransaction();
    }

    private function needToProcessAccount($switchTime)
    {
        $timeToProcess = new DateTime();
        $timeToProcess->setTimestamp($switchTime);
        $timeToProcess->setTime(23, 59, 59);
        $timeToProcess->modify("+31 day");
        return $timeToProcess->getTimestamp() < time();
    }

    private function getAccountsFromProcessList()
    {
        $accounts = $this->globals->accountToProcess;
        if (!is_array($accounts)) {
            return array();
        }
        usort($accounts, function ($record1, $record2) {
            return $record1->date - $record2->date;
        });
        return $accounts;
    }

    /**
     * @param $record
     * @throws Exception
     */
    private function removeAccountFromProcessList($record)
    {
        Log::getInstance()->notice('Removing the account %s from the processing list.', $record->accountId);
        if (!is_array($this->globals->accountToProcess)) {
            throw new Exception(Str::format('Processing list is empty.'));
        }
        $this->globals->accountToProcess = array_values(array_filter(
            $this->globals->accountToProcess,
            function ($account) use ($record) {
                return $account->accountId != $record->accountId ||
                    $account->date != $record->date ||
                    $account->type != $record->type;
            }
        ));
        $this->apsController->updateResource($this->globals);
    }

    private function processSubscriptions()
    {
        // todo: add paging for $tenants
        $tenants = $this->fetchTenants();
        if (!count($tenants)) {
            Log::getInstance()->notice('There are no subscriptions to process.');
            return;
        }
        Log::getInstance()->notice('Processing subscriptions.');
        foreach ($tenants as $tenant) {
            Log::getInstance()->notice(str_repeat('=', 80));
            try {
                $this->processSubscription($tenant);
            } catch (Exception $e) {
                Log::getInstance()->error($e->getMessage());
            }
        }
        Log::getInstance()->notice(str_repeat('=', 80));
        Log::getInstance()->notice('All subscriptions have been processed.');
    }

    /**
     * @param $tenant
     * @throws Exception
     */
    private function processSubscription($tenant)
    {
        $subscription = $tenant->subscription;

        $accountId = $subscription->account->id;
        $subscriptionId = $subscription->subscriptionId;

        Log::getInstance()->notice(
            'Processing the subscription "%s" for the account "%s".',
            $subscriptionId,
            $accountId
        );

        $accountInfo = $this->fetchAccountInfo($accountId);
        if (is_null($accountInfo->getStatementDay())) {
            throw new Exception(
                Str::format('Subscription "%s" was skipped because of "Statement Day" is not set for the account.', $subscriptionId)
            );
        }
        $this->initPeriodByAccountInfo($accountInfo);
        if (!$this->needToProcessSubscription($accountInfo, $tenant)) {
            Log::getInstance()->notice(
                'Subscription "%s" was skipped because of there are no billing orders for it during the period from %s to %s.',
                $subscriptionId,
                $this->startPeriodDate->format('c'),
                $this->endPeriodDate->format('c')
            );
            return;
        }

        $subscriptionResources = $subscription->resources();
        $availableTiersResources = $this->filterSubscriptionResources(
            $subscriptionResources,
            \APS\TypeLibrary::getTypeByServiceId('minimalCommitmentTier')
        );
        if (!count($availableTiersResources)) {
            Log::getInstance()->notice(
                'Subscription "%s" does not support volume pricing.',
                $subscriptionId
            );
            return;
        }

        $enabledTiersResources = $this->filterSubscriptionResourcesByLimitNotEqual($availableTiersResources, 0);
        $enabledTiersCount = count($enabledTiersResources);
        if (!$enabledTiersCount) {
            throw new Exception(Str::format(
                'Subscription "%s" does not have any minimum commitment tier included.',
                $subscriptionId
            ));
        }
        if ($enabledTiersCount > 1) {
            throw new Exception(Str::format(
                'Subscription "%s" has multiple minimum commitment tiers included.',
                $subscriptionId
            ));
        }
        $currentTierResource = reset($enabledTiersResources);
        $currentTier = $tenant->minimalCommitmentTier;

        if ($currentTier->aps->id !== $currentTierResource->apsId) {
            throw new Exception(Str::format(
                'The tenant is not linked with the tier "%s" which is represented by the resource "%s" included into the subscription.',
                $currentTier->aps->id,
                $currentTierResource->apsId
            ));
        }

        Log::getInstance()->notice('Getting currency for account "%s".', $accountId);
        $currencyInfo = $this->pbaApi->getVendorCurrency($accountId);
        $currency = $currencyInfo->getCurrencyID();
        Log::getInstance()->notice('Currency for account "%s" is "%s".', $accountId, $currency);

        Log::getInstance()->notice('Getting minimum commitment from tier "%s" by currency "%s".', $currentTier->resourceTypeId, $currency);
        $minimalCommitment = $this->getMinimalCommitment($currentTier, $currency);
        if (is_null($minimalCommitment)) {
            throw new Exception(Str::format('Minimum commitment tier "%s" does not have a value for currency "%s".', $currentTier->resourceTypeId, $currency));
        }
        Log::getInstance()->notice('Minimum commitment for subscription "%s" is %s %s.', $subscriptionId, $minimalCommitment, $currency);

        Log::getInstance()->notice('Calculating total amount for the subscription "%s" by billing orders.', $subscriptionId);
        Log::getInstance()->notice(
            'Processing orders for the period from %s to %s.',
            $this->startPeriodDate->format('c'),
            $this->endPeriodDate->format('c')
        );

        $billingOrders = $this->fetchSubscriptionBillingOrders($subscriptionId);
        $totalAmount = $this->calculateBillingOrderTotalAmount($subscriptionId, $billingOrders);
        Log::getInstance()->notice('Total amount for the subscription "%s" is %s %s.', $subscriptionId, $totalAmount, $currency);
        if ($totalAmount < $minimalCommitment) {
            $orderAmount = $minimalCommitment - $totalAmount;
            Log::getInstance()->notice('Placing a billing order for the subscription "%s" on %s %s.', $subscriptionId, $orderAmount, $currency);
            $orderInfo = $this->placeBillingOrder($accountId, $subscriptionId, $orderAmount);
            Log::getInstance()->notice('%s %s', $orderInfo->getOrderNumber(), $orderInfo->getComments());
            return;
        }

        Log::getInstance()->notice('Finding the next tier for the subscription "%s".', $subscriptionId);
        $nextSuitableTier = $this->findNextSuitableTier($minimalCommitment, $totalAmount, $currency, $availableTiersResources);
        if (is_null($nextSuitableTier)) {
            Log::getInstance()->notice('The next tier for the subscription "%s" was not found.', $subscriptionId);
            return;
        }
        Log::getInstance()->notice('The next tier "%s" for the subscription "%s" is found.', $nextSuitableTier->resourceTypeId, $subscriptionId);

        Log::getInstance()->notice('Finding suitable plan for the subscription "%s" by tier "%s".', $subscriptionId, $nextSuitableTier->resourceTypeId);
        $planId = $this->findAvailablePlanWithResource($subscriptionId, $nextSuitableTier->resourceTypeId);
        if (is_null($planId)) {
            throw new Exception('Cannot find any plan for the subscription upgrade.');
        }

        Log::getInstance()->notice(
            'Placing the order for switching the subscription "%s" to the plan "%s".',
            $subscriptionId,
            $planId
        );
        $orderInfo = $this->placePlanSwitchOrder($subscriptionId, $planId);
        Log::getInstance()->notice('%s %s', $orderInfo->getOrderNumber(), $orderInfo->getDescription());

        Log::getInstance()->notice('Changing status for the order "%s".', $orderInfo->getOrderNumber());
        $result = $this->changeOrderStatus($orderInfo->getOrderId(), BillingHelper::ORDER_STATUS_OPEN);
        Log::getInstance()->notice('The result of the status change is "%s".', $result->getStatus());
    }

    /**
     * @param int $appInstanceId
     * @param string $apsInstanceId
     * @return \APS\ControllerProxy
     */
    private function createApsControllerInstance($appInstanceId, $apsInstanceId)
    {
        $token = $this->openApi->getApplicationInstanceToken($appInstanceId);
        return $this->apiFactory->getApsController(
            $token[BillingHelper::PROPERTY_CONTROLLER_URI],
            null,
            $apsInstanceId
        );
    }

    /**
     * @return Service_PbaApi_Api
     */
    private function createPbaApiInstance()
    {
        return $this->apiFactory->getPbaApi(Settings::getInstance()->getBillingConfig());
    }

    /**
     * @return Service_OpenApi_Api
     */
    private function createOpenApiInstance()
    {
        return $this->apiFactory->getOpenApi(OPEN_API_URI);
    }

    /**
     * @return \APS\ResourceBase[]
     */
    private function fetchTenants()
    {
        $filter = RqlBuilder::createInstance()
            ->implementing(\APS\TypeLibrary::getTypeByServiceId('tenant'))
            ->select(array('subscription', 'minimalCommitmentTier'))
            ->eq('aps.status', Service_APSApi_Api::APS_READY)
            ->isFalse('subscription.trial')
            ->isNotNull('minimalCommitmentTier.aps.id')
            ->ge('minimalCommitmentTier.level', 0)
            ->isNotNull('startProductionTime')
            ->getQuery();
        return $this->apsController->getResources($filter);
    }

    /**
     * @param array $subscriptionResources
     * @param string $apsType
     * @param string $apsProperty
     * @return array
     */
    private function filterSubscriptionResources(array $subscriptionResources, $apsType, $apsProperty = null)
    {
        return array_filter($subscriptionResources, function ($resource) use ($apsType, $apsProperty) {
            $typeIsMatched = isset($resource->apsType) && $resource->apsType === $apsType;
            $propertyIsMatched = is_null($apsProperty)
                ? !isset($resource->property)
                : isset($resource->property) && $resource->property === $apsProperty;

            return $typeIsMatched && $propertyIsMatched;
        });
    }

    /**
     * @param array $subscriptionResources
     * @param int | null $limit
     * @return array
     */
    private function filterSubscriptionResourcesByLimitNotEqual(array $subscriptionResources, $limit)
    {
        return array_filter($subscriptionResources, function ($resource) use ($limit) {
            $resourceLimit = isset($resource->limit) ? $resource->limit : BillingHelper::APS_UNLIMITED_VALUE;
            return $resourceLimit !== $limit;
        });
    }

    /**
     * @param array $subscriptionResources
     * @return array
     */
    private function fetchSubscriptionTiers($subscriptionResources)
    {
        $ids = array_map(function ($resource) {
            return $resource->apsId;
        }, $subscriptionResources);
        $filter = RqlBuilder::createInstance()
            ->implementing(\APS\TypeLibrary::getTypeByServiceId('minimalCommitmentTier'))
            ->in('aps.id', $ids)
            ->getQuery();
        $tiers = $this->apsController->getResources($filter);
        return $tiers;
    }

    /**
     * @param \APS\ResourceBase $tier
     * @param string $currency
     * @return int | null
     */
    private function getMinimalCommitment($tier, $currency)
    {
        if (!is_array($tier->minimalCommitments)) {
            return null;
        }
        foreach ($tier->minimalCommitments as $minimalCommitment) {
            if ($minimalCommitment->currency === $currency) {
                return $minimalCommitment->value;
            }
        }
        return null;
    }

    /**
     * @param int $subscriptionId
     * @param Service_PbaApi_Api_Response_Struct_GetSubscriptionSalesOrdersListInterface[] $billingOrders
     * @return float|int
     * @throws Exception
     */
    private function calculateBillingOrderTotalAmount($subscriptionId, $billingOrders)
    {
        $totalAmount = 0;
        $startPeriodTime = $this->startPeriodDate->getTimestamp();
        $endPeriodTime = $this->endPeriodDate->getTimestamp();

        foreach ($billingOrders as $orderInfo) {
            $details = $this->pbaApi->getOrderDetails($orderInfo->getOrderId());
            foreach ($details as $detail) {
                if ($detail->getOrderDetailStartDate() < $startPeriodTime ||
                    $detail->getOrderDetailEndDate() - 1 > $endPeriodTime
                ) {
                    continue;
                }
                if ($detail->getSubscription() == $subscriptionId) {
                    $totalAmount += $detail->getExtendedPrice();
                }
            }
        }
        return $totalAmount;
    }

    /**
     * @param int $subscriptionId
     * @return array
     * @throws Exception
     */
    private function fetchSubscriptionBillingOrders($subscriptionId)
    {
        // sort by order date (-4) is important
        $subscriptionOrdersList = $this->pbaApi->getSubscriptionSalesOrdersList($subscriptionId, -4);

        $startPeriodTime = $this->startPeriodDate->getTimestamp();

        $billingOrders = array();
        foreach ($subscriptionOrdersList as $listItem) {
            $orderTime = $listItem->getOrderDate();
            if ($orderTime < $startPeriodTime) {
                break;
            }
            $pattern = '/^(' . preg_quote(BillingHelper::ORDER_TYPE_BO, '/') . '|' . preg_quote(BillingHelper::ORDER_TYPE_CH, '/') . ')\d+$/';
            if (!preg_match($pattern, $listItem->getOrderNumber())) {
                continue;
            }
            $billingOrders[] = $listItem;
        }
        return $billingOrders;
    }

    /**
     * @param $accountId
     * @param $subscriptionId
     * @param $amount
     * @return Service_PbaApi_Api_Response_Struct_GetOrder
     */
    private function placeBillingOrder($accountId, $subscriptionId, $amount)
    {
        $orderDate = clone $this->endPeriodDate;
        $orderDate->setTime(0, 0, 0);

        $orderAttributes = new Service_PbaApi_Api_Request_Struct_OrderAttributes();
        $orderAttributes->setAccountId($accountId);
        $orderAttributes->setOrderTypeId(BillingHelper::ORDER_TYPE_BO);
        $orderAttributes->setOrderStatusId(BillingHelper::ORDER_STATUS_NEW);
        $orderAttributes->setConsolidationType(0);
        $orderAttributes->setFromIp($this->endpointHost);
        $orderAttributes->setDescription(BillingHelper::BILLING_ORDER_NAME);
        $orderAttributes->setOrderDate($orderDate->getTimestamp());

        $orderDetail = new Service_PbaApi_Api_Request_Struct_OrderDetailAttributes();
        $orderDetail->setExtendedPrice($amount);
        $orderDetail->setTaxAmount(0);
        $orderDetail->setTaxIncludedAmount(0);
        $orderDetail->setTaxAlgorithm(0);
        $orderDetail->setUnitPrice($amount);
        $orderDetail->setServiceQuantity(1);
        $orderDetail->setServiceUnitMeasure('-');
        $orderDetail->setDuration(1);
        $orderDetail->setDurationBillPeriod(1);
        $orderDetail->setDurationBillPeriodType(4);
        $orderDetail->setSubscriptionId($subscriptionId);
        $orderDetail->setDescription(BillingHelper::BILLING_ORDER_NAME);

        $orderId = $this->pbaApi->createManualOrder($orderAttributes, array($orderDetail));
        return $this->pbaApi->getOrder($orderId);
    }

    /**
     * @return string
     */
    private function fetchEndPointHost()
    {
        $ehif = new EndpointHostIpFetcher($this->apsController);
        return $ehif->fetchApsEndpointIpAddressOrFallback();
    }

    /**
     * @param $subscriptionId
     * @param $planId
     * @return Service_PbaApi_Api_Response_Struct_PlacePlanPeriodSwitchOrder
     * @throws Exception
     */
    private function placePlanSwitchOrder($subscriptionId, $planId)
    {
        $subscriptionInfo = $this->pbaApi->getSubscriptionDetailsEx($subscriptionId);
        $periodDuration = $subscriptionInfo->getPeriod();
        $periodType = $subscriptionInfo->getPeriodType();
        $period = $this->findPlanPeriod($planId, $periodDuration, $periodType);
        if (is_null($period)) {
            throw new Exception(Str::format(
                'Plan "%s" does not have any non-trial period with duration "%s" and of type "%s" (1 - Day(s), 2 - Month(s), 3 - Year(s))',
                $planId, $periodDuration, $periodType
            ));
        }
        return $this->pbaApi->placePlanPeriodSwitchOrder($subscriptionId, $planId, $period->getPlanPeriodID(), 20);
    }

    /**
     * @param $planId
     * @param $periodDuration
     * @param $periodType
     * @return null|Service_PbaApi_Api_Response_Struct_PlanPeriodListItem
     */
    private function findPlanPeriod($planId, $periodDuration, $periodType)
    {
        $periods = $this->pbaApi->getPlanPeriodList($planId);
        foreach ($periods as $period) {
            if ($period->getTrial() || !$period->getEnabled()) {
                continue;
            }
            if ($period->getPeriod() == $periodDuration && $period->getPeriodType() == $periodType) {
                return $period;
            }
        }
        return null;
    }

    /**
     * @param $currentMinimalCommitment
     * @param $totalAmount
     * @param $currency
     * @param array $availableTiersResources
     * @return \APS\ResourceBase|null
     * @throws Exception
     */
    private function findNextSuitableTier($currentMinimalCommitment, $totalAmount, $currency, array $availableTiersResources)
    {
        $availableTiers = $this->fetchSubscriptionTiers($availableTiersResources);
        if (empty($availableTiers)) {
            return null;
        }
        $values = array();
        foreach ($availableTiers as $availableTier) {
            if ($availableTier->level < 0) {
                continue;
            }
            $value = $this->getMinimalCommitment($availableTier, $currency);
            if (is_null($value)) {
                $message = Str::format('Minimum commitment tier "%s" does not have a value for currency "%s".', $availableTier->name, $currency);
                Log::getInstance()->error($message);
                throw new Exception($message);
            }
            if ($value <= $currentMinimalCommitment || $value > $totalAmount) {
                continue;
            }
            $values[$value] = $availableTier;
        }
        if (empty($values)) {
            return null;
        }
        $value = max(array_keys($values));
        return $values[$value];
    }

    /**
     * @param $subscriptionId
     * @param $resourceTypeId
     * @return int | null
     */
    private function findAvailablePlanWithResource($subscriptionId, $resourceTypeId)
    {
        $plansList = $this->pbaApi->getSubscriptionAvailablePlanList($subscriptionId);
        foreach ($plansList as $planInfo) {
            $ratesList = $this->pbaApi->getPlanRateList($planInfo->getPlanId());
            foreach ($ratesList as $rateInfo) {
                if ($rateInfo->getResourceId() == $resourceTypeId && $rateInfo->getIncludedValue() != 0) {
                    return $planInfo->getPlanId();
                }
            }
        }
        return null;
    }

    /**
     * @param int $orderId
     * @param string $status
     * @return Service_PbaApi_Api_Response_Struct_Status
     */
    private function changeOrderStatus($orderId, $status)
    {
        $order = $this->pbaApi->getOrder($orderId);
        $result = $this->pbaApi->changeOrderStatus(
            $order->getOrderID(),
            $status,
            md5(
                $order->getOrderID() .
                trim($order->getOrderNumber()) .
                $order->getCreationTime() .
                sprintf('%s %.2f', trim($order->getCurrencyID()), $order->getTotal()) .
                $order->getComments()
            )
        );
        return $result;
    }

    /**
     * @return array
     */
    private function fetchApsApplicationInstances()
    {
        $token = $this->openApi->getAccountToken(BillingHelper::PROVIDER_ACCOUNT_ID, BillingHelper::PROVIDER_SUBSCRIPTION_ID);

        $url = $token[BillingHelper::PROPERTY_CONTROLLER_URI] . BillingHelper::APS_APPLICATIONS_PATH;
        $headers = array(BillingHelper::APS_TOKEN_HEADER . $token[BillingHelper::PROPERTY_APS_TOKEN]);

        $applications = json_decode($this->doHttpGetRequest($url, $headers));

        return array_filter($applications, function ($application) {
            return $application->aps->type === APSResource::PACKAGE_ID;
        });
    }

    /**
     * @param string $url
     * @param array $headers
     * @return mixed
     * @throws Exception
     */
    private function doHttpGetRequest($url, array $headers = array())
    {
        Log::getInstance()->event('Request GET %s', $url);
        $curl = new Curl();
        $curl->setOption(CURLOPT_URL, $url);
        $curl->setOption(CURLOPT_HTTPHEADER, $headers);
        $curl->setOption(CURLOPT_RETURNTRANSFER, true);
        $curl->setOption(CURLOPT_FOLLOWLOCATION, true);
        $curl->setOption(CURLOPT_SSL_VERIFYPEER, false);
        $curl->setOption(CURLOPT_SSL_VERIFYHOST, false);
        $return = $curl->execute();
        $errorNo = $curl->getErrorNo();
        if ($errorNo) {
            $curl->refresh();
            throw new Exception($curl->getError(), $errorNo);
        }
        $code = $curl->getInfo(CURLINFO_HTTP_CODE);
        Log::getInstance()->event('Response %s %s', $code, $return);
        if ($code < 200 || $code >= 300) {
            $curl->refresh();
            throw new Exception($return, $code);
        }
        $curl->refresh();
        return $return;
    }

    /**
     * @param $instances
     * @param $resourceId
     * @return array
     */
    private function findApsInstanceIdByResourceId($instances, $resourceId)
    {
        $instances = array_filter($instances, function ($application) use ($resourceId) {
            return $application->globals->aps->id === $resourceId;
        });
        $instance = reset($instances);
        return $instance->aps->id;
    }

    /**
     * @param $accountId
     * @return Service_PbaApi_Api_Response_Struct_Account
     */
    private function fetchAccountInfo($accountId)
    {
        if (!isset($this->accountsInfo[$accountId])) {
            $accountInfo = $this->pbaApi->getAccount($accountId);
            $this->accountsInfo[$accountId] = $accountInfo;
        }

        return $this->accountsInfo[$accountId];
    }

    /**
     * @param $accountInfo
     * @param $tenant
     * @return bool
     */
    private function needToProcessSubscription(Service_PbaApi_Api_Response_Struct_Account $accountInfo, $tenant)
    {
        if ($tenant->startProductionTime > $this->startPeriodDate->getTimestamp()) {
            return false;
        }
        return ($accountInfo->getLastBillingDate() >= $this->endPeriodDate->getTimestamp());
    }

    /**
     * @param $accountInfo
     */
    private function initPeriodByAccountInfo(Service_PbaApi_Api_Response_Struct_Account $accountInfo)
    {
        $statementDay = $accountInfo->getStatementDay();

        $currentDate = new DateTime();
        $currentDay = (int)$currentDate->format('d');

        $this->startPeriodDate = clone $currentDate;
        $this->endPeriodDate = clone $currentDate;

        list($year, $month) = explode('-', $this->startPeriodDate->format('Y-m'));
        $this->startPeriodDate->setDate($year, $month, 1)->setTime(0, 0, 0)->modify('-1 month');
        $this->endPeriodDate->setDate($year, $month, 1)->setTime(23, 59, 59);
        if ($currentDay < $statementDay) {
            $this->startPeriodDate->modify('-1 month');
            $this->endPeriodDate->modify('-1 month');
        }

        //thank you, php
        list($y, $m) = explode('-', $this->startPeriodDate->format('Y-m'));
        if (checkdate($m, $statementDay, $y)) {
            $this->startPeriodDate->setDate($y, $m, $statementDay);
        } else {
            $this->startPeriodDate->modify('last day of this month');
        }

        list($y, $m) = explode('-', $this->endPeriodDate->format('Y-m'));
        if (checkdate($m, $statementDay, $y)) {
            $this->endPeriodDate->setDate($y, $m, $statementDay);
        } else {
            $this->endPeriodDate->modify('last day of this month');
        }
        $this->endPeriodDate->modify('-1 day');
    }

    /**
     * @param $accountId
     * @param Service_PbaApi_Api_Request_Struct_OrderDetailAttributes[] $orderDetails
     * @return Service_PbaApi_Api_Response_Struct_GetOrder
     */
    private function placeAdjustingOrderForAccount($accountId, $orderDetails)
    {
        Log::getInstance()->notice('Placing an adjustment order for the account %s.', $accountId);
        $this->printOrderDetails($orderDetails);

        $orderAttributes = new Service_PbaApi_Api_Request_Struct_OrderAttributes();
        $orderAttributes->setAccountId($accountId);
        $orderAttributes->setOrderTypeId(BillingHelper::ORDER_TYPE_BO);
        $orderAttributes->setOrderStatusId(BillingHelper::ORDER_STATUS_NEW);
        $orderAttributes->setConsolidationType(10);
        $orderAttributes->setFromIp($this->endpointHost);
        $orderAttributes->setDescription('Correcting order for Acronis Volume Pricing Program');
        $orderAttributes->setOrderDate(time());

        $orderId = $this->pbaApi->createManualOrder($orderAttributes, $orderDetails);
        return $this->pbaApi->getOrder($orderId);
    }

    /**
     * @param $accountId
     * @return int | false
     */
    private function getDiscountIdForAccount($accountId)
    {
        $discountRelations = $this->globals->discountRelations;
        $discountRelation = reset(array_filter($discountRelations, function ($discountRelation) use ($accountId) {
            return $discountRelation->accountId == $accountId;
        }));
        if (!$discountRelation) {
            return false;
        }
        return $discountRelation->discountId;
    }

    /**
     * @param $discountId
     * @return Service_PbaApi_Api_Response_Struct_DiscountedPlanListItem[]
     */
    private function getDiscountedPlans($discountId)
    {
        $discountedPlanList = $this->pbaApi->getDiscountedPlanList($discountId);
        return array_map(function ($discountedPlan) {
            return $discountedPlan->getPlanId();
        }, $discountedPlanList);
    }

    /**
     * @param $discountId
     * @return Service_PbaApi_Api_Response_Struct_DiscountedResourceListItem[]
     */
    private function getDiscountedResources($discountId)
    {
        $discountedResourceList = $this->pbaApi->getDiscountedResourceList($discountId);
        return array_map(function ($discountedResource) {
            return $discountedResource->getResourceId();
        }, $discountedResourceList);
    }

    /**
     * @param $accountId
     * @param $switchTime
     * @return Service_PbaApi_Api_Response_Struct_DeliverLine[]
     * @throws Exception
     */
    private function getDeliverLinesForDate($accountId, $switchTime)
    {
        $switchDate = new DateTime();
        $switchDate->setTimestamp($switchTime);
        $switchDate->setTime(0, 0, 0);

        $startPeriodDate = clone $switchDate;
        $startPeriodDate->modify('-31 day');

        $endPeriodDate = clone $switchDate;
        $endPeriodDate->modify('+31 day');

        $startPeriodTime = $startPeriodDate->getTimestamp();
        $endPeriodTime = $endPeriodDate->getTimestamp();
        $switchTime = $switchDate->getTimestamp();

        $this->pbaApi->startTransaction();
        try {
            Log::getInstance()->event(str_repeat('-', 80));
            Log::getInstance()->event(Str::format(
                "Getting all Deliver Lines for the period from %s to %s by account %s with split in %s.",
                $startPeriodDate->format('Y-m-d'),
                $endPeriodDate->format('Y-m-d'),
                $accountId,
                $switchDate->format('Y-m-d')
            ));

            $this->pbaApi->splitDeliverLinesForVendorByDate($switchTime, $accountId);
            $deliverLines = $this->pbaApi->getDeliverLinesForPeriodByVendorList($startPeriodTime, $endPeriodTime, $accountId);

            $this->printDeliverLines($deliverLines, 'event');
            Log::getInstance()->event(str_repeat('-', 80));

            Log::getInstance()->event(str_repeat('-', 80));
            Log::getInstance()->event(Str::format(
                "Getting all suitable Deliver Lines for the period from %s to %s by account %s with split in %s.",
                $startPeriodDate->format('Y-m-d'),
                $endPeriodDate->format('Y-m-d'),
                $accountId,
                $switchDate->format('Y-m-d')
            ));

            $deliverLines = array_filter($deliverLines, function ($deliverLine) use ($switchTime) {
                if ($deliverLine->getDeliverLineStatus() != 1) {
                    return false;
                }

                $detailType = $deliverLine->getDetailTypeId();

                $isOverusage = in_array($detailType, array(
                        BillingHelper::DETAIL_TYPE_RESOURCE_OVERUSAGE
                    )) && $deliverLine->getDeliverLineEndDate() == $switchTime;

                $isRecurring = in_array($detailType, array(
                        BillingHelper::DETAIL_TYPE_PLAN_RECURRING,
                        BillingHelper::DETAIL_TYPE_RESOURCE_RECURRING,
                    )) && $deliverLine->getDeliverLineStartDate() == $switchTime;

                return $isOverusage || $isRecurring;
            });
            $this->printDeliverLines($deliverLines, 'event');
            Log::getInstance()->event(str_repeat('-', 80));

        } catch (Exception $e) {
            $this->pbaApi->rollbackTransaction();
            throw $e;
        }
        // only rollback to not affect deliver lines!
        $this->pbaApi->rollbackTransaction();
        return $deliverLines;
    }

    /**
     * @param $detailId
     * @return Service_PbaApi_Api_Response_Struct_GetOrderDetails
     */
    private function fetchOrderDetail($detailId)
    {
        $detail = $this->pbaApi->getOrderDetail($detailId);
        $details = $this->pbaApi->getOrderDetails($detail->getOrderId());
        return reset(array_filter($details, function ($detail) use ($detailId) {
            return $detail->getDetailID() == $detailId;
        }));
    }

    /**
     * @param int $vendorAccountId
     * @param Service_PbaApi_Api_Response_Struct_DeliverLine[] $deliverLines
     * @param string $type
     * @param array $subscriptions
     * @return Service_PbaApi_Api_Request_Struct_OrderDetailAttributes[]
     * @throws Exception
     */
    private function makeOrderDetailsFromDeliverLines($vendorAccountId, $deliverLines, $type, array $subscriptions)
    {
        $discountId = $this->getDiscountIdForAccount($vendorAccountId);
        if (!$discountId) {
            throw new Exception(Str::format('There are no discounts for the vendor %s.', $vendorAccountId));
        }

        $vendorDiscountedPlans = $this->getDiscountedPlans($discountId);
        $vendorDiscountedResources = $this->getDiscountedResources($discountId);
        if (empty($vendorDiscountedPlans) && empty($vendorDiscountedResources)) {
            throw new Exception(Str::format('There are neither discounted plans nor discounted resources in the discount %s.', $discountId));
        }

        $suitableDeliverLines = array();
        $orderDetails = array();
        foreach ($deliverLines as $deliverLine) {
            if (!in_array($deliverLine->getSubscriptionId(), $subscriptions)) {
                Log::getInstance()->event(Str::format(
                    "Skipping the deliver line %s. The subscription %s does not need to be corrected.",
                    $deliverLine->getDeliverLineId(),
                    $deliverLine->getSubscriptionId()
                ));
                continue;
            }

            if ($deliverLine->getDeliverLineStatus() != 1) {
                Log::getInstance()->event(Str::format(
                    "Skipping the deliver line %s. The line is disabled.",
                    $deliverLine->getDeliverLineId()
                ));
                continue;
            }

            $detailId = $deliverLine->getDetailId();
            if (!$detailId) {
                Log::getInstance()->event(Str::format(
                    "Skipping the deliver line %s. The line is merged.",
                    $deliverLine->getDeliverLineId()
                ));
                continue;
            }

            $vendorPlanId = $this->getVendorPlanId($deliverLine->getPlanId());
            if (!$vendorPlanId) {
                throw new Exception(Str::format('Cannot get the vendor plan by child %s.', $deliverLine->getPlanId()));
            }
            Log::getInstance()->event(Str::format(
                "The parent plan for %s is %s.",
                $deliverLine->getPlanId(),
                $vendorPlanId
            ));

            $resourceId = $deliverLine->getResourceId();

            if ($resourceId) {
                if (!in_array($vendorPlanId, $vendorDiscountedPlans) && !in_array($resourceId, $vendorDiscountedResources)) {
                    Log::getInstance()->event(Str::format(
                        "Skipping the deliver line %s. Neither vendor plan %s nor resource %s is not in the discount %s.",
                        $deliverLine->getDeliverLineId(),
                        $vendorPlanId,
                        $resourceId,
                        $discountId
                    ));
                    continue;
                }
            } else {
                if (!in_array($vendorPlanId, $vendorDiscountedPlans)) {
                    Log::getInstance()->event(Str::format(
                        "Skipping the deliver line %s. The vendor plan %s is not in the discount %s.",
                        $deliverLine->getDeliverLineId(),
                        $vendorPlanId,
                        $discountId
                    ));
                    continue;
                }
            }

            $detail = $this->fetchOrderDetail($detailId);
            if (!$detail) {
                throw new Exception(Str::format('Cannot fetch detail %s.', $detailId));
            }

            $detailTypeId = $deliverLine->getDetailTypeId();
            if ($detailTypeId == BillingHelper::DETAIL_TYPE_PLAN_RECURRING) {
                $period = $this->fetchPlanPeriod($detail->getPlanPeriodId());
                $vendorPlanPeriod = $this->fetchPlanPeriodByTypeAndDuration($vendorPlanId, $period->getPeriodType(), $period->getPeriod());
                if (!$vendorPlanPeriod) {
                    throw new Exception(Str::format(
                        'Cannot fetch plan period for the vendor plan %s by type %s and period %s.',
                        $vendorPlanId,
                        $period->getPeriodType(),
                        $period->getPeriod()
                    ));
                }

                $unitPrice = $vendorPlanPeriod->getSubscriptionFee();
            } elseif ($detailTypeId == BillingHelper::DETAIL_TYPE_RESOURCE_RECURRING || $detailTypeId == BillingHelper::DETAIL_TYPE_RESOURCE_OVERUSAGE) {
                $vendorPlanRate = $this->fetchPlanRateByResourceId($vendorPlanId, $detail->getResourceID());
                if (!$vendorPlanRate) {
                    throw new Exception(Str::format(
                        'Cannot fetch plan rate for the vendor plan by resource %s.',
                        $vendorPlanId,
                        $detail->getResourceID()
                    ));
                }

                $unitPrice = $detailTypeId == BillingHelper::DETAIL_TYPE_RESOURCE_RECURRING ?
                    $vendorPlanRate->getRecurringFee() :
                    $vendorPlanRate->getOveruseFee();

            } else {
                throw new Exception(Str::format('Unexpected detail type %s.', $detailTypeId));
            }

            $debit = ($detailTypeId == BillingHelper::DETAIL_TYPE_RESOURCE_OVERUSAGE);
            if ($type !== BillingHelper::ACCOUNT_PROCESS_TYPE_ONBOARDING) {
                $debit = !$debit;
            }

            $description = $this->makeOrderDetailDescription($type, $deliverLine);
            $serviceQuantity = $deliverLine->getExtendedPrice() / ($deliverLine->getUnitPrice() * $deliverLine->getPeriodInMonths());

            $orderDetails[] = $this->createOrderDetailAttributes(
                $description,
                $deliverLine->getPeriodInMonths(),
                $deliverLine->getBillingPeriod(),
                $deliverLine->getBillingPeriodType(),
                $serviceQuantity,
                $detail->getUom(),
                $unitPrice,
                $debit
            );

            $suitableDeliverLines[] = $deliverLine;
        }

        Log::getInstance()->notice(str_repeat('-', 80));
        Log::getInstance()->notice(Str::format("Suitable deliver lines for a correcting order for the account %s.", $vendorAccountId));
        $this->printDeliverLines($suitableDeliverLines);
        Log::getInstance()->notice(str_repeat('-', 80));

        return $orderDetails;
    }

    /**
     * @param $planId
     * @return int|null
     * @throws Exception
     */
    private function getVendorPlanId($planId)
    {
        $this->pbaApi->startTransaction();
        try {
            $publicationStatus = $this->pbaApi->getPlanPublicationStatusForUpdate($planId);
            $showPriority = $publicationStatus->getShowPriority();
            $showPriority = $showPriority < 0 ? 0 : $showPriority;
            $this->pbaApi->updatePlanPublicationStatus(
                $planId,
                $publicationStatus->getPublished(),
                $showPriority + 1,
                $publicationStatus->getDefaultPlanPeriodId()
            );
            $this->pbaApi->updatePlanPublicationStatus(
                $planId,
                $publicationStatus->getPublished(),
                $showPriority,
                $publicationStatus->getDefaultPlanPeriodId()
            );
            $historyRecord = reset($this->pbaApi->getPlanHistoryList($planId, -3));  // -3 - sort by date DESC
            $plan = $this->pbaApi->getPlanFromArchive($planId, $historyRecord->getUserId(), $historyRecord->getDate());
        } catch (Exception $e) {
            $this->pbaApi->rollbackTransaction();
            throw $e;
        }
        // only rollback to not affect plans!
        $this->pbaApi->rollbackTransaction();
        return $plan->getCopiedFromPlanId();
    }

    /**
     * @param string $description
     * @param float $duration
     * @param int $period
     * @param int $periodType
     * @param float $quantity
     * @param string $uom
     * @param float $price
     * @param bool $debit
     * @return Service_PbaApi_Api_Request_Struct_OrderDetailAttributes
     */
    private function createOrderDetailAttributes(
        $description, $duration, $period, $periodType,
        $quantity, $uom, $price, $debit = true
    )
    {
        $detail = new Service_PbaApi_Api_Request_Struct_OrderDetailAttributes();

        $detail->setDescription($description);

        $detail->setDuration($duration);
        $detail->setDurationBillPeriod($period);
        $detail->setDurationBillPeriodType($periodType);

        $detail->setTaxAlgorithm(0);
        $detail->setTaxAmount(0.0);
        $detail->setTaxIncludedAmount(0.0);

        $detail->setDiscountAmount(0.0);

        $detail->setServiceQuantity($quantity);
        $detail->setServiceUnitMeasure($uom);

        $detail->setUnitPrice($price);

        $extendedPrice = $quantity * $duration * $price;

        $detail->setExtendedPrice($debit ? $extendedPrice : -$extendedPrice);
        return $detail;
    }

    /**
     * @param $planId
     * @param $periodType
     * @param $periodDuration
     * @return false|Service_PbaApi_Api_Response_Struct_PlanPeriodListItem
     */
    private function fetchPlanPeriodByTypeAndDuration($planId, $periodType, $periodDuration)
    {
        $periods = $this->pbaApi->getPlanPeriodList($planId);
        return reset(array_filter($periods, function ($period) use ($periodType, $periodDuration) {
            /**
             * @var Service_PbaApi_Api_Response_Struct_PlanPeriodListItem $period
             */
            return $period->getPeriod() == $periodDuration && $period->getPeriodType() == $periodType;
        }));
    }

    /**
     * @param $planId
     * @param $resourceId
     * @return false|Service_PbaApi_Api_Response_Struct_PlanRate
     */
    private function fetchPlanRateByResourceId($planId, $resourceId)
    {
        $rates = $this->pbaApi->getPlanRates($planId);
        return reset(array_filter($rates, function ($rate) use ($resourceId) {
            return $rate->getResourceId() == $resourceId;
        }));
    }

    /**
     * @param $periodId
     * @return Service_PbaApi_Api_Response_Struct_PlanPeriod
     */
    private function fetchPlanPeriod($periodId)
    {
        return $this->pbaApi->getPlanPeriod($periodId);
    }

    /**
     * @param $type
     * @param Service_PbaApi_Api_Response_Struct_DeliverLine $deliverLine
     * @return string|null
     */
    private function makeOrderDetailDescription($type, $deliverLine)
    {
        $detailType = $deliverLine->getDetailTypeId();

        $subscriptionId = $deliverLine->getSubscriptionId();
        if ($subscriptionId) {
            $subscriptionDetails = $this->pbaApi->getSubscriptionDetails($subscriptionId);
            $subscriptionName = $subscriptionDetails->getSubscriptionName();
        } else {
            $subscriptionName = '';
        }

        $resourceId = $deliverLine->getResourceId();
        if ($resourceId) {
            $resourceDetails = $this->pbaApi->getResource($resourceId);
            $resourceName = $resourceDetails->getName();
        } else {
            $resourceName = '';
        }

        $startDate = new DateTime();
        $startDate->setTimestamp($deliverLine->getDeliverLineStartDate());
        $startDateString = $startDate->format('d-M-Y');

        $endTime = new DateTime();
        $endTime->setTimestamp($deliverLine->getDeliverLineEndDate());
        $endDateString = $endTime->format('d-M-Y');

        $description = '';
        if ($type === BillingHelper::ACCOUNT_PROCESS_TYPE_ONBOARDING) {
            if ($detailType == BillingHelper::DETAIL_TYPE_PLAN_RECURRING) {
                $description = 'Return for {subscriptionName}, Subscription ID {subscriptionId}, Plan Recurring, from {startDate} through {endDate}.';
            }
            if ($detailType == BillingHelper::DETAIL_TYPE_RESOURCE_RECURRING) {
                $description = 'Return for {resourceName}, Subscription ID {subscriptionId}, Resource ID {resourceId}, Resource Recurring from {startDate} through {endDate}.';
            }
            if ($detailType == BillingHelper::DETAIL_TYPE_RESOURCE_OVERUSAGE) {
                $description = '{resourceName}, Subscription ID {subscriptionId}, Resource ID {resourceId}, Resource Overusage Addition from {startDate} through {endDate}.';
            }
        } elseif ($type === BillingHelper::ACCOUNT_PROCESS_TYPE_CANCELLATION) {
            if ($detailType == BillingHelper::DETAIL_TYPE_PLAN_RECURRING) {
                $description = '{subscriptionName}, Subscription ID {subscriptionId}, Plan Recurring Addition, from {startDate} through {endDate}.';
            }
            if ($detailType == BillingHelper::DETAIL_TYPE_RESOURCE_RECURRING) {
                $description = '{resourceName}, Subscription ID {subscriptionId}, Resource ID {resourceId}, Resource Recurring Addition from {startDate} through {endDate}.';
            }
            if ($detailType == BillingHelper::DETAIL_TYPE_RESOURCE_OVERUSAGE) {
                $description = 'Return for {resourceName}, Subscription ID {subscriptionId}, Resource ID {resourceId}, Resource Overusage from {startDate} through {endDate}.';
            }
        }

        return Str::format($description, array(
            '{subscriptionId}' => $subscriptionId,
            '{subscriptionName}' => $subscriptionName,
            '{resourceId}' => $resourceId,
            '{resourceName}' => $resourceName,
            '{startDate}' => $startDateString,
            '{endDate}' => $endDateString,
        ));
    }

    /**
     * @param Service_PbaApi_Api_Response_Struct_DeliverLine[] $deliverLines
     * @param string $level
     */
    private function printDeliverLines($deliverLines, $level = 'notice')
    {
        /**
         * 1000017 | 01 Jun #################### 21 Jun | Acronis Backup Service Recurring
         * 1000018 |       07 Jun ############## 21 Jun | Acronis Backup Service Recurring
         * 1000019 |              14 Jun ####### 21 Jun | Acronis Backup Service Recurring
         * 1000017 | 01 Jun #################### 21 Jun | Cloud Storage (Acronis) Overusage
         * 1000018 |       07 Jun ############## 21 Jun | Cloud Storage (Acronis) Overusage
         * 1000019 |              14 Jun ####### 21 Jun | Cloud Storage (Acronis) Overusage
         */

        if (empty($deliverLines)) {
            return;
        }

        $minDate = min(array_map(function ($deliverLine) {
            return BillingHelper::toDays($deliverLine->getDeliverLineStartDate());
        }, $deliverLines));
        $maxDate = max(array_map(function ($deliverLine) {
            return BillingHelper::toDays($deliverLine->getDeliverLineEndDate());
        }, $deliverLines));


        uasort($deliverLines, function ($deliverLine1, $deliverLine2) {
            $r = $deliverLine1->getSubscriptionId() - $deliverLine2->getSubscriptionId();
            if ($r !== 0) {
                return $r;
            }
            $r = $deliverLine1->getDetailTypeId() - $deliverLine2->getDetailTypeId();
            if ($r !== 0) {
                return $r;
            }
            $r = $deliverLine1->getDeliverLineStartDate() - $deliverLine2->getDeliverLineStartDate();
            if ($r !== 0) {
                return $r;
            }
            return $deliverLine1->getDeliverLineEndDate() - $deliverLine2->getDeliverLineEndDate();
        });

        foreach ($deliverLines as $deliverLine) {
            $lpad = BillingHelper::toDays($deliverLine->getDeliverLineStartDate()) - $minDate;
            $rpad = $maxDate - BillingHelper::toDays($deliverLine->getDeliverLineEndDate());
            $length = BillingHelper::toDays($deliverLine->getDeliverLineEndDate()) - BillingHelper::toDays($deliverLine->getDeliverLineStartDate());
            $marker = $deliverLine->getDeliverLineStatus() == 1 ? "#" : "-";

            $line = $deliverLine->getSubscriptionId();
            $line .= ' | ';
            $line .= $lpad > 0 ? str_repeat(" ", $lpad) : '';
            $line .= date("d M", $deliverLine->getDeliverLineStartDate()) . ' ';
            $line .= str_repeat($marker, $length);
            $line .= ' ' . date("d M", $deliverLine->getDeliverLineEndDate());
            $line .= $rpad > 0 ? str_repeat(" ", $rpad) : '';
            $line .= ' | ';
            $line .= $deliverLine->getDeliverLineId() . "\t";
            $line .= $deliverLine->getDescription();

            Log::getInstance()->{$level}($line);
        }
    }

    /**
     * @param Service_PbaApi_Api_Request_Struct_OrderDetailAttributes[] $orderDetails
     */
    private function printOrderDetails($orderDetails)
    {
        if (empty($orderDetails)) {
            return;
        }
        foreach ($orderDetails as $orderDetail) {
            $line = "Quantity: " . $orderDetail->getServiceQuantity() . ' ' . $orderDetail->getServiceUnitMeasure() . "\t";
            $line .= "Duration: " . $orderDetail->getDuration() . " month(s)\t";
            $line .= "Price: " . $orderDetail->getUnitPrice() . "\t";
            $line .= "Total: " . $orderDetail->getExtendedPrice() . "\t";
            $line .= $orderDetail->getDescription();
            Log::getInstance()->notice($line);
        }
    }

    public static function toDays($time)
    {
        return ceil($time / 60 / 60 / 24);
    }
}
