<?php
/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

require_once(__DIR__ . '/app/bootstrap.php');

/**
 * @type("http://acronis.com/aps/cloud-backup-v1/datacenter/4.0")
 * @implements("http://aps-standard.org/types/core/resource/1.0")
 */
class datacenter extends APSResource
{
    const RQL_NE_APS_ID_AND_EQ_ID = 'ne(aps.id,%s),eq(id,%s),limit(0,1)';
    const RQL_EQ_DATACENTER_ID = 'eq(datacenterId,%s),limit(0,1)';

    const REGEX_INTEGER = '/^[1-9]\d*$/';

    const EXCEPTION_BAD_REQUEST_CODE = 400;
    const EXCEPTION_CONFLICT_CODE = 409;
    const EXCEPTION_LOCKED_CODE = 423;

    const EXCEPTION_ID_IS_ALREADY_IN_USE_MESSAGE = "ID '%s' is already in use. Please specify another ID.";
    const EXCEPTION_INVALID_FORMAT_OF_ID_MESSAGE = "The ID format is invalid.";
    const EXCEPTION_EMPTY_CREDENTIALS_MESSAGE = "Please specify all connection settings for the datacenter.";
    const EXCEPTION_CANNOT_REMOVE_DATACENTER_MESSAGE = "Cannot remove datacenter '%s - %s' because there are subscriptions that are provisioned in it.";
    const EXCEPTION_CANNOT_REMOVE_DATACENTER_RESELLER_MESSAGE = "Cannot remove datacenter '%s - %s' because there are resellers that are provisioned in it.";
    const EXCEPTION_CANNOT_RECEIVE_STORAGES_MESSAGE = 'Cannot receive storages';
    const EXCEPTION_CANNOT_CHANGE_CREDENTIALS_MESSAGE = 'There is no such account in this group. Please specify another account.';
    const EXCEPTION_CANNOT_REMOVE_DATACENTER_STORAGE_MESSAGE = 'Cannot remove datacenter \'%s - %s\' because its storage is already in use';
    const EXCEPTION_CANNOT_REMOVE_DATACENTER_PROFILE_MESSAGE = 'Cannot remove datacenter \'%s - %s\' because it is already in use';

    const RQL_IMPLEMENTING_TENANT_EQ_DATACENTER_ID = 'implementing(%s),eq(datacenterId,%s)';

    const PARAMETER_SUBSCRIPTION_COUNT = 'subscriptionCount';

    const TYPE = "type";
    const SUCCESS = "success";
    const DESCRIPTION = "description";
    const ERROR = "error";
    const HOST = "host";
    const SERVER_NAME = 'SERVER_NAME';

    const LOG_CALLED_UPGRADE = 'Datacenter "%s" upgrade was called.';

    /**
     * @link("http://acronis.com/aps/cloud-backup-v1/globals/9.0")
     * @required
     */
    public $global;

    /**
     * @type("string")
     */
    public $id;

    /**
     * @type("string")
     */
    public $name;

    /**
     * @type("string")
     */
    public $url;

    /**
     * @type("string")
     */
    public $login;

    /**
     * @type("string")
     * @encrypted
     */
    public $password;

    public function provision()
    {
        Log::getInstance()->event(self::LOG_CALLED_METHOD_WAS_CALLED, __METHOD__);

        if (empty($this->id)) {
            $this->id = $this->getNextId();
        }

        $this->_checkFormatId($this->id);
        $this->_checkUniqueId($this->id);
        $this->_checkCredentials($this);
        $this->updatePackageVersionInfo();
    }

    /**
     * @param $resource
     * @throws Exception
     */
    public function configure($resource)
    {
        Log::getInstance()->event(self::LOG_CALLED_METHOD_WAS_CALLED, __METHOD__);

        if ($this->id !== $resource->id) {
            $this->_checkFormatId($resource->id);
            $this->_checkUniqueId($resource->id);
        }

        if ($this->url !== $resource->url || $this->login !== $resource->login || $this->password !== $resource->password) {
            try {
                $this->_checkLinkedTenants();
            } catch (Exception $e) {
                if ($e->getCode() != self::EXCEPTION_LOCKED_CODE) {
                    throw $e;
                }
                $this->checkAvailabilityChangeCredentials($resource);
            }

            $this->_checkCredentials($resource);
            $this->updatePackageVersionInfo();
        }

        parent::configure($resource);
    }

    public function unprovision()
    {
        Log::getInstance()->event(self::LOG_CALLED_METHOD_WAS_CALLED, __METHOD__);

        $this->checkStorageUsage();
        $this->_checkLinkedTenants();
        $this->_checkLinkedResellers();
    }

    public function upgrade()
    {
        Log::getInstance()->event(self::LOG_CALLED_UPGRADE, $this->aps->id);

        if ($this->aps->type === self::CLOUD_BACKUP_DATACENTER_TYPE) {
            return;
        }

        $this->aps->type = self::CLOUD_BACKUP_DATACENTER_TYPE;
        if (isset($this->timezone)) {
            unset($this->timezone);
        }
        $this->update();
    }

    /**
     * @verb(POST)
     * @path("/check_connection")
     * @param(string, query)
     * @param(object, body)
     */
    public function checkConnection($type, $credentials)
    {
        Log::getInstance()->event(self::LOG_CALLED_METHOD_WAS_CALLED, __METHOD__);

        if (empty($credentials->login)) {
            $credentials->login = $this->login;
        }
        if (empty($credentials->password)) {
            $credentials->password = $this->password;
        }
        if (empty($credentials->url)) {
            $credentials->url = $this->url;
        }

        $publicApiHost = defined('OPEN_API_URI') ? OPEN_API_URI : parse_url(\APS\Proto::getControllerURI(), PHP_URL_HOST);
        $connectionChecker = new Model_ConnectionChecker($credentials->login, $credentials->password, $credentials->url, $publicApiHost);
        $checker = $connectionChecker->getCheckerByType($type);
        $success = $checker->check();
        return array(
            self::TYPE => $type,
            self::SUCCESS => $success,
            self::DESCRIPTION => $checker->getCheckDescription(),
            self::ERROR => $checker->getCheckError(),
            self::HOST => $_SERVER[self::SERVER_NAME]
        );
    }

    /**
     * @verb(GET)
     * @path("/storages")
     * @return(array)
     */
    public function getStorages()
    {
        Log::getInstance()->event(self::LOG_CALLED_METHOD_WAS_CALLED, __METHOD__);
        $response = array();
        $storages = $this->getMspApi()->getStorages();
        if (is_null($storages)) {
            throw new Exception(self::EXCEPTION_CANNOT_RECEIVE_STORAGES_MESSAGE);
        }

        foreach ($storages as $storage) {
            /** @var Service_MspApiClient_Response_Struct_Storage $storage */
            $response[] = (object)array(
                'id' => (string)$storage->getId(),
                'name' => $storage->getName(),
                'inherited' => $storage->getInherited(),
                'uid' => $storage->getUid(),
                'spaceUsage' => $storage->getSpaceUsage(),
            );
        }
        return $response;
    }

    /**
     * @verb(GET)
     * @path("/info")
     * @return(array)
     */
    public function getInfo()
    {
        Log::getInstance()->event(self::LOG_CALLED_METHOD_WAS_CALLED, __METHOD__);
        $response = array();

        $response[self::PARAMETER_SUBSCRIPTION_COUNT] = $this->_getApsApi()->countResources(
            Str::format(
                self::RQL_IMPLEMENTING_TENANT_EQ_DATACENTER_ID,
                \APS\TypeLibrary::getTypeByServiceId('tenant'),
                $this->id
            )
        );

        return $response;
    }

    /**
     * @return PartnerCredentials
     */
    public function getPartnerCredentials()
    {
        Log::getInstance()->event(self::LOG_CALLED_METHOD_WAS_CALLED, __METHOD__);
        $partnerCredentials = new PartnerCredentials();
        $partnerCredentials->baasUrl = $this->url;
        $partnerCredentials->partnerLogin = $this->login;
        $partnerCredentials->partnerPassword = $this->password;
        return $partnerCredentials;
    }

    /**
     * @param string $id
     * @throws \Rest\RestException
     */
    private function _checkFormatId($id)
    {
        if (!preg_match(self::REGEX_INTEGER, $id)) {
            throw new \Rest\RestException(
                self::EXCEPTION_BAD_REQUEST_CODE,
                self::EXCEPTION_INVALID_FORMAT_OF_ID_MESSAGE
            );
        }
    }

    /**
     * @param string $id
     * @throws \Rest\RestException
     */
    private function _checkUniqueId($id)
    {
        $link = $this->global->datacenters->getLink();
        $datacenters = $this->_apsc->getResources(Str::format(self::RQL_NE_APS_ID_AND_EQ_ID, $this->aps->id, $id), $link->href);
        if (count($datacenters)) {
            throw new \Rest\RestException(
                self::EXCEPTION_CONFLICT_CODE,
                Str::format(self::EXCEPTION_ID_IS_ALREADY_IN_USE_MESSAGE, $id)
            );
        }
    }

    /**
     * @param object $resource
     * @throws \Rest\RestException
     */
    private function _checkCredentials($resource)
    {
        if (empty($resource->url) || empty($resource->login) || empty($resource->password)) {
            throw new \Rest\RestException(
                self::EXCEPTION_BAD_REQUEST_CODE,
                self::EXCEPTION_EMPTY_CREDENTIALS_MESSAGE
            );
        }

        $mspApi = $this->_facade->getMspApi($resource->url);
        $mspApi->setCredentials($resource->login, $resource->password, $resource->url, true);
        $mspApi->login();
    }

    private function _checkLinkedTenants()
    {
        $link = $this->global->tenant->getLink();
        $tenants = $this->_apsc->getResources(Str::format(self::RQL_EQ_DATACENTER_ID, $this->id), $link->href);
        if (count($tenants)) {
            throw new \Rest\RestException(
                self::EXCEPTION_LOCKED_CODE,
                Str::format(self::EXCEPTION_CANNOT_REMOVE_DATACENTER_MESSAGE, $this->id, $this->name)
            );
        }
    }

    private function getNextId()
    {
        $datacenters = $this->global->datacenters;
        $maxId = 0;
        foreach ($datacenters as $datacenter) {
            if ($datacenter->id > $maxId) {
                $maxId = (int)$datacenter->id;
            }
        }
        $maxId++;
        return $maxId;
    }

    /**
     * @param $resource
     * @throws Exception
     * @throws Service_BaseClient_HTTPException
     * @throws \Rest\RestException
     */
    private function checkAvailabilityChangeCredentials($resource)
    {
        $mspApi = $this->getMspApi();
        $group = $mspApi->getGroupById('self');
        $instanceId = Str::format(
            self::UNDERSCORE_DELIMITER,
            $this->_apsc->getInstanceId(),
            md5($resource->login . $resource->url)
        );

        $mspApiNew = $this->_facade->getMspApi($resource->url, $instanceId);
        $mspApiNew->setCredentials($resource->login, $resource->password, $resource->url);
        $mspApiNew->login();
        $newGroup = $mspApiNew->getGroupById('self');

        if ($group->getId() != $newGroup->getId()) {
            throw new \Rest\RestException(409, self::EXCEPTION_CANNOT_CHANGE_CREDENTIALS_MESSAGE);
        }
    }

    private function checkStorageUsage()
    {
        $profiles = $this->global->storages;

        if (!count($profiles)) {
            return;
        }
        foreach ($profiles as $profile) {
            if (!is_null($profile->storages)) {
                /** @var BackupProfileStorage $storage */
                foreach ($profile->storages as $storage) {
                    if ($storage->datacenterId == $this->id) {
                        throw new \Rest\RestException(
                            self::EXCEPTION_LOCKED_CODE,
                            Str::format(self::EXCEPTION_CANNOT_REMOVE_DATACENTER_STORAGE_MESSAGE, $this->id, $this->name)
                        );
                    }
                }
            } elseif (!is_null($profile->datacenters)) {
                /** @var BackupProfileStorage $storage */
                foreach ($profile->datacenters as $datacenter) {
                    if ($datacenter->datacenterId == $this->id) {
                        throw new \Rest\RestException(
                            self::EXCEPTION_LOCKED_CODE,
                            Str::format(self::EXCEPTION_CANNOT_REMOVE_DATACENTER_PROFILE_MESSAGE, $this->id, $this->name)
                        );
                    }
                }
            }
        }
    }

    private function _checkLinkedResellers()
    {
        $resellers = $this->global->resellerRelations;

        if (!is_array($resellers)) {
            return;
        }

        foreach ($resellers as $reseller) {
            if ($reseller->datacenterId == $this->id) {
                throw new \Rest\RestException(
                    self::EXCEPTION_LOCKED_CODE,
                    Str::format(self::EXCEPTION_CANNOT_REMOVE_DATACENTER_RESELLER_MESSAGE, $this->id, $this->name)
                );
            }
        }
    }

    /**
     * @throws Exception
     */
    private function updatePackageVersionInfo()
    {
        $mspGroupManager = new Model_MspGroupManager($this->getMspApi(), $this->_getApsApi());
        $mspGroupManager->updatePackageVersionInfo('self');
    }
}
