<?php
/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

class ExternalAccountsImporter
{
    const PROPERTY_APS_TOKEN = "aps_token";
    const PROPERTY_CONTROLLER_URI = "controller_uri";

    const ORDER_STATUS_NEW = 'NW';
    const ORDER_STATUS_OPEN = 'OP';

    const MINIMUM_TIER_LEVEL = -1;

    const TRIAL_PLAN_PERIOD = 1;

    const ATTRIBUTE_EXTERNAL_GROUP_INFO = 'ExternalGroupInfo';
    const ATTRIBUTE_EXTERNAL_GROUP_INFO_FORMAT = "Datacenter=%s;Group=%s";

    const SUBSCRIPTION_STATUS_ORDERED = 10;
    const SUBSCRIPTION_STATUS_TERMINATED = 60;
    const SUBSCRIPTION_STATUS_CANCELED = 70;

    /**
     * @var ApiFactory
     */
    private $apiFactory;

    /**
     * @var Service_OpenApi_Api
     */
    private $openApi;

    /**
     * @var Service_PbaApi_Api
     */
    private $pbaApi;

    /**
     * @var string
     */
    private $httpApsInstanceId;

    /**
     * @var int
     */
    private $applicationInstanceId;

    /**
     * @var array
     */
    private $_cachedDatacenters = array();

    /**
     * @var array
     */
    private $_cachedTiers = array();

    /**
     * @var Model_OSA_AccountAttributesManager
     */
    private $accountAttributesManager;

    /**
     * @param ApiFactory $apiFactory
     * @param string $httpApsInstanceId
     * @param int $applicationInstanceId
     */
    public function __construct(ApiFactory $apiFactory, $httpApsInstanceId, $applicationInstanceId)
    {
        $this->apiFactory = $apiFactory;

        $this->httpApsInstanceId = $httpApsInstanceId;
        $this->applicationInstanceId = $applicationInstanceId;

        $this->openApi = $this->createOpenApiInstance();
        $this->pbaApi = $this->createPbaApiInstance();

        $this->accountAttributesManager = new Model_OSA_AccountAttributesManager($this->pbaApi);
    }

    /**
     * @param array $accountsImportData
     */
    public function import($accountsImportData)
    {
        if (!$accountsImportData) {
            Log::getInstance()->notice('There are no accounts to import');
            return;
        }

        Log::getInstance()->notice(str_repeat('=', 50));
        foreach ($accountsImportData as $accountImportData) {
            Log::getInstance()->notice('Importing %s ...', json_encode($accountImportData));
            try {
                $this->checkAccountImportData($accountImportData);
                $this->importAccountChildren($accountImportData);
            } catch (\Exception $e) {
                Log::getInstance()->error($e->getMessage());
            }
            Log::getInstance()->notice(str_repeat('=', 50));
        }
    }

    /**
     * @param array $accountImportData
     * @throws Exception
     */
    private function importAccountChildren($accountImportData)
    {
        $parentAccountId = $accountImportData['accountId'];

        $accountInfo = $this->openApi->getAccountInfo($parentAccountId);
        if ($accountInfo->getAccountType() == 'C') {
            throw new Exception(Str::format('The account "%s" is customer. You cannot import account under customer.', $parentAccountId));
        }

        $datacenterId = $accountImportData['datacenterId'];
        $datacenter = $this->fetchDatacenterById($datacenterId);
        if (!$datacenter) {
            throw new Exception("There is no datacenter with id \"{$datacenterId}\".");
        }

        $mspApi = $this->createMspApiForDatacenter($datacenter);
        $rootGroupId = $accountImportData['groupId'];
        try {
            $allGroups = $mspApi->getGroups($rootGroupId);
        } catch (Service_BaseClient_HTTPException $e) {
            if ($e->getCode() === Service_BaseClient_HTTPException::HTTP_NOT_FOUND_CODE) {
                throw new Exception(
                    Str::format('Group "%s" s not found.', $rootGroupId),
                    Service_BaseClient_HTTPException::HTTP_NOT_FOUND_CODE
                );
            }
            throw $e;
        }
        $serviceProviders = array_filter(
            $allGroups,
            function (Service_MspApiClient_Response_Struct_GroupItem $group) {
                return $group->getKind() == Service_MspApiClient_Request_Struct_CreateGroupInterface::RESELLER_GROUP_KIND_ID;
            }
        );
        Log::getInstance()->notice(
            "There are %s service providers were fetched from Acronis Backup Cloud.",
            count($serviceProviders)
        );

        $this->cancelSubscriptionsForNotExistingServiceProviders($parentAccountId, $datacenterId, $serviceProviders);

        if (!$serviceProviders) {
            Log::getInstance()->notice('There are no service providers to import.');
            return;
        }

        foreach ($serviceProviders as $reseller) {
            try {
                $this->importServiceProvider($datacenter, $parentAccountId, $reseller, $accountImportData['aan']);
            } catch (\Exception $e) {
                Log::getInstance()->error($e->getMessage());
            }
            Log::getInstance()->notice(str_repeat('-', 50));
        }
    }

    /**
     * @param $datacenter
     * @param int $parentAccountId
     * @param Service_MspApiClient_Response_Struct_GroupItem $group
     * @param string $acronisAccountNumber
     */
    private function importServiceProvider(
        $datacenter,
        $parentAccountId,
        Service_MspApiClient_Response_Struct_GroupItem $group,
        $acronisAccountNumber
    )
    {
        Log::getInstance()->notice(Str::format(
            "Importing group (service provider) with id \"%s\" ...",
            $group->getId()
        ));

        $tierLevelInBackupCloud = $this->getValidatedTierLevelFromGroup($group);
        Log::getInstance()->notice("Tier level in Acronis Backup Cloud is \"{$tierLevelInBackupCloud}\"");

        $minimalCommitmentMonth = $this->getMinComMonthFromGroup($group);
        Log::getInstance()->notice("Minimum commitment apply month is \"{$minimalCommitmentMonth}\"");

        $tenant = $this->fetchTenantByGroupId($datacenter->id, $group->getId());

        if (!$tenant) {
            Log::getInstance()->notice(
                "Tenant for group \"%s\" does not exist. So we will create account and place order for him",
                $group->getId()
            );
            $this->createAccountAndPlaceOrder(
                $datacenter,
                $parentAccountId,
                $group,
                $tierLevelInBackupCloud,
                $acronisAccountNumber,
                $minimalCommitmentMonth
            );
            return;
        }

        $accountId = $tenant->accountId;
        $curMinimalCommitmentMonth = $this->accountAttributesManager->getAttributeMinimalCommitmentMonth($accountId);
        if (!is_null($minimalCommitmentMonth) && $minimalCommitmentMonth != $curMinimalCommitmentMonth) {
            $this->accountAttributesManager->setAttributeMinimalCommitmentMonth($accountId, $minimalCommitmentMonth);
        }

        $curTierLevel = $tenant->minimalCommitmentTier->isEmpty() ? null : $tenant->minimalCommitmentTier->level;

        Log::getInstance()->notice(
            "Tenant for group \"%s\" exists: %s",
            $group->getId(),
            json_encode(array(
                'apsId' => $tenant->aps->id,
                'accountId' => $tenant->account->id,
                'subscriptionId' => $tenant->subscription->subscriptionId,
                'curTierLevel' => $curTierLevel,
            ))
        );

        if ($tierLevelInBackupCloud == $curTierLevel) {
            Log::getInstance()->notice("Levels are equal so we don't need to switch plan");
            return;
        };

        $this->switchSubscriptionToPlanWithNeededTierLevel($tenant->subscription, $tierLevelInBackupCloud);
    }

    /**
     * @return Service_PbaApi_Api
     */
    private function createPbaApiInstance()
    {
        return $this->apiFactory->getPbaApi(Settings::getInstance()->getBillingConfig());
    }

    /**
     * @return Service_OpenApi_Api
     */
    private function createOpenApiInstance()
    {
        return $this->apiFactory->getOpenApi(OPEN_API_URI);
    }

    private function createApsController()
    {
        $token = $this->openApi->getApplicationInstanceToken($this->applicationInstanceId);
        return $this->apiFactory->getApsController(
            $token[self::PROPERTY_CONTROLLER_URI],
            $token[self::PROPERTY_APS_TOKEN],
            $this->httpApsInstanceId
        );
    }

    /**
     * @param $datacenter
     * @return Service_MspApiClient_Api
     */
    private function createMspApiForDatacenter($datacenter)
    {
        return $this->apiFactory->getMspApi($datacenter, $this->httpApsInstanceId);
    }

    private function fetchDatacenterById($id)
    {
        if (!array_key_exists($id, $this->_cachedDatacenters)) {
            $rqlQuery = RqlBuilder::createInstance()
                ->implementing(APSResource::CLOUD_BACKUP_DATACENTER_TYPE)
                ->eq('aps.status', Service_APSApi_Api::APS_READY)
                ->eq('id', $id)
                ->getQuery();
            $datacenters = $this->createApsController()->getResources($rqlQuery);
            if (count($datacenters) > 1) {
                throw new \Exception("There are more than one datacenters with id \"{$id}\"");
            }
            $this->_cachedDatacenters[$id] = $datacenters ? reset($datacenters) : null;
        }
        return $this->_cachedDatacenters[$id];
    }

    private function fetchTenantByGroupId($datacenterId, $groupId)
    {
        $rqlQuery = RqlBuilder::createInstance()
            ->implementing(APSResource::CLOUD_BACKUP_TENANT_TYPE_CCPv2)
            ->select(array('account', 'subscription', 'minimalCommitmentTier'))
            ->eq('aps.status', Service_APSApi_Api::APS_READY)
            ->eq('type', tenantAcronis::TYPE_EXTERNAL)
            ->eq('datacenter.id', $datacenterId)
            ->eq('mspGroupId', $groupId)
            ->getQuery();
        $tenants = $this->createApsController()->getResources($rqlQuery);
        if (!$tenants) {
            return null;
        }
        if (count($tenants) > 1) {
            throw new \Exception("There are more than one datacenters with mspGroupId \"{$groupId}\"");
        }
        return reset($tenants);
    }

    private function fetchTenantsForAccountsByDatacenterAndNotEqualGroup($accountIds, $datacenterId, array $groupIds = array())
    {
        $rqlQuery = RqlBuilder::createInstance()
            ->implementing(APSResource::CLOUD_BACKUP_TENANT_TYPE_CCPv2)
            ->select(array('account', 'subscription', 'minimalCommitmentTier'))
            ->eq('aps.status', Service_APSApi_Api::APS_READY)
            ->eq('type', tenantAcronis::TYPE_EXTERNAL)
            ->in('account.id', $accountIds)
            ->eq('datacenter.id', $datacenterId);
        if ($groupIds) {
            $rqlQuery->out('mspGroupId', $groupIds);
        }
        return $this->createApsController()->getResources($rqlQuery->getQuery());
    }

    /**
     * @param Service_MspApiClient_Response_Struct_GroupItem $group
     * @return int
     * @throws Exception
     */
    private function getValidatedTierLevelFromGroup(Service_MspApiClient_Response_Struct_GroupItem $group)
    {
        $decodedInternalTag = Arr::decode($group->getInternalTag());

        $tierLevel = isset($decodedInternalTag[Model_MspGroupManager::TIER_LEVEL]) ?
            $decodedInternalTag[Model_MspGroupManager::TIER_LEVEL] :
            minimalCommitmentTier::LEVEL_NO_COMMITMENT;

        $filterVarOptions = array('options' => array('min_range' => self::MINIMUM_TIER_LEVEL));
        if (filter_var($tierLevel, FILTER_VALIDATE_INT, $filterVarOptions) === false) {
            throw new \Exception(Str::format(
                "Group with id \"%s\" has invalid value of \"%s\" property: %s",
                $group->getId(),
                Model_MspGroupManager::TIER_LEVEL,
                $tierLevel
            ));
        }

        return (int)$tierLevel;
    }

    /**
     * @param Service_MspApiClient_Response_Struct_GroupItem $group
     * @return int
     * @throws Exception
     */
    private function getMinComMonthFromGroup(Service_MspApiClient_Response_Struct_GroupItem $group)
    {
        $decodedInternalTag = Arr::decode($group->getInternalTag());
        if (!isset($decodedInternalTag[Model_MspGroupManager::START_PROD_DATE])) {
            return null;
        }

        $startProductionTime = $decodedInternalTag[Model_MspGroupManager::START_PROD_DATE];

        $date = new DateTime();
        $date->setTimestamp($startProductionTime);

        $day = $date->format('d');

        $date->setDate($date->format('Y'), $date->format('m'), 1)
            ->setTime(0, 0, 0);

        // skip not full month
        if ($day != '01') {
            $date->modify('+1 month');
        }

        return $date->format(Model_OSA_AccountAttributesManagerInterface::MINIMAL_COMMITMENT_MONTH_FORMAT);
    }

    private function fetchTierResourceToSwitch($level)
    {
        if (!array_key_exists($level, $this->_cachedTiers)) {
            $rqlQuery = RqlBuilder::createInstance()
                ->implementing(APSResource::CLOUD_BACKUP_MINIMAL_COMMITMENT_TIER)
                ->getQuery();
            $tiers = $this->createApsController()->getResources($rqlQuery);
            $foundTiers = array();
            foreach ($tiers as $tier) {
                if ($tier->level == $level) {
                    $foundTiers [] = $tier;
                }
            }
            if (count($foundTiers) > 1) {
                throw new \Exception("There are more than one MC Tier resources with level {$level}");
            }
            $this->_cachedTiers[$level] = $foundTiers ? reset($foundTiers) : null;
        }
        return $this->_cachedTiers[$level];
    }

    /**
     * @param Service_PbaApi_Api_Response_Struct_SubscriptionAvailablePlanListItem[] | Service_PbaApi_Api_Response_Struct_PlanInterface[] $plans
     * @param $resource
     * @return Service_PbaApi_Api_Response_Struct_SubscriptionAvailablePlanListItem | Service_PbaApi_Api_Response_Struct_PlanInterface | null
     */
    private function searchPlanWithIncludedResource($plans, $resource)
    {
        $foundPlans = array();
        foreach ($plans as $plan) {
            $ratesList = $this->pbaApi->getPlanRateListFull($plan->getPlanId());
            foreach ($ratesList as $rateInfo) {
                if ($rateInfo->getResourceId() == $resource->resourceTypeId && $rateInfo->getIncludedValue() != 0) {
                    $foundPlans [] = $plan;
                }
            }
        }
        if (count($foundPlans) > 1) {
            throw new \Exception("There are more than one plan with included resource \"{$resource->resourceTypeId}\"");
        }
        return $foundPlans ? reset($foundPlans) : null;
    }

    /**
     * @param $subscription
     * @param Service_PbaApi_Api_Response_Struct_SubscriptionAvailablePlanListItem $plan
     * @return Service_PbaApi_Api_Response_Struct_PlacePlanPeriodSwitchOrder
     * @throws Exception
     */
    private function placePlanSwitchOrderForSubscription(
        $subscription,
        Service_PbaApi_Api_Response_Struct_SubscriptionAvailablePlanListItem $plan
    )
    {
        $subscriptionInfo = $this->pbaApi->getSubscriptionDetailsEx($subscription->subscriptionId);

        $periodDuration = $subscriptionInfo->getPeriod();
        $periodType = $subscriptionInfo->getPeriodType();
        $period = $this->findPlanPeriod($plan->getPlanId(), $periodDuration, $periodType);
        if (is_null($period)) {
            throw new Exception(Str::format(
                'Plan "%s" does not have an enabled non-trial period with duration "%s" and of type "%s" (1 - Day(s), 2 - Month(s), 3 - Year(s))',
                $plan->getPlanId(), $periodDuration, $periodType
            ));
        }

        return $this->pbaApi->placePlanPeriodSwitchOrder(
            $subscription->subscriptionId,
            $plan->getPlanId(),
            $period->getPlanPeriodID(),
            20
        );
    }

    /**
     * @param $planId
     * @param $periodDuration
     * @param $periodType
     * @return null|Service_PbaApi_Api_Response_Struct_PlanPeriodListItem
     */
    private function findPlanPeriod($planId, $periodDuration, $periodType)
    {
        $periods = $this->pbaApi->getPlanPeriodList($planId);
        foreach ($periods as $period) {
            if ($period->getTrial() || !$period->getEnabled()) {
                continue;
            }
            if ($period->getPeriod() == $periodDuration && $period->getPeriodType() == $periodType) {
                return $period;
            }
        }
        return null;
    }

    /**
     * @param int $orderId
     */
    private function openOrder($orderId)
    {
        Log::getInstance()->notice('Changing status for order "%s" to "%s".', $orderId, self::ORDER_STATUS_OPEN);
        $order = $this->pbaApi->getOrder($orderId);
        if ($order->getOrderStatusID() != self::ORDER_STATUS_NEW) {
            Log::getInstance()->notice('The status of the order "%s" is already "%s".', $orderId, $order->getOrderStatusID());
            return;
        }
        $result = $this->pbaApi->changeOrderStatus(
            $order->getOrderID(),
            self::ORDER_STATUS_OPEN,
            md5(
                $order->getOrderID() .
                trim($order->getOrderNumber()) .
                $order->getCreationTime() .
                sprintf('%s %.2f', trim($order->getCurrencyID()), $order->getTotal()) .
                $order->getComments()
            )
        );
        Log::getInstance()->notice('Result of status changing is "%s".', $result->getStatus());
    }

    private function switchSubscriptionToPlanWithNeededTierLevel($subscription, $tierLevel)
    {
        $tierResource = $this->fetchTierResourceToSwitch($tierLevel);
        if (!$tierResource) {
            throw new \Exception("There is no minimal commitment tier resource with level {$tierLevel}");
        }

        $availablePlans = $this->pbaApi->getSubscriptionAvailablePlanList($subscription->subscriptionId);
        Log::getInstance()->notice(Str::format(
            "There are %s available plans for subscription",
            count($availablePlans)
        ));

        $plan = $this->searchPlanWithIncludedResource($availablePlans, $tierResource);
        if (!$plan) {
            throw new \Exception(Str::format(
                "There is no available plan with resource \"%s\" for subscription \"%s\"",
                $tierResource->resourceTypeId,
                $subscription->subscriptionId
            ));
        }

        Log::getInstance()->notice(
            'Placing order to switch subscription "%s" to plan "%s"',
            $subscription->subscriptionId,
            $plan->getPlanId()
        );

        $orderInfo = $this->placePlanSwitchOrderForSubscription($subscription, $plan);
        Log::getInstance()->notice('Order was placed: %s %s', $orderInfo->getOrderNumber(), $orderInfo->getDescription());

        $this->openOrder($orderInfo->getOrderId());
    }

    /**
     * @param $datacenter
     * @param int $parentAccountId
     * @param Service_MspApiClient_Response_Struct_GroupItem $group
     * @param int $tierLevel
     * @param string $acronisAccountNumber
     * @param $minimalCommitmentMonth
     * @throws Exception
     */
    private function createAccountAndPlaceOrder(
        $datacenter,
        $parentAccountId,
        Service_MspApiClient_Response_Struct_GroupItem $group,
        $tierLevel,
        $acronisAccountNumber,
        $minimalCommitmentMonth
    )
    {
        $tierResource = $this->fetchTierResourceToSwitch($tierLevel);
        if (!$tierResource) {
            throw new \Exception("There is no resource with tier level {$tierLevel}");
        }

        $plans = $this->pbaApi->getServicePlans($parentAccountId);
        $plan = $this->searchPlanWithIncludedResource($plans, $tierResource);
        if (!$plan) {
            throw new \Exception("There is no plan with resource {$tierResource->resourceTypeId}");
        }

        $planPeriod = $this->findFirstNotTrialPeriodForPlan($plan);
        if (!$planPeriod) {
            throw new \Exception(Str::format(
                "There is no not trial plan period for plan %s",
                $plan->getPlanId()
            ));
        }

        $this->validateAccountAttributes();
        $this->validateSubscriptionParameters($datacenter, $group, $plan);

        $parameters = $this->buildRequestParametersToCreateAccountAndPlaceOrder(
            $plan,
            $planPeriod,
            $datacenter,
            $group,
            $acronisAccountNumber
        );

        $placeOrderResult = $this->pbaApi->placeOrderAndAuthorize(
            $parentAccountId,
            $parameters
        );
        $accountId = $placeOrderResult[0];
        $orderNumber = $placeOrderResult[10];
        Log::getInstance()->notice(
            "Account was created, order was placed: %s",
            json_encode(compact('accountId', 'orderNumber'))
        );

        if (!is_null($minimalCommitmentMonth)) {
            $this->accountAttributesManager->setAttributeMinimalCommitmentMonth($accountId, $minimalCommitmentMonth);
        }
    }

    /**
     * @param Service_PbaApi_Api_Response_Struct_PlanInterface $plan
     * @param Service_PbaApi_Api_Response_Struct_PlanPeriodListItem $planPeriod
     * @param $datacenter
     * @param Service_MspApiClient_Response_Struct_GroupItem $group
     * @param string $acronisAccountNumber
     * @return array
     */
    private function buildRequestParametersToCreateAccountAndPlaceOrder(
        Service_PbaApi_Api_Response_Struct_PlanInterface $plan,
        Service_PbaApi_Api_Response_Struct_PlanPeriodListItem $planPeriod,
        $datacenter,
        Service_MspApiClient_Response_Struct_GroupItem $group,
        $acronisAccountNumber
    )
    {
        $provisioningItemsDefaultValue = 1;
        $provisioningDataSlotCounter = 1;
        $provisioningItemIdDefaultValue = 0;
        $provisioningParametersCount = 3;

        $params = array(
            $provisioningItemsDefaultValue,
            "{$plan->getPlanId()}={$planPeriod->getPlanPeriodID()}=0=-1",
            $provisioningDataSlotCounter,
            $provisioningItemIdDefaultValue,
            $provisioningParametersCount,
            Model_OSA_SubscriptionParametersManagerInterface::PARAMETER_DATACENTER . '=' . $datacenter->id,
            Model_OSA_SubscriptionParametersManagerInterface::PARAMETER_GROUP_ID . '=' . $group->getId(),
            Model_OSA_SubscriptionParametersManagerInterface::PARAMETER_IS_EXTERNAL_GROUP . '=1',
        );

        $accountId = $this->fetchServiceProviderAccountId($datacenter, $group);
        if ($accountId) {
            $contactDataParametersCount = 2;
            // use an existing account
            $contactData = array(
                $contactDataParametersCount,
                "AccountID={$accountId}",
                "PromoCodeID=",
            );
        } else {
            $contactDataParametersCount = 24;

            $login = static::generateLogin();
            $password = static::generatePassword(12);
            $email = $login . '@fake.acronis.com';

            $mspApi = $this->createMspApiForDatacenter($datacenter);
            $groupId = $group->getId();
            try {
                $groupInfo = $mspApi->getGroupById($groupId);
            } catch (Service_BaseClient_HTTPException $e) {
                if ($e->getCode() === Service_BaseClient_HTTPException::HTTP_NOT_FOUND_CODE) {
                    throw new Exception(
                        Str::format('Group "%s" s not found.', $groupId),
                        Service_BaseClient_HTTPException::HTTP_NOT_FOUND_CODE
                    );
                }
                throw $e;
            }

            $groupContact = $groupInfo->getContact();

            $companyName = $groupInfo->getName();
            $firstName = $groupContact->getFirstname() ?: 'FirstName';
            $middleName = '';
            $lastName = $groupContact->getLastname() ?: 'LastName';
            $country = $groupContact->getCountry() ?: 'US';
            $state = $groupContact->getState() ?: '';
            $city = $groupContact->getCity() ?: 'City';
            $zip = $groupContact->getZipcode() ?: ($country === 'CA' ? 'a1a 1a1' : '11111');
            $address = $groupContact->getAddress1() ?: ($groupContact->getAddress2() ?: 'Address');
            $phone = $groupContact->getPhone() ?: '+11111111111';

            // create a new account
            $contactData = array(
                $contactDataParametersCount,
                "LoginID={$login}",
                "XXXPasswordID={$password}",
                "CompanyNameID={$companyName}",
                "FirstNameID={$firstName}",
                "MiddleNameID={$middleName}",
                "LastNameID={$lastName}",
                "AddressID={$address}",
                "Address2ID=",
                "CityID={$city}",
                "StateID={$state}",
                "ZipID={$zip}",
                "CountryID={$country}",
                "EmailID={$email}",
                "PhoneCountryID=1",
                "PhoneAreaID=0",
                "PhoneNumberID={$phone}",
                "PhoneExtensionID=0",
                "FaxCountryID=",
                "FaxAreaID=",
                "FaxNumberID=",
                "FaxExtensionID=",
                "PromoCodeID=",
                Model_OSA_AccountAttributesManager::ATTRIBUTE_ACRONIS_ACCOUNT_NUMBER . '=' . $acronisAccountNumber,
                self::ATTRIBUTE_EXTERNAL_GROUP_INFO . '=' . Str::format(
                    self::ATTRIBUTE_EXTERNAL_GROUP_INFO_FORMAT, $datacenter->id, $groupId
                ),
            );
        }
        $params = array_merge($params, $contactData);

        $payToolCount = 3;
        $payToolTypeId = 3; // Check/Cash
        $ipAddress = '127.0.0.1';

        $params = array_merge($params, array(
            $payToolCount,
            "PayToolTypeID={$payToolTypeId}",
            "PluginID=0",
            "IPAddressID={$ipAddress}",
        ));

        return $params;
    }

    /**
     * @param Service_PbaApi_Api_Response_Struct_PlanInterface $plan
     * @return Service_PbaApi_Api_Response_Struct_PlanPeriodListItem | null
     */
    private function findFirstNotTrialPeriodForPlan($plan)
    {
        $planPeriods = $this->pbaApi->getPlanPeriodList($plan->getPlanId());
        foreach ($planPeriods as $planPeriod) {
            if (!$planPeriod->getTrial()) {
                return $planPeriod;
            }
        }
        return null;
    }

    /**
     * @return string
     */
    private static function generateLogin()
    {
        return uniqid();
    }

    /**
     * @param int $length
     * @return string
     */
    private static function generatePassword($length)
    {
        $digits = '0123456789';
        $letters = 'abcdefghijklmnopqrstuvwxyz';
        $specialChars = '!@#$%&*?';
        $pool = str_repeat($digits, 3)
            . $letters
            . strtoupper($letters)
            . str_repeat($specialChars, 3);
        return substr(str_shuffle(str_repeat($pool, $length)), 0, $length);
    }

    /**
     * @param $datacenter
     * @param Service_MspApiClient_Response_Struct_GroupItem $group
     * @return int|null
     */
    private function fetchServiceProviderAccountId($datacenter, Service_MspApiClient_Response_Struct_GroupItem $group)
    {
        $externalGroupInfo = Str::format(self::ATTRIBUTE_EXTERNAL_GROUP_INFO_FORMAT, $datacenter->id, $group->getId());
        $accounts = $this->pbaApi->getAccountListByCustomAttribute(self::ATTRIBUTE_EXTERNAL_GROUP_INFO, $externalGroupInfo);
        if (empty($accounts)) {
            return null;
        }

        if (count($accounts) > 1) {
            Log::getInstance()->warning(
                'Misconfiguration is detected. There are several accounts (%s) with the same account attribute %s="%s". The first account will be used.',
                implode(', ', array_map(function ($accountInfo) {
                    /**
                     * @var Service_PbaApi_Api_Response_Struct_GetAccountListByCustomAttribute $accountInfo
                     */
                    return $accountInfo->getAccountId();
                }, $accounts)),
                self::ATTRIBUTE_EXTERNAL_GROUP_INFO,
                $externalGroupInfo
            );
        }

        $accountInfo = reset($accounts);
        return $accountInfo->getAccountId();
    }

    private function cancelSubscription($subscriptionId)
    {
        Log::getInstance()->notice(
            'Place a cancellation order for the subscription "%s".',
            $subscriptionId
        );

        $pbaApi = $this->pbaApi;
        $subscriptionDetails = $pbaApi->getSubscriptionDetails($subscriptionId);
        $status = $subscriptionDetails->getStatus();
        if ($status == self::SUBSCRIPTION_STATUS_ORDERED) {
            Log::getInstance()->warning('Cannot place a cancellation order for the subscription "%s". The subscription has status 10 (Ordered).', $subscriptionId);
            return;
        }
        if ($status == self::SUBSCRIPTION_STATUS_TERMINATED || $status == self::SUBSCRIPTION_STATUS_CANCELED) {
            Log::getInstance()->notice('Skip the subscription "%s". The subscription is already terminated.', $subscriptionId);
            return;
        }
        list($status, $orderId) = $pbaApi->placeSubscriptionCancellationOrder($subscriptionId, 30, 1);

        Log::getInstance()->notice('Status: %s Order ID: %s', $status, $orderId);

        $this->openOrder($orderId);
    }

    /**
     * @param $accountImportData
     * @throws Exception
     */
    private function checkAccountImportData($accountImportData)
    {
        $stringParameter = array('aan');
        $integerParameter = array('datacenterId', 'groupId', 'accountId');

        foreach (array_merge($integerParameter, $stringParameter) as $parameter) {
            if (!isset($accountImportData[$parameter]) || $accountImportData[$parameter] == '') {
                throw new Exception(Str::format('Please specify the parameter "%s".', $parameter));
            }
        }

        foreach ($integerParameter as $parameter) {
            if (!is_int($accountImportData[$parameter])) {
                throw new Exception(Str::format('The parameter "%s" must be integer.', $parameter));
            }
        }
    }

    /**
     * @param $parentAccountId
     * @param $datacenterId
     * @param $existingServiceProviders
     */
    private function cancelSubscriptionsForNotExistingServiceProviders($parentAccountId, $datacenterId, $existingServiceProviders)
    {
        Log::getInstance()->notice(str_repeat('-', 50));

        $customerIds = array_map(function ($accountInfo) {
            return $accountInfo['account_id'];
        }, $this->openApi->getCustomerIds($parentAccountId));

        if (empty($customerIds)) {
            Log::getInstance()->notice(
                "There are no service providers to destroy."
            );
            Log::getInstance()->notice(str_repeat('-', 50));
            return;
        }

        $tenantsToDestroy = $this->fetchTenantsForAccountsByDatacenterAndNotEqualGroup(
            $customerIds,
            $datacenterId,
            array_map(function (Service_MspApiClient_Response_Struct_GroupItem $group) {
                return $group->getId();
            }, $existingServiceProviders)
        );

        Log::getInstance()->notice(
            "There are %s service providers to destroy.",
            count($tenantsToDestroy)
        );
        foreach ($tenantsToDestroy as $tenant) {
            try {
                $this->cancelSubscription($tenant->subscription->subscriptionId);
            } catch (\Exception $e) {
                Log::getInstance()->error($e->getMessage());
            }
            Log::getInstance()->notice(str_repeat('-', 50));
        }
    }

    /**
     * @param $datacenter
     * @param Service_MspApiClient_Response_Struct_GroupItem $group
     * @param $plan
     * @throws Exception
     */
    private function validateSubscriptionParameters($datacenter, Service_MspApiClient_Response_Struct_GroupItem $group, $plan)
    {
        $datacenterParameter = new Service_PbaApi_Api_Request_Struct_ServiceTemplateParameter();
        $datacenterParameter->setId(Model_OSA_SubscriptionParametersManagerInterface::PARAMETER_DATACENTER);
        $datacenterParameter->setValue($datacenter->id);

        $groupParameter = new Service_PbaApi_Api_Request_Struct_ServiceTemplateParameter();
        $groupParameter->setId(Model_OSA_SubscriptionParametersManagerInterface::PARAMETER_GROUP_ID);
        $groupParameter->setValue($group->getId());

        $isExternalGroupParameter = new Service_PbaApi_Api_Request_Struct_ServiceTemplateParameter();
        $isExternalGroupParameter->setId(Model_OSA_SubscriptionParametersManagerInterface::PARAMETER_IS_EXTERNAL_GROUP);
        $isExternalGroupParameter->setValue(1);

        $validationStatuses = $this->pbaApi->getAndValidatePlanParameters($plan->getPlanId(), $plan->getServiceTemplateId(), array(
            $datacenterParameter, $groupParameter, $isExternalGroupParameter
        ));

        $mandatorySubscriptionParameters = array(
            Model_OSA_SubscriptionParametersManagerInterface::PARAMETER_DATACENTER => false,
            Model_OSA_SubscriptionParametersManagerInterface::PARAMETER_GROUP_ID => false,
            Model_OSA_SubscriptionParametersManagerInterface::PARAMETER_IS_EXTERNAL_GROUP => false,
        );

        foreach ($validationStatuses as $validationStatus) {
            $parameterId = $validationStatus->getParameterId();
            if (!isset($mandatorySubscriptionParameters[$parameterId])) {
                continue;
            }
            if ($validationStatus->getStatus() != Service_PbaApi_Api_Response_Struct_GetAndValidatePlanParams::PARAMETER_VALIDATION_STATUS_VALID) {
                throw new Exception(Str::format(
                    'The value of the parameter "%s" is invalid. %s',
                    $parameterId,
                    $validationStatus->getMessage()
                ));
            }
            $mandatorySubscriptionParameters[$parameterId] = true;
        }
        $absentSubscriptionParameters = array_filter($mandatorySubscriptionParameters, function ($status) {
            return !$status;
        });
        if (!empty($absentSubscriptionParameters)) {
            throw new Exception(Str::format(
                'The follow parameters (%s) must be presented in the service template %s.',
                implode(', ', array_keys($absentSubscriptionParameters)),
                $plan->getServiceTemplateId()
            ));
        }
    }

    private function validateAccountAttributes()
    {
        $mandatoryAccountAttributes = array(
            Model_OSA_AccountAttributesManagerInterface::ATTRIBUTE_ACRONIS_ACCOUNT_NUMBER => false,
            self::ATTRIBUTE_EXTERNAL_GROUP_INFO => false
        );

        $attributes = $this->pbaApi->getAttributesTypeList();
        foreach ($attributes as $attribute) {
            $attributeId = $attribute->getAttributeId();
            if (!isset($mandatoryAccountAttributes[$attributeId]) ||
                $attribute->getActive() != 1 ||
                $attribute->getAttributeType() != 0
            ) {
                continue;
            }
            $mandatoryAccountAttributes[$attributeId] = true;
        }

        $absentAccountAttributes = array_filter($mandatoryAccountAttributes, function ($status) {
            return !$status;
        });
        if (!empty($absentAccountAttributes)) {
            throw new Exception(Str::format(
                'The follow attributes (%s) must be presented, active and available for accounts.',
                implode(', ', array_keys($absentAccountAttributes))
            ));
        }
    }
}
