<?php
/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

use Monolog\Formatter\LineFormatter;
use Monolog\Handler\StreamHandler;
use Monolog\Logger;

require_once(__DIR__ . "/../app/bootstrap.php");
require_once(__DIR__ . "/place_billing_orders/src/BillingInternalHelper.php");

chdir(DIR_ROOT);
define('APS_RUNTIME_HOME', DIR_ROOT);
define('APS_RUNTIME_NO_PROCESS', true);

const OUTPUT_OPTION_UNDEFINED = "Option '%s' undefined \n";

const OPTION_PARSE_CHAR = ':';
const OPTION_APPLICATION = 'a';
const OPTION_RESOURCES = 'r';
const OPTION_BILL_MODE = 'm';
const OPTION_DRY_RUN_MODE = 'd';
const OPTION_VERBOSE = 'v';
const OPTION_HELP = 'h';

/**
 *  Prints usage to the screen
 */
function printScriptUsage()
{
    $fileName = basename($_SERVER['PHP_SELF']);
    echo <<<USAGE
Usage: php {$fileName} [OPTIONS]
Example: php {$fileName} -a 10 -r 100001 -r 100002 -r 100003 -r 100004 -m ALL -d ALL -vv
Options:
  -a APPLICATION    Acronis Backup Service application ID.
  -r RESOURCE       Billable resource.
  -m ALL | LAST     Calculation mode.
                    Options: 
                      ALL - Summarize all orders in period;
                      LAST - Use the latest order only.
  -d ALL | CH | BO  Disable a generation of orders.
                    Options: 
                      ALL - all types of orders;
                      CH - disable a generation of orders to change a plan for a subscription;
                      BO - disable a generation of billing orders.
  -v                Verbose.

USAGE;
}

function validateCommandLineParameters($opt)
{
    if (isset($opt[OPTION_HELP])) {
        return true;
    }

    if (!isset($opt[OPTION_APPLICATION]) || !ctype_digit($opt[OPTION_APPLICATION])) {
        return false;
    }

    if (!isset($opt[OPTION_RESOURCES])) {
        return false;
    }

    $resources = (array) $opt[OPTION_RESOURCES];
    foreach ($resources as $resource) {
        if (!ctype_digit($resource)) {
            return false;
        }
    }

    if (!isset($opt[OPTION_BILL_MODE]) || !is_string($opt[OPTION_BILL_MODE]) || !in_array(strtoupper($opt[OPTION_BILL_MODE]), array(
            BillingInternalHelper::BILL_MODE_ALL,
            BillingInternalHelper::BILL_MODE_LAST
        ))
    ) {
        return false;
    }

    if (isset($opt[OPTION_DRY_RUN_MODE])) {
        if (!is_string($opt[OPTION_DRY_RUN_MODE]) || !in_array(strtoupper($opt[OPTION_DRY_RUN_MODE]), array(
                BillingInternalHelper::DRY_RUN_MODE_ALL,
                BillingInternalHelper::DRY_RUN_MODE_CH,
                BillingInternalHelper::DRY_RUN_MODE_BO
            ))
        ) {
            return false;
        }
    }

    return true;
}

/**
 * @param string $optionName
 * @param array $options
 * @return mixed | null
 */
function getOptions($optionName, $options)
{
    $option = null;
    switch ($optionName) {
        case OPTION_HELP:
            $option = isset($options[OPTION_HELP]);
            break;
        case OPTION_APPLICATION:
            $option = intval($options[OPTION_APPLICATION]);
            break;
        case OPTION_RESOURCES:
            $option = is_array($options[OPTION_RESOURCES]) ? $options[OPTION_RESOURCES] : array($options[OPTION_RESOURCES]);
            break;
        case OPTION_BILL_MODE:
            $option = strtoupper($options[OPTION_BILL_MODE]);
            break;
        case OPTION_DRY_RUN_MODE:
            $option = isset($options[OPTION_DRY_RUN_MODE]) ? strtoupper($options[OPTION_DRY_RUN_MODE]) : null;
            break;
        case OPTION_VERBOSE:
            if (isset($options[OPTION_VERBOSE])) {
                $option = is_array($options[OPTION_VERBOSE]) ? count($options[OPTION_VERBOSE]) : 1;
            } else {
                $option = 0;
            }
            break;
        default:
            echo sprintf(OUTPUT_OPTION_UNDEFINED, $optionName);
            printScriptUsage();
            exit;
    }
    return $option;
}

$opt = getopt(
    OPTION_APPLICATION . OPTION_PARSE_CHAR .
    OPTION_RESOURCES . OPTION_PARSE_CHAR .
    OPTION_BILL_MODE . OPTION_PARSE_CHAR .
    OPTION_DRY_RUN_MODE . OPTION_PARSE_CHAR .
    OPTION_VERBOSE .
    OPTION_HELP
);

if (!validateCommandLineParameters($opt)) {
    printScriptUsage();
    exit(1);
};

if (getOptions(OPTION_HELP, $opt)) {
    printScriptUsage();
    exit;
}

$applicationId = getOptions(OPTION_APPLICATION, $opt);
$resources = getOptions(OPTION_RESOURCES, $opt);
$billMode = getOptions(OPTION_BILL_MODE, $opt);
$dryRunMode = getOptions(OPTION_DRY_RUN_MODE, $opt);
$verbose = getOptions(OPTION_VERBOSE, $opt);

$output = Log::FIELD_TIME . ' ' . Log::FIELD_REQUEST . ' ' . Log::FIELD_TYPE . ' ' . Log::FIELD_MESSAGE . PHP_EOL;
$formatter = new LineFormatter($output);

$stream = new StreamHandler('php://stdout', Logger::NOTICE + $verbose*50);
$stream->setFormatter($formatter);
Log::getInstance()->getLogger()->pushHandler($stream);

try {
    $apiFactory = new ApiFactory();
    $helper = new BillingInternalHelper($apiFactory, $applicationId, $resources, $billMode, $dryRunMode);
    $helper->process();
} catch (Exception $e) {
    Log::getInstance()->error($e->getMessage());
    exit(1);
}
