<?php

/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */
class BillingInternalHelper
{
    const PROPERTY_CONTROLLER_URI = "controller_uri";
    const PROPERTY_APS_TOKEN = "aps_token";

    const APS_UNLIMITED_VALUE = -1;
    const APS_APPLICATIONS_PATH = '/aps/2/applications';
    const APS_TOKEN_HEADER = 'APS-Token: ';

    const BILLING_ORDER_NAME = 'Minimum commitment';

    const ORDER_TYPE_BO = 'BO';
    const ORDER_TYPE_CH = 'CH';
    const ORDER_TYPE_CL = 'CL';

    const ORDER_STATUS_NEW = 'NW';
    const ORDER_STATUS_OPEN = 'OP';

    const ORDER_DETAIL_TYPE_OVERUSAGE = 140;

    const PROVIDER_ACCOUNT_ID = 1;
    const PROVIDER_SUBSCRIPTION_ID = 1;

    const FREE_COMMITMENT_PERIOD_IN_MONTHS = 3;

    const DRY_RUN_MODE_ALL = 'ALL';
    const DRY_RUN_MODE_CH = 'CH';
    const DRY_RUN_MODE_BO = 'BO';

    const DRY_RUN_MESSAGE = 'DRY RUN MODE IS ENABLED. THE ORDER PLACING WILL BE SKIPPED.';

    const BILL_MODE_ALL = 'ALL';
    const BILL_MODE_LAST = 'LAST';

    const SUBSCRIPTION_STATUS_ORDERED = 10;
    const SUBSCRIPTION_STATUS_ACTIVE = 30;
    const SUBSCRIPTION_STATUS_TERMINATED = 60;

    const SUBSCRIPTION_SERVICE_STATUS_NOT_PROVISIONED = 10;
    const SUBSCRIPTION_SERVICE_STATUS_REMOVED = 90;

    const PERIOD_TYPE_DAYS = 1;

    /**
     * @var ApiFactory
     */
    private $apiFactory;

    /**
     * @var Service_OpenApi_Api
     */
    private $openApi;

    /**
     * @var Service_PbaApi_Api
     */
    private $pbaApi;

    /**
     * @var Model_OSA_AccountAttributesManagerInterface
     */
    private $accountAttributesManager;

    /**
     * @var Model_OSA_SubscriptionParametersManagerInterface
     */
    private $subscriptionParametersManager;

    /**
     * @var \APS\ControllerProxy
     */
    private $apsController;

    /**
     * @var int
     */
    private $applicationId;

    /**
     * @var DateTime
     */
    private $startPeriodDate;

    /**
     * @var DateTime
     */
    private $endPeriodDate;

    /**
     * @var array
     */
    private $instanceAvailableTiers = array();

    /**
     * @var string
     */
    private $billMode;

    /**
     * @var null|string
     */
    private $dryRunMode;

    /**
     * @var array
     */
    private $billableResources;

    /**
     * BillingHelper constructor.
     * @param ApiFactory $apiFactory
     * @param int $applicationId
     * @param array $billableResources
     * @param string $billMode
     * @param string | null $dryRunMode
     */
    public function __construct($apiFactory, $applicationId, $billableResources, $billMode = self::BILL_MODE_LAST, $dryRunMode = null)
    {
        $this->apiFactory = $apiFactory;
        $this->applicationId = $applicationId;
        $this->billableResources = $billableResources;
        $this->billMode = $billMode;
        $this->dryRunMode = $dryRunMode;

        $this->openApi = $this->createOpenApiInstance();
        $this->pbaApi = $this->createPbaApiInstance();

        $this->accountAttributesManager = new Model_OSA_AccountAttributesManager($this->pbaApi);
        $this->subscriptionParametersManager = new Model_OSA_SubscriptionParametersManager($this->pbaApi);
    }

    public function process()
    {
        $this->startPeriodDate = new DateTime('first day of previous month');
        $this->startPeriodDate->setTime(0, 0, 0);

        $this->endPeriodDate = new DateTime('last day of previous month');
        $this->endPeriodDate->setTime(23, 59, 59);

        Log::getInstance()->notice(
            'Processing subscriptions for the period from %s to %s.',
            $this->startPeriodDate->format('c'),
            $this->endPeriodDate->format('c')
        );

        $applicationInstances = $this->openApi->getApplicationInstances($this->applicationId);
        $apsApplicationInstances = $this->fetchApsApplicationInstances();
        foreach ($applicationInstances as $applicationInstance) {
            $appInstanceId = $applicationInstance['application_instance_id'];
            $appResourceId = $applicationInstance['application_resource_id'];
            $apsInstanceId = $this->findApsInstanceIdByResourceId($apsApplicationInstances, $appResourceId);
            Log::getInstance()->notice(str_repeat('=', 80));
            Log::getInstance()->notice('Processing subscriptions for application instance "%s".', $appInstanceId);
            try {
                $this->processApplicationInstance($appInstanceId, $apsInstanceId);
            } catch (Exception $e) {
                Log::getInstance()->error($e->getMessage());
            }
        }
        Log::getInstance()->notice(str_repeat('=', 80));
        Log::getInstance()->notice('The instances are processed.');
    }

    private function processApplicationInstance($appInstanceId, $apsInstanceId)
    {
        $this->apsController = $this->createApsControllerInstance($appInstanceId, $apsInstanceId);

        $this->instanceAvailableTiers = $this->fetchInstanceTiers();

        $limit = 100;
        $offset = 0;
        do {
            $accounts = $this->fetchAccounts($limit, $offset);
            if (!count($accounts) && !$offset) {
                Log::getInstance()->notice('There are no accounts to process.');
                return;
            }
            foreach ($accounts as $account) {
                try {
                    Log::getInstance()->notice(str_repeat('-', 80));
                    Log::getInstance()->notice('Processing subscriptions for account "%s".', $account->id);
                    $this->processAccount($account);
                } catch (Service_PbaApi_Api_Exception $e) {
                    $message = $e->getMessage();
                    $previous = $e->getPrevious();
                    if ($previous) {
                        $message .= ' ' . $previous->getMessage();
                    }
                    Log::getInstance()->error($message);
                } catch (Exception $e) {
                    Log::getInstance()->error($e->getMessage());
                }
            }
            $offset += $limit;
        } while (count($accounts) === $limit);
        Log::getInstance()->notice(str_repeat('-', 80));
        Log::getInstance()->notice('The accounts are processed.');
    }

    /**
     * @param $account
     * @throws Exception
     */
    private function processAccount($account)
    {
        $accountId = $account->id;
        $accountName = $account->companyName;

        $currentTier = $this->getMinimalCommitmentTierForAccount($accountId);
        if (is_null($currentTier)) {
            throw new Exception(Str::format(
                'Account "%s" does not have the account attribute "%s".',
                $accountId,
                Model_OSA_AccountAttributesManagerInterface::ATTRIBUTE_MINIMAL_COMMITMENT_TIER
            ));
        }

        $subscriptions = $this->fetchAccountSubscriptions($accountId, $currentTier->resourceTypeId);
        if (!count($subscriptions)) {
            Log::getInstance()->warning('The account "%s" has no subscriptions to process.', $accountId);
            return;
        }

        Log::getInstance()->notice('Getting currency for account "%s".', $accountId);
        $currencyInfo = $this->pbaApi->getVendorCurrency($accountId);
        $currencyName = $currencyInfo->getCurrencyID();
        Log::getInstance()->notice('Currency for account "%s" is "%s".', $accountId, $currencyName);

        Log::getInstance()->notice('Getting minimum commitment from tier "%s" by currency "%s".', $currentTier->resourceTypeId, $currencyName);
        $minimalCommitment = $this->getMinimalCommitmentForAccount($accountId, $currentTier, $currencyName);
        Log::getInstance()->notice('Minimum commitment for account "%s" is %g %s.', $accountId, $minimalCommitment, $currencyName);

        Log::getInstance()->notice('Calculating total amount for account "%s" by billing orders.', $accountId);
        $totalAmount = $this->calculateBillingOrderTotalAmount($subscriptions, $currencyName);
        Log::getInstance()->notice('Total amount for account "%s %s" is %g %s.', $accountId, $accountName, $totalAmount, $currencyName);

        $this->billAccountForMinimumCommitment($account, $totalAmount, $minimalCommitment, $accountName, $currentTier, $subscriptions);
        $this->switchAccountToNewTier($account, $totalAmount, $minimalCommitment, $currencyName, $currentTier, $subscriptions);
    }

    private function getMinimalCommitmentForAccount($accountId, $currentTier, $currency)
    {
        if ($currentTier->level == minimalCommitmentTier::LEVEL_FIXED_COMMITMENT) {
            $minimalCommitment = $this->accountAttributesManager->getAttributeMinimalCommitmentValue($accountId);
        } elseif ($currentTier->level == minimalCommitmentTier::LEVEL_NO_COMMITMENT) {
            $minimalCommitment = 0;
        } else {
            $minimalCommitment = $this->getMinimalCommitment($currentTier, $currency);
        }
        if (is_null($minimalCommitment)) {
            throw new Exception(Str::format('Minimum commitment tier "%s" does not have a value for currency "%s".', $currentTier->resourceTypeId, $currency));
        }
        return $minimalCommitment;
    }

    private function billAccountForMinimumCommitment($account, $totalAmount, $minimalCommitment, $currency, $currentTier, $subscriptions)
    {
        $accountId = $account->id;
        $accountName = $account->companyName;

        if ($currentTier->level == minimalCommitmentTier::LEVEL_NO_COMMITMENT) {
            Log::getInstance()->notice('Account "%s" has no commitment.', $accountId);
            return;
        }

        $mcStartDate = $this->getMinimalCommitmentStartDate($accountId);
        if (is_null($mcStartDate)) {
            Log::getInstance()->warning(
                'Account "%s" does not have the account attribute "%s". The account will be skipped.',
                $accountId,
                Model_OSA_AccountAttributesManagerInterface::ATTRIBUTE_MINIMAL_COMMITMENT_MONTH
            );
            return;
        }
        if ($mcStartDate->getTimestamp() > $this->endPeriodDate->getTimestamp()) {
            Log::getInstance()->notice(
                'Account "%s" has free period till %s inclusive.',
                $accountId,
                $mcStartDate->format(Model_OSA_AccountAttributesManagerInterface::MINIMAL_COMMITMENT_MONTH_FORMAT));
            return;
        }

        if ($totalAmount == $minimalCommitment) {
            Log::getInstance()->notice('Total amount for account "%s" equals to minimum commitment amount %g %s.', $accountId, $totalAmount, $currency);
            return;
        }
        if ($totalAmount > $minimalCommitment) {
            return;
        }

        $orderAmount = $minimalCommitment - $totalAmount;
        Log::getInstance()->notice('Place billing order for account "%s %s" on %g %s.', $accountId, $accountName, $orderAmount, $currency);
        if ($this->dryRunMode == BillingInternalHelper::DRY_RUN_MODE_ALL ||
            $this->dryRunMode == BillingInternalHelper::DRY_RUN_MODE_BO
        ) {
            Log::getInstance()->notice(self::DRY_RUN_MESSAGE);
            return;
        }
        $subscriptionId = min(array_map(function (Service_PbaApi_Api_Response_Struct_SubscriptionDetailsEx $subscription) {
            return $subscription->getSubscriptionId();
        }, $subscriptions));
        $orderInfo = $this->getSubscriptionMinimalCommitmentOrder($subscriptionId);
        if (is_null($orderInfo)) {
            $orderInfo = $this->placeBillingOrder($accountId, $subscriptionId, $currentTier->resourceTypeId, $orderAmount);
        } else {
            Log::getInstance()->notice(
                'A minimum commitment order is already placed for the period from %s to %s.',
                $this->startPeriodDate->format('Y-m-d'),
                $this->endPeriodDate->format('Y-m-d')
            );
        }
        Log::getInstance()->notice('%s %s', $orderInfo->getOrderNumber(), $orderInfo->getComments());
    }

    private function switchAccountToNewTier($account, $totalAmount, $minimalCommitment, $currency, $currentTier, $subscriptions)
    {
        $accountId = $account->id;

        $availableTiers = $this->fetchAvailableTiers($subscriptions);

        if ($currentTier->level < minimalCommitmentTier::LEVEL_NO_COMMITMENT) {
            Log::getInstance()->notice('Account "%s" has custom commitment. Account will not be moved to the next tier.', $accountId);
            return;
        }

        Log::getInstance()->notice('Find a next suitable tier for account "%s".', $accountId);
        $nextSuitableTier = $this->findNextSuitableTier($minimalCommitment, $totalAmount, $currency, $availableTiers);
        if (is_null($nextSuitableTier)) {
            Log::getInstance()->notice('A suitable tier for account "%s" is not found.', $accountId);
            return;
        }
        Log::getInstance()->notice('A suitable tier "%s" for account "%s" is found.', $nextSuitableTier->resourceTypeId, $accountId);

        $switchingIsAllowed = (bool)$this->accountAttributesManager->getAttributeEnableAutoSwitch($accountId);
        if (!$switchingIsAllowed) {
            Log::getInstance()->warning('Switching to the next tier is not allowed for the account "%s".', $accountId);
            return;
        }

        foreach ($subscriptions as $subscription) {
            try {
                $subscriptionId = $subscription->getSubscriptionId();

                // skip not active subscription
                if ($subscription->getStatus() != self::SUBSCRIPTION_STATUS_ACTIVE) {
                    Log::getInstance()->notice(
                        'Subscription %s is not active. A tier will not be changed for this subscription.',
                        $subscriptionId
                    );
                    continue;
                }

                // skip external subscriptions
                $isExternal = $this->subscriptionParametersManager->getParameterIsExternal($subscriptionId);
                if ($isExternal) {
                    Log::getInstance()->notice(
                        'Subscription %s is external. A tier will not be changed for this subscription.',
                        $subscriptionId
                    );
                    continue;
                }

                $this->switchSubscriptionToNewTier($accountId, $subscriptionId, $nextSuitableTier->resourceTypeId);
            } catch (Service_PbaApi_Api_Exception $e) {
                $message = $e->getMessage();
                $previous = $e->getPrevious();
                if ($previous) {
                    $message .= ' ' . $previous->getMessage();
                }
                Log::getInstance()->error($message);
            } catch (Exception $e) {
                Log::getInstance()->error($e->getMessage());
            }
        }
    }

    /**
     * @param int $appInstanceId
     * @param string $apsInstanceId
     * @return \APS\ControllerProxy
     */
    private function createApsControllerInstance($appInstanceId, $apsInstanceId)
    {
        $token = $this->openApi->getApplicationInstanceToken($appInstanceId);
        return $this->apiFactory->getApsController(
            $token[self::PROPERTY_CONTROLLER_URI],
            null,
            $apsInstanceId
        );
    }

    /**
     * @return Service_PbaApi_Api
     */
    private function createPbaApiInstance()
    {
        return $this->apiFactory->getPbaApi(Settings::getInstance()->getBillingConfig());
    }

    /**
     * @return Service_OpenApi_Api
     */
    private function createOpenApiInstance()
    {
        return $this->apiFactory->getOpenApi(OPEN_API_URI);
    }

    /**
     * @param int $limit
     * @param int $offset
     * @return \APS\ResourceBase[]
     */
    private function fetchAccounts($limit = 1000, $offset = 0)
    {
        $filter = RqlBuilder::createInstance()
            ->implementing('http://parallels.com/aps/types/pa/account/1.0')
            ->eq('aps.status', Service_APSApi_Api::APS_READY)
            ->eq('type', 'CLIENT')
            ->limit($offset, $limit)
            ->getQuery();
        return $this->apsController->getResources($filter);
    }

    /**
     * @param array $subscriptionResources
     * @param string $apsType
     * @param string|null $apsProperty
     * @return array
     */
    private function filterSubscriptionResources(array $subscriptionResources, $apsType, $apsProperty = null)
    {
        return array_filter($subscriptionResources, function ($resource) use ($apsType, $apsProperty) {
            $typeIsMatched = isset($resource->apsType) && $resource->apsType === $apsType;
            $propertyIsMatched = is_null($apsProperty)
                ? !isset($resource->property)
                : isset($resource->property) && $resource->property === $apsProperty;

            return $typeIsMatched && $propertyIsMatched;
        });
    }

    /**
     * @param array $subscriptionResources
     * @param int | null $limit
     * @return array
     */
    private function filterSubscriptionResourcesByLimitNotEqual(array $subscriptionResources, $limit)
    {
        return array_filter($subscriptionResources, function ($resource) use ($limit) {
            $resourceLimit = isset($resource->limit) ? $resource->limit : BillingInternalHelper::APS_UNLIMITED_VALUE;
            return $resourceLimit !== $limit;
        });
    }

    /**
     * @return array
     */
    private function fetchInstanceTiers()
    {
        $filter = RqlBuilder::createInstance()
            ->implementing(\APS\TypeLibrary::getTypeByServiceId('minimalCommitmentTier'))
            ->getQuery();
        $resources = $this->apsController->getResources($filter);
        $tiers = array();
        foreach ($resources as $tier) {
            $id = $tier->resourceTypeId;
            $tiers[$id] = $tier;
        }
        return $tiers;
    }

    /**
     * @param \APS\ResourceBase $tier
     * @param string $currency
     * @return int|null
     */
    private function getMinimalCommitment($tier, $currency)
    {
        if (!is_array($tier->minimalCommitments)) {
            return null;
        }
        foreach ($tier->minimalCommitments as $minimalCommitment) {
            if ($minimalCommitment->currency === $currency) {
                return $minimalCommitment->value;
            }
        }
        return null;
    }

    /**
     * @param Service_PbaApi_Api_Response_Struct_SubscriptionDetailsEx[] $subscriptions
     * @param string $currency
     * @return float|int
     * @throws Exception
     */
    private function calculateBillingOrderTotalAmount(array $subscriptions, $currency)
    {
        $totalAmount = 0;
        foreach ($subscriptions as $subscription) {
            $subscriptionId = $subscription->getSubscriptionId();
            $billingOrders = $this->fetchSubscriptionBillingOrders($subscriptionId);
            if (!count($billingOrders)) {
                throw new Exception(Str::format('Subscription "%s" does not have billing orders.', $subscriptionId));
            }
            $totalAmount += $this->calculateBillingOrderTotalAmountForSubscription($subscriptionId, $currency, $billingOrders);
        }
        return $totalAmount;
    }

    /**
     * @param int $subscriptionId
     * @param string $currency
     * @param Service_PbaApi_Api_Response_Struct_GetOrder[] $billingOrders
     * @return float|int
     */
    private function calculateBillingOrderTotalAmountForSubscription($subscriptionId, $currency, array $billingOrders)
    {
        $startPeriodTime = $this->startPeriodDate->getTimestamp();
        $endPeriodTime = $this->endPeriodDate->getTimestamp();

        $discount = $this->getSubscriptionDiscount($subscriptionId);
        $precision = $this->getCurrencyPrecision($currency);

        $totalAmount = 0;
        foreach ($billingOrders as $orderInfo) {
            $details = $this->pbaApi->getOrderDetails($orderInfo->getOrderId());
            foreach ($details as $detail) {
                if ($detail->getOrderDetailStartDate() < $startPeriodTime ||
                    $detail->getOrderDetailEndDate() - 1 > $endPeriodTime
                ) {
                    continue;
                }
                if ($detail->getSubscription() != $subscriptionId) {
                    continue;
                }
                if ($detail->getDetailType() != self::ORDER_DETAIL_TYPE_OVERUSAGE) {
                    continue;
                }
                if (!in_array($detail->getResourceID(), $this->billableResources)) {
                    continue;
                }

                $duration = ($this->billMode === self::BILL_MODE_LAST) ? 1.0 : $detail->getDuration();
                $totalAmount += round(
                    $duration * $detail->getQuantity() * $detail->getUnitPrice() * (1 - $discount),
                    $precision
                );
            }
        }
        return $totalAmount;
    }

    /**
     * @param int $subscriptionId
     * @return Service_PbaApi_Api_Response_Struct_GetOrder[]
     * @throws Exception
     */
    private function fetchSubscriptionBillingOrders($subscriptionId)
    {
        // sort by order date (-4) is important
        $subscriptionOrdersList = $this->pbaApi->getSubscriptionSalesOrdersList($subscriptionId, -4);

        $startPeriodTime = $this->startPeriodDate->getTimestamp();
        $endPeriodTime = $this->endPeriodDate->getTimestamp();

        $billingOrders = array();
        foreach ($subscriptionOrdersList as $listItem) {
            $orderTime = $listItem->getOrderDate();
            if ($orderTime < $startPeriodTime) {
                break;
            }
            if (($orderTime - 1) > $endPeriodTime) {
                continue;
            }

            $pattern = '/^(' . self::ORDER_TYPE_BO . '|' . self::ORDER_TYPE_CH . '|' . self::ORDER_TYPE_CL . ')\d+$/';
            if (!preg_match($pattern, $listItem->getOrderNumber())) {
                continue;
            }

            $order = $this->pbaApi->getOrder($listItem->getOrderId());
            if ($order->getComments() === self::BILLING_ORDER_NAME) {
                continue;
            }

            $billingOrders[] = $order;

            if ($this->billMode === self::BILL_MODE_LAST) {
                break;
            }
        }
        return $billingOrders;
    }

    /**
     * @param int $accountId
     * @param int $subscriptionId
     * @param float $amount
     * @return Service_PbaApi_Api_Response_Struct_GetOrder
     */
    private function placeBillingOrder($accountId, $subscriptionId, $resourceId, $amount)
    {
        $orderAttributes = new Service_PbaApi_Api_Request_Struct_OrderAttributes();
        $orderAttributes->setAccountId($accountId);
        $orderAttributes->setOrderTypeId(self::ORDER_TYPE_BO);
        $orderAttributes->setOrderStatusId(self::ORDER_STATUS_NEW);
        $orderAttributes->setConsolidationType(10);
        $orderAttributes->setFromIp($this->fetchEndPointHost());
        $orderAttributes->setDescription(self::BILLING_ORDER_NAME);
        $orderAttributes->setOrderDate($this->endPeriodDate->getTimestamp());

        $orderDetail = new Service_PbaApi_Api_Request_Struct_OrderDetailAttributes();
        $orderDetail->setExtendedPrice(0);
        $orderDetail->setTaxAmount(0);
        $orderDetail->setTaxIncludedAmount(0);
        $orderDetail->setTaxAlgorithm(0);
        $orderDetail->setUnitPrice($amount);
        $orderDetail->setServiceQuantity(1);
        $orderDetail->setServiceUnitMeasure('-');
        $orderDetail->setDuration(1);
        $orderDetail->setDurationBillPeriod(1);
        $orderDetail->setDurationBillPeriodType(4);
        $orderDetail->setSubscriptionId($subscriptionId);
        $orderDetail->setResourceId($resourceId);
        $orderDetail->setDescription(self::BILLING_ORDER_NAME);

        $orderId = $this->pbaApi->createManualOrder($orderAttributes, array($orderDetail));
        return $this->pbaApi->getOrder($orderId);
    }

    private function fetchEndPointHost()
    {
        $fetcher = new EndpointHostIpFetcher($this->apsController);
        return $fetcher->fetchApsEndpointIpAddressOrFallback();
    }

    private function placePlanSwitchOrder($subscriptionId, $planId, $planPeriodId)
    {
        return $this->pbaApi->placePlanPeriodSwitchOrder($subscriptionId, $planId, $planPeriodId, 20);
    }

    /**
     * @param $planId
     * @param $periodDuration
     * @param $periodType
     * @return null|Service_PbaApi_Api_Response_Struct_PlanPeriodListItem
     */
    private function findPlanPeriod($planId, $periodDuration, $periodType)
    {
        $periods = $this->pbaApi->getPlanPeriodList($planId);
        foreach ($periods as $period) {
            if ($period->getTrial() || !$period->getEnabled()) {
                continue;
            }
            if ($period->getPeriod() == $periodDuration && $period->getPeriodType() == $periodType) {
                return $period;
            }
        }
        return null;
    }

    private function findPlanRate($planId, $resourceTypeId)
    {
        $ratesList = $this->pbaApi->getPlanRateListFull($planId);
        foreach ($ratesList as $rateInfo) {
            if ($rateInfo->getResourceId() == $resourceTypeId && $rateInfo->getIncludedValue() != 0) {
                return $rateInfo;
            }
        }
        return null;
    }

    /**
     * @param $currentMinimalCommitment
     * @param $totalAmount
     * @param $currency
     * @param array $availableTiers
     * @return \APS\ResourceBase|null
     * @throws Exception
     */
    private function findNextSuitableTier($currentMinimalCommitment, $totalAmount, $currency, array $availableTiers)
    {
        $values = array();
        foreach ($availableTiers as $availableTier) {
            if ($availableTier->level < 0) {
                continue;
            }
            $value = $this->getMinimalCommitment($availableTier, $currency);
            if (is_null($value)) {
                $message = Str::format('Minimum commitment tier "%s" does not have a value for currency "%s".', $availableTier->name, $currency);
                Log::getInstance()->error($message);
                throw new Exception($message);
            }
            if ($value <= $currentMinimalCommitment || $value > $totalAmount) {
                continue;
            }
            $values[$value] = $availableTier;
        }
        if (empty($values)) {
            return null;
        }
        $value = max(array_keys($values));
        return $values[$value];
    }

    /**
     * @param $subscriptionId
     * @return Service_PbaApi_Api_Response_Struct_PlanDetails[]
     */
    private function fetchSubscriptionAvailablePlans($subscriptionId)
    {
        $plans = array();
        $plansList = $this->pbaApi->getSubscriptionAvailablePlanList($subscriptionId);
        foreach ($plansList as $plansListItem) {
            $plans [] = $this->pbaApi->getPlanDetails($plansListItem->getPlanId());
        }
        return $plans;
    }

    /**
     * @param Service_PbaApi_Api_Response_Struct_GetOrder $order
     * @param string $status
     * @return Service_PbaApi_Api_Response_Struct_Status
     */
    private function changeOrderStatus($order, $status)
    {
        Log::getInstance()->notice('Changing status for order "%s" to "%s".', $order->getOrderID(), $status);
        $result = $this->pbaApi->changeOrderStatus(
            $order->getOrderID(),
            $status,
            md5(
                $order->getOrderID() .
                trim($order->getOrderNumber()) .
                $order->getCreationTime() .
                sprintf('%s %.2f', trim($order->getCurrencyID()), $order->getTotal()) .
                $order->getComments()
            )
        );
        Log::getInstance()->notice('Result of status changing is "%s".', $result->getStatus());
        return $result;
    }

    /**
     * @param int $orderId
     * @return null|Service_PbaApi_Api_Response_Struct_Status
     */
    private function openOrder($orderId)
    {
        $order = $this->pbaApi->getOrder($orderId);
        if ($order->getOrderStatusID() != self::ORDER_STATUS_NEW) {
            Log::getInstance()->notice('The status of the order "%s" is already "%s".', $orderId, $order->getOrderStatusID());
            return null;
        }
        return $this->changeOrderStatus($order, self::ORDER_STATUS_OPEN);
    }

    private function fetchApsApplicationInstances()
    {
        $token = $this->openApi->getAccountToken(self::PROVIDER_ACCOUNT_ID, self::PROVIDER_SUBSCRIPTION_ID);

        $url = $token[self::PROPERTY_CONTROLLER_URI] . self::APS_APPLICATIONS_PATH;
        $headers = array(self::APS_TOKEN_HEADER . $token[self::PROPERTY_APS_TOKEN]);

        $applications = json_decode($this->doHttpGetRequest($url, $headers));

        return array_filter($applications, function ($application) {
            return $application->aps->type === APSResource::PACKAGE_ID;
        });
    }

    /**
     * @param string $url
     * @param array $headers
     * @return mixed
     * @throws Exception
     */
    private function doHttpGetRequest($url, array $headers = array())
    {
        Log::getInstance()->event('Request GET %s', $url);
        $curl = new Curl();
        $curl->setOption(CURLOPT_URL, $url);
        $curl->setOption(CURLOPT_HTTPHEADER, $headers);
        $curl->setOption(CURLOPT_RETURNTRANSFER, true);
        $curl->setOption(CURLOPT_FOLLOWLOCATION, true);
        $curl->setOption(CURLOPT_SSL_VERIFYPEER, false);
        $curl->setOption(CURLOPT_SSL_VERIFYHOST, false);
        $return = $curl->execute();
        $errorNo = $curl->getErrorNo();
        if ($errorNo) {
            $curl->refresh();
            throw new Exception($curl->getError(), $errorNo);
        }
        $code = $curl->getInfo(CURLINFO_HTTP_CODE);
        Log::getInstance()->event('Response %s %s', $code, $return);
        if ($code < 200 || $code >= 300) {
            $curl->refresh();
            throw new Exception($return, $code);
        }
        $curl->refresh();
        return $return;
    }

    /**
     * @param $instances
     * @param $resourceId
     * @return array
     */
    private function findApsInstanceIdByResourceId($instances, $resourceId)
    {
        $instances = array_filter($instances, function ($application) use ($resourceId) {
            return $application->globals->aps->id === $resourceId;
        });
        $instance = reset($instances);
        return $instance->aps->id;
    }

    /**
     * @param Service_PbaApi_Api_Response_Struct_SubscriptionDetailsEx[] $subscriptions
     * @return array
     * @throws Exception
     */
    private function fetchAvailableTiers(array $subscriptions)
    {
        $activeTierId = null;
        $availableTiersIds = array();
        $isFirstSubscription = true;
        foreach ($subscriptions as $subscriptionDetails) {
            $subscriptionId = $subscriptionDetails->getSubscriptionId();
            if ($subscriptionDetails->getStatus() != self::SUBSCRIPTION_STATUS_ACTIVE) {
                continue;
            }

            $subscriptionResources = $this->fetchSubscriptionResources($subscriptionId);
            $subscriptionAvailableTiersResources = $this->filterSubscriptionResources(
                $subscriptionResources,
                \APS\TypeLibrary::getTypeByServiceId('minimalCommitmentTier')
            );
            if (!count($subscriptionAvailableTiersResources)) {
                throw new Exception(Str::format(
                    'Subscription "%s" does not support minimum commitment tiers.',
                    $subscriptionId
                ));
            }

            $subscriptionAvailableTiersIds = array_map(function ($resource) {
                return $resource->id;
            }, $subscriptionAvailableTiersResources);
            sort($subscriptionAvailableTiersIds);

            if (!$isFirstSubscription && $subscriptionAvailableTiersIds != $availableTiersIds) {
                throw new Exception(Str::format(
                    'The subscription "%s" has a different sequence of minimum commitment tiers.',
                    $subscriptionId
                ));
            }

            $subscriptionActiveTiersResources = $this->filterSubscriptionResourcesByLimitNotEqual($subscriptionAvailableTiersResources, 0);
            $subscriptionActiveTiersCount = count($subscriptionActiveTiersResources);
            if (!$subscriptionActiveTiersCount) {
                throw new Exception(Str::format(
                    'Subscription "%s" does not have an enabled minimum commitment tier.',
                    $subscriptionId
                ));
            }
            if ($subscriptionActiveTiersCount > 1) {
                throw new Exception(Str::format(
                    'Subscription "%s" has multiple enabled minimum commitment tiers.',
                    $subscriptionId
                ));
            }
            $subscriptionActiveTierResource = reset($subscriptionActiveTiersResources);
            $subscriptionActiveTierId = $subscriptionActiveTierResource->id;

            if (!$isFirstSubscription && $subscriptionActiveTierId != $activeTierId) {
                throw new Exception(Str::format(
                    'The subscription "%s" has a different enabled minimum commitment tier.',
                    $subscriptionId
                ));
            }

            if ($isFirstSubscription) {
                $activeTierId = $subscriptionActiveTierId;
                $availableTiersIds = $subscriptionAvailableTiersIds;
                $isFirstSubscription = false;
            }
        }

        $availableTiers = array();
        foreach ($availableTiersIds as $tierId) {
            $availableTiers[] = $this->getTierResource($tierId);
        }
        return $availableTiers;
    }

    /**
     * @param int $accountId
     * @param int $subscriptionId
     * @param int $newTierId
     * @throws Exception
     */
    private function switchSubscriptionToNewTier($accountId, $subscriptionId, $newTierId)
    {
        Log::getInstance()->notice('Searching for a next suitable plan for subscription "%s" by tier "%s".', $subscriptionId, $newTierId);
        $suitablePlans = $this->findSuitablePlansToUpgrade($subscriptionId, $newTierId);
        if (empty($suitablePlans)) {
            throw new Exception(Str::format('A suitable plan is not found for subscription "%s".', $subscriptionId));
        }
        if (count($suitablePlans) > 1) {
            throw new Exception(Str::format('Subscription "%s" has several suitable plans.', $subscriptionId));
        }

        $suitablePlan = reset($suitablePlans);
        /**
         * @var Service_PbaApi_Api_Response_Struct_PlanDetails $planDetails
         */
        $planDetails = $suitablePlan['details'];
        /**
         * @var Service_PbaApi_Api_Response_Struct_PlanPeriodListItem $planPeriod
         */
        $planPeriod = $suitablePlan['period'];


        $planId = $planDetails->getPlanId();
        $planName = $planDetails->getName();
        $planPeriodId = $planPeriod->getPlanPeriodID();

        Log::getInstance()->notice('Placing the order for re-assigning account "%s" to plan "%s - %s".', $accountId, $planId, $planName);
        if ($this->dryRunMode == BillingInternalHelper::DRY_RUN_MODE_ALL ||
            $this->dryRunMode == BillingInternalHelper::DRY_RUN_MODE_CH
        ) {
            Log::getInstance()->notice(self::DRY_RUN_MESSAGE);
            return;
        }
        $orderInfo = $this->placePlanSwitchOrder($subscriptionId, $planId, $planPeriodId);
        Log::getInstance()->notice('%s %s', $orderInfo->getOrderNumber(), $orderInfo->getDescription());

        $this->openOrder($orderInfo->getOrderId());
    }

    /**
     * @param int $accountId
     * @return \APS\ResourceBase
     * @throws Exception
     */
    private function getMinimalCommitmentTierForAccount($accountId)
    {
        $resourceTypeId = $this->accountAttributesManager->getAttributeMinimalCommitmentTier($accountId);
        if (empty($resourceTypeId)) {
            return null;
        }

        $tier = $this->getTierResource($resourceTypeId);
        if (!$tier) {
            throw new Exception(Str::format(
                'There are no minimum commitment tiers with id "%s".',
                $resourceTypeId
            ));
        }
        return $tier;
    }

    /**
     * @param int $accountId
     * @throws Exception
     */
    private function setMinimalCommitmentTierForAccount($accountId, $resourceTypeId)
    {
        $this->accountAttributesManager->setAttributeMinimalCommitmentTier($accountId, $resourceTypeId);
    }

    /**
     * @param int $accountId
     * @return DateTime|null
     * @throws Exception
     */
    private function getMinimalCommitmentStartDate($accountId)
    {
        $minComMonth = $this->accountAttributesManager->getAttributeMinimalCommitmentMonth($accountId);
        if (empty($minComMonth)) {
            return null;
        }

        if (!preg_match('/^(\d+)-(\d+)$/', $minComMonth, $matches)) {
            throw new Exception(
                Str::format(
                    'Account\'s (%s) attribute "%s" has invalid format.',
                    $accountId,
                    Model_OSA_AccountAttributesManagerInterface::ATTRIBUTE_MINIMAL_COMMITMENT_MONTH
                )
            );
        }

        $date = new DateTime();
        $date->setDate($matches[2], $matches[1], 1);
        $date->setTime(0, 0, 0);

        return $date;
    }

    /**
     * @param $subscriptionId
     * @return float
     * @throws Service_PbaApi_Api_Exception
     */
    private function getSubscriptionDiscount($subscriptionId)
    {
        $percent = $this->subscriptionParametersManager->getParameterDiscount($subscriptionId);
        return $percent / 100;
    }

    private function getCurrencyPrecision($currency)
    {
        if (in_array($currency, array('AOA', 'CNY', 'HKD', 'MOP', 'TWD'))) {
            return 1;
        }
        if (in_array($currency, array('BHD', 'JOD', 'KWD', 'LYD', 'OMR', 'TND'))) {
            return 3;
        }
        if (in_array($currency, array(
            'AMD', 'BIF', 'BYR', 'CLP', 'COP', 'DJF', 'GNF', 'HUF', 'IDR', 'IQD', 'IRR',
            'ISK', 'JPY', 'KHR', 'KMF', 'KPW', 'KRW', 'LAK', 'LBP', 'MMK', 'PYG', 'RWF',
            'SLL', 'STD', 'UGX', 'VND', 'VUV', 'XAF', 'XOF', 'XPF', 'YER', 'ZMK'
        ))) {
            return 0;
        }
        return 2;
    }

    /**
     * @param $subscriptionId
     * @param $resourceTypeId
     * @return array
     * @throws Exception
     */
    private function findSuitablePlansToUpgrade($subscriptionId, $resourceTypeId)
    {
        $plans = $this->fetchSubscriptionAvailablePlans($subscriptionId);
        if (empty($plans)) {
            return $plans;
        }

        $subscriptionDetails = $this->pbaApi->getSubscriptionDetailsEx($subscriptionId);
        $currentPlanDetails = $this->pbaApi->getPlanDetails($subscriptionDetails->getPlanId());
        $currentCategoryId = $currentPlanDetails->getCategoryId();
        $periodDuration = $subscriptionDetails->getPeriod();
        $periodType = $subscriptionDetails->getPeriodType();

        $suitablePlans = array();
        foreach ($plans as $plan) {
            if ($plan->getCategoryId() != $currentCategoryId) {
                continue;
            }

            $period = $this->findPlanPeriod($plan->getPlanId(), $periodDuration, $periodType);
            if (is_null($period)) {
                continue;
            }

            $rate = $this->findPlanRate($plan->getPlanId(), $resourceTypeId);
            if (empty($rate)) {
                continue;
            }
            $suitablePlans [] = array(
                'details' => $plan,
                'period' => $period,
                'rate' => $rate,
            );
        }
        return $suitablePlans;
    }

    /**
     * @param $subscriptionId
     * @return Service_PbaApi_Api_Response_Struct_GetOrder|null
     */
    private function getSubscriptionMinimalCommitmentOrder($subscriptionId)
    {
        // sort by order date (-4) is important
        $subscriptionOrdersList = $this->pbaApi->getSubscriptionSalesOrdersList($subscriptionId, -4);

        $startPeriodTime = $this->startPeriodDate->getTimestamp();
        $endPeriodTime = $this->endPeriodDate->getTimestamp();

        foreach ($subscriptionOrdersList as $listItem) {
            $orderTime = $listItem->getOrderDate();
            if ($orderTime < $startPeriodTime) {
                break;
            }

            if ($orderTime > $endPeriodTime) {
                continue;
            }

            $pattern = '/^' . self::ORDER_TYPE_BO . '\d+$/';
            if (!preg_match($pattern, $listItem->getOrderNumber())) {
                continue;
            }

            $order = $this->pbaApi->getOrder($listItem->getOrderId());
            if ($order->getComments() === self::BILLING_ORDER_NAME) {
                return $order;
            }
        }
        return null;
    }

    /**
     * @param $accountId
     * @param $accountTierId
     * @return Service_PbaApi_Api_Response_Struct_SubscriptionDetailsEx[]
     * @throws Exception
     */
    private function fetchAccountSubscriptions($accountId, $accountTierId)
    {
        $subscriptionList = $this->pbaApi->getCustomerSubscriptionList($accountId);

        $subscriptionIds = array_map(function ($subscriptionItem) {
            return $subscriptionItem->getSubscriptionId();
        }, $subscriptionList);

        Log::getInstance()->notice('Account\'s subscriptions: %s.', implode(', ', $subscriptionIds));

        $startPeriodDate = $this->startPeriodDate->getTimestamp();
        $subscriptionParametersManager = $this->subscriptionParametersManager;

        $subscriptions = array();
        foreach ($subscriptionList as $subscriptionItem) {
            $subscriptionId = $subscriptionItem->getSubscriptionId();
            // skip trial subscriptions
            if ($subscriptionItem->getPeriodType() == self::PERIOD_TYPE_DAYS) {
                Log::getInstance()->notice(
                    'Subscription %s has trial period. The subscription will be skipped.',
                    $subscriptionId
                );
                continue;
            }

            // skip not provisioned or terminated subscriptions
            $status = $subscriptionItem->getStatus();
            $serviceStatus = $subscriptionItem->getServiceStatus();
            if (
                (
                    $status == self::SUBSCRIPTION_STATUS_ORDERED &&
                    $serviceStatus == self::SUBSCRIPTION_SERVICE_STATUS_NOT_PROVISIONED
                ) || (
                    $status == self::SUBSCRIPTION_STATUS_TERMINATED &&
                    (
                        $serviceStatus == self::SUBSCRIPTION_SERVICE_STATUS_NOT_PROVISIONED ||
                        $serviceStatus == self::SUBSCRIPTION_SERVICE_STATUS_REMOVED
                    )
                )
            ) {
                Log::getInstance()->notice(
                    'Subscription %s is Not Provisioned or Terminated. The subscription will be skipped.',
                    $subscriptionId
                );
                continue;
            }

            $subscriptionDetails = $this->pbaApi->getSubscriptionDetailsEx($subscriptionId);

            // skip not billed subscriptions
            if ($subscriptionDetails->getLastBillDate() < $startPeriodDate) {
                Log::getInstance()->notice(
                    'Subscription %s was not billed in this period. The subscription will be skipped.',
                    $subscriptionId
                );
                continue;
            }

            // skip excluded subscriptions
            $needExclude = $subscriptionParametersManager->getParameterExcludedFromReport($subscriptionId);
            if ($needExclude != Model_OSA_SubscriptionParametersManager::EXCLUDE_FROM_REPORT_OFF) {
                Log::getInstance()->notice(
                    'Subscription %s is excluded from report. The subscription will be skipped.',
                    $subscriptionId
                );
                continue;
            }

            // skip not tear subscriptions
            $tierResources = array_keys($this->instanceAvailableTiers);
            $ratesList = $this->pbaApi->getPlanRateListFull($subscriptionDetails->getPlanId());
            $tierResourceRates = array_filter($ratesList, function ($rateInfo) use ($tierResources) {
                return in_array($rateInfo->getResourceId(), $tierResources);
            });
            if (empty($tierResourceRates)) {
                Log::getInstance()->notice(
                    'Subscription %s does not have minimal commitment tier resources. The subscription will be skipped.',
                    $subscriptionId
                );
                continue;
            }

            $activeTierResourceRates = array_filter($tierResourceRates, function ($rateInfo) {
                return $rateInfo->getIncludedValue() != 0;
            });
            if (empty($activeTierResourceRates)) {
                throw new Exception(Str::format(
                    'Subscription "%s" does not have an enabled minimum commitment tier.',
                    $subscriptionId
                ));
            }
            if (count($activeTierResourceRates) > 1) {
                throw new Exception(Str::format(
                    'Subscription "%s" has multiple enabled minimum commitment tiers.',
                    $subscriptionId
                ));
            }

            $subscriptionActiveTierResourceRate = reset($activeTierResourceRates);
            if ($subscriptionActiveTierResourceRate->getResourceId() != $accountTierId) {
                throw new Exception(Str::format(
                    'Subscription "%s" is on tier "%s" while account is on tier "%s".',
                    $subscriptionId,
                    $subscriptionActiveTierResourceRate->getResourceId(),
                    $accountTierId
                ));
            }
            $subscriptions[$subscriptionId] = $subscriptionDetails;
        }
        return $subscriptions;
    }

    private function getTierResource($resourceId)
    {
        return isset($this->instanceAvailableTiers[$resourceId])
            ? $this->instanceAvailableTiers[$resourceId]
            : null;
    }

    private function fetchSubscriptionResources($subscriptionId)
    {
        $filter = RqlBuilder::createInstance()
            ->implementing('http://parallels.com/aps/types/pa/subscription/1.0')
            ->eq('aps.status', Service_APSApi_Api::APS_READY)
            ->eq('subscriptionId', $subscriptionId)
            ->getQuery();
        $resources = $this->apsController->getResources($filter);
        if (empty($resources)) {
            throw new Exception(Str::format(
                'There is no APS resource for subscription "%s".',
                $subscriptionId
            ));
        }
        if (count($resources) > 1) {
            throw new Exception(Str::format(
                'There are several APS resources for subscription "%s".',
                $subscriptionId
            ));
        }
        $subscription = reset($resources);
        return $subscription->resources();
    }
}