<?php

/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */
class ReportHelper
{
    const PROPERTY_CONTROLLER_URI = "controller_uri";
    const PROPERTY_APS_TOKEN = "aps_token";

    const APS_APPLICATIONS_PATH = '/aps/2/applications';
    const APS_TOKEN_HEADER = 'APS-Token: ';

    const PROVIDER_ACCOUNT_ID = 1;
    const PROVIDER_SUBSCRIPTION_ID = 1;

    const WHEN_OFFSET_END_DATE = '+2 hours';
    const DEFAULT_TIMEZONE = 'UTC';

    /**
     * @var ApiFactory
     */
    private $apiFactory;

    /**
     * @var Service_OpenApi_Api
     */
    private $openApi;

    /**
     * @var \APS\ControllerProxy
     */
    private $apsController;

    /**
     * @var int
     */
    private $applicationId;


    /**
     * BillingHelper constructor.
     * @param ApiFactory $apiFactory
     * @param int $applicationId
     * @param boolean $dryRunMode
     */
    public function __construct($apiFactory, $applicationId, $dryRunMode = false)
    {
        $this->apiFactory = $apiFactory;
        $this->applicationId = $applicationId;
        $this->dryRunMode = $dryRunMode;

        $this->openApi = $this->createOpenApiInstance();
    }


    public function process()
    {
        $applicationInstances = $this->openApi->getApplicationInstances($this->applicationId);
        $apsApplicationInstances = $this->fetchApsApplicationInstances();
        foreach ($applicationInstances as $applicationInstance) {
            $appInstanceId = $applicationInstance['application_instance_id'];
            $appResourceId = $applicationInstance['application_resource_id'];
            $apsInstanceId = $this->findApsInstanceIdByResourceId($apsApplicationInstances, $appResourceId);
            Log::getInstance()->notice(str_repeat('=', 80));
            Log::getInstance()->notice('Processing datacenters for application instance "%s".', $appInstanceId);
            try {
                $this->processApplicationInstance($appInstanceId, $apsInstanceId);
            } catch (Exception $e) {
                Log::getInstance()->error($e->getMessage());
            }
        }
        Log::getInstance()->notice(str_repeat('=', 80));
        Log::getInstance()->notice('The instances are processed.');
    }


    /**
     * @param int $appInstanceId
     * @param string $apsInstanceId
     * @return \APS\ControllerProxy
     */
    private function createApsControllerInstance($appInstanceId, $apsInstanceId)
    {
        $token = $this->openApi->getApplicationInstanceToken($appInstanceId);
        return $this->apiFactory->getApsController(
            $token[self::PROPERTY_CONTROLLER_URI],
            null,
            $apsInstanceId
        );
    }


    /**
     * @return Service_OpenApi_Api
     */
    private function createOpenApiInstance()
    {
        return $this->apiFactory->getOpenApi(OPEN_API_URI);
    }


    private function processApplicationInstance($appInstanceId, $apsInstanceId)
    {
        $this->apsController = $this->createApsControllerInstance($appInstanceId, $apsInstanceId);

        $query = RqlBuilder::createInstance()->implementing(\APS\TypeLibrary::getTypeByServiceId('datacenter'))->getQuery();
        $datacenters = $this->apsController->getResources($query);
        foreach ($datacenters as $datacenter) {
            try {
                $this->requestReportForDatacenter($apsInstanceId, $datacenter);
            } catch (Exception $e) {
                Log::getInstance()->notice($e->getMessage());
            }
        }

        Log::getInstance()->notice(str_repeat('-', 80));
        Log::getInstance()->notice('The datacenters are processed.');
    }


    private function fetchApsApplicationInstances()
    {
        $token = $this->openApi->getAccountToken(self::PROVIDER_ACCOUNT_ID, self::PROVIDER_SUBSCRIPTION_ID);

        $url = $token[self::PROPERTY_CONTROLLER_URI] . self::APS_APPLICATIONS_PATH;
        $headers = array(self::APS_TOKEN_HEADER . $token[self::PROPERTY_APS_TOKEN]);

        $applications = json_decode($this->doHttpGetRequest($url, $headers));

        return array_filter($applications, function ($application) {
            return $application->aps->type === APSResource::PACKAGE_ID;
        });
    }

    /**
     * @param string $url
     * @param array $headers
     * @return mixed
     * @throws Exception
     */
    private function doHttpGetRequest($url, array $headers = array())
    {
        Log::getInstance()->event('Request GET %s', $url);
        $curl = new Curl();
        $curl->setOption(CURLOPT_URL, $url);
        $curl->setOption(CURLOPT_HTTPHEADER, $headers);
        $curl->setOption(CURLOPT_RETURNTRANSFER, true);
        $curl->setOption(CURLOPT_FOLLOWLOCATION, true);
        $curl->setOption(CURLOPT_SSL_VERIFYPEER, false);
        $curl->setOption(CURLOPT_SSL_VERIFYHOST, false);
        $return = $curl->execute();
        $errorNo = $curl->getErrorNo();
        if ($errorNo) {
            $curl->refresh();
            throw new Exception($curl->getError(), $errorNo);
        }
        $code = $curl->getInfo(CURLINFO_HTTP_CODE);
        Log::getInstance()->event('Response %s %s', $code, $return);
        if ($code < 200 || $code >= 300) {
            $curl->refresh();
            throw new Exception($return, $code);
        }
        $curl->refresh();
        return $return;
    }


    /**
     * @param $instances
     * @param $resourceId
     * @return array
     */
    private function findApsInstanceIdByResourceId($instances, $resourceId)
    {
        $instances = array_filter($instances, function ($application) use ($resourceId) {
            return $application->globals->aps->id === $resourceId;
        });
        $instance = reset($instances);
        return $instance->aps->id;
    }

    /**
     * @param Service_MspApiClient_ApiInterface $mspApi
     * @param int $groupId
     * @param DateTime $startDate
     * @param DateTime $endDate
     * @return void
     */
    private function requestUsageReport(Service_MspApiClient_ApiInterface $mspApi, $groupId, $startDate, $endDate)
    {
        $when = clone $endDate;
        $when->modify(self::WHEN_OFFSET_END_DATE);
        Log::getInstance()->notice(
            'Request to generate report for group %s with report attributes: period %s-%s, when %s',
            $groupId,
            $startDate->format('Y-m-d H:i:s'),
            $endDate->format('Y-m-d H:i:s'),
            $when->format('Y-m-d H:i:s')
        );
        if ($this->dryRunMode) {
            return;
        }

        $requestUsageReport = $this->createRequestUsageReportStruct($startDate, $endDate, $when);
        $result = $mspApi->requestUsageReport($requestUsageReport, $groupId);

        Log::getInstance()->notice(
            'Report requested. report id - %s, version - %s',
            $result->getReportId(),
            $result->getVersion()
        );
        return;
    }

    /**
     * @param DateTime $startDate
     * @param DateTime $endDate
     * @param DateTime|null $when
     * @param int $kind
     * @param string $resultAction
     * @param int $level
     * @return Service_MspApiClient_Request_Struct_RequestUsageReport
     */
    private function createRequestUsageReportStruct(
        DateTime $startDate,
        DateTime $endDate,
        $when = null,
        $kind = Service_MspApiClient_Request_Struct_RequestUsageReport::KIND_DAILY,
        $resultAction = Service_MspApiClient_Request_Struct_RequestUsageReport::RESULT_ACTION_SAVE,
        $level = Service_MspApiClient_Request_Struct_RequestUsageReport::LEVEL_UP_TO_EUC
    )
    {
        $when = is_null($when) ? Service_MspApiClient_Request_Struct_RequestUsageReport::WHEN_ASAP : $when->getTimestamp();

        $requestUsageReport = new Service_MspApiClient_Request_Struct_RequestUsageReport();
        $requestUsageReport->setBeginDate($startDate->getTimestamp());
        $requestUsageReport->setEndDate($endDate->getTimestamp());
        $requestUsageReport->setWhen($when);
        $requestUsageReport->setKind($kind);
        $requestUsageReport->setResultAction($resultAction);
        $requestUsageReport->setLevel($level);
        return $requestUsageReport;
    }

    /**
     * @param $apsInstanceId
     * @param $datacenter
     */
    private function requestReportForDatacenter($apsInstanceId, $datacenter)
    {
        $mspApi = $this->apiFactory->getMspApi($datacenter, $apsInstanceId);
        $rootGroupId = $mspApi->getGroupById('self')->getId();

        $timezone = new DateTimeZone(self::DEFAULT_TIMEZONE);

        $startDate = new DateTime('now', $timezone);
        $startDate->setTime(0, 0, 0);
        $endDate = new DateTime('now', $timezone);
        $endDate->setTime(23, 59, 59);

        $startDate->modify('-1 day');
        $endDate->modify('-1 day');
        $this->requestUsageReport($mspApi, $rootGroupId, $startDate, $endDate);
    }
}
