<?php
/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

require_once(__DIR__ . '/app/bootstrap.php');

/**
 * Class users
 * @type("http://acronis.com/aps/cloud-backup-v1/users/7.5")
 * @implements("http://aps-standard.org/types/core/resource/1.0")
 * @implements("http://aps-standard.org/types/core/user/service/1.0")
 */
class usersCCPv2 extends users
{
    const ON_USER_AVAILABLE = 'onUserAvailable';

    /* ======= RESOURCE PROPERTIES ======= */

    /**
     * @link("http://parallels.com/aps/types/pa/user/1.0")
     * @required
     */
    public $user;

    /* ======= OTHER PROPERTIES ======= */

    /**
     * @var Model_UserScreenNotificationManagerInterface
     */
    private $notificationManager;

    /* ======= RESOURCE METHODS ======= */

    public function provision()
    {
        try {
            parent::provision();
        } catch (\Rest\Accepted $exception) {
            // do nothing
        }

        // finally section
        try {
            $notificationManager = $this->getScreenNotificationManager();
            $notificationManager->provisionHasStarted();
        } catch (Exception $e) {
            Log::getInstance()->error($e->getMessage());
        }

        if (isset($exception)) {
            throw $exception;
        }
    }

    public function provisionAsync()
    {
        if (!$this->tenant->ismigration && $this->isServiceUserHasSubscription()) {
            $this->userstatus = self::USER_STATUS_ERROR;
            $this->subscribeToOnUserAvailableEvent();
            return;
        }
        try {
            parent::provisionAsync();
        } catch (\Rest\Accepted $e) {
            // continue async
            throw $e;
        } catch (\Exception $exception) {
            Log::getInstance()->error($exception->getMessage());
            // do nothing
        }

        // finally section
        try {
            $notificationManager = $this->getScreenNotificationManager();
            if (isset($exception)) {
                $notificationManager->provisionHasFailed();
            } else {
                $notificationManager->provisionHasFinished();
            }
        } catch (Exception $e) {
            Log::getInstance()->error($e->getMessage());
        }

        if (isset($exception)) {
            if ($this->tenant->ismigration) {
                throw $exception;
            }

            $this->userstatus = self::USER_STATUS_ERROR;
            $this->subscribeToOnUserAvailableEvent();
            // Do not throw exception
            // Remove user in onUserAvailable
        }
    }

    /**
     * @param self|null $new
     * @throws Exception
     */
    public function configure($new = null)
    {
        $notificationManager = $this->getScreenNotificationManager();
        $notificationManager->configureHasStarted();
        try {
            parent::configure($new);
        } catch (\Exception $e) {
            $notificationManager->configureHasFailed();
            throw $e;
        }
        $notificationManager->configureHasFinished();
    }

    public function unprovision()
    {
        Log::getInstance()->event(sprintf(
            self::LOG_CALLED_UNPROVISION,
            $this->subscriptionId,
            $this->endUserLogin
        ));

        if ($this->userstatus == self::USER_STATUS_ERROR) {
            parent::unprovision();
            return;
        }

        $notificationManager = $this->getScreenNotificationManager();
        $notificationManager->unprovisionHasStarted();
        try {
            parent::unprovision();
        } catch (Exception $e) {
            $notificationManager->unprovisionHasFailed();
            throw $e;
        }
        $notificationManager->unprovisionHasFinished();
    }

    /* ======= RESOURCE EVENTS ======= */

    /**
     * We define the operation on post event
     * @verb(POST)
     * @path("/onUserAvailable")
     *
     * @param("http://aps-standard.org/types/core/resource/1.0#Notification",body)
     */
    public function onUserAvailable($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        Log::getInstance()->event(sprintf(self::STAFF_MEMBER_AVAILABLE, $this->endUserLogin));

        if ($this->userstatus == self::USER_STATUS_ERROR) {
            $this->_apsc->unprovisionResource($event->source->id);
        }
        $this->unsubscribeFromOnUserAvailableEvent();
    }

    /* ======= RESOURCE OPERATIONS ======= */

    /**
     * @verb(POST)
     * @path("/changePassword")
     * @param(object,body)
     * @access(admin,true)
     * @access(owner,true)
     * @access(referrer,true)
     */
    public function changePassword($configuration)
    {
        Log::getInstance()->event(self::LOG_CALLED_CHANGE_PASSWORD);
        $notificationManager = $this->getScreenNotificationManager();
        $notificationManager->changePasswordHasStarted();
        try {
            parent::changePassword($configuration);
        } catch (Exception $e) {
            $notificationManager->changePasswordHasFailed();
            throw $e;
        }
        $notificationManager->changePasswordHasFinished();
    }

    /**
     * @verb(POST)
     * @path("/activateAccount")
     * @param(object,body)
     * @access(admin,true)
     * @access(owner,true)
     * @access(referrer,true)
     */
    public function activateAccount($configuration)
    {
        Log::getInstance()->event(self::LOG_CALLED_ACTIVATE_ACCOUNT);
        $notificationManager = $this->getScreenNotificationManager();
        $notificationManager->activateAccountHasStarted();
        try {
            parent::activateAccount($configuration);
        } catch (Exception $e) {
            $notificationManager->activateAccountHasFailed();
            throw $e;
        }
        $notificationManager->activateAccountHasFinished();
    }

    /* ======= OTHER METHODS ======= */

    /**
     * @return Model_UserScreenNotificationManagerInterface
     */
    private function getScreenNotificationManager()
    {
        if (!isset($this->notificationManager)) {
            $this->notificationManager = $this->_managerFactory->getUserScreenNotificationManager($this);
        }

        return $this->notificationManager;
    }

    protected function isServiceUserHasSubscription()
    {
        $notificationManager = $this->getScreenNotificationManager();

        $rql = RqlBuilder::createInstance()
            ->implementing(\APS\TypeLibrary::getTypeByServiceId('users'))
            ->linkedWith($this->serviceUserId)
            ->ne('aps.id', $this->aps->id)
            ->limit(0, 1)
            ->getQuery();

        $assignedUsers = $this->_apsc->impersonate($this->tenant)->getResources($rql);
        if (count($assignedUsers) > 0) {
            $notificationManager->provisionHasFailedSubscriptionAlreadyAssigned(
                $this->subscriptionId,
                $assignedUsers[0]->subscriptionId
            );
            return true;
        }
        return false;
    }

    /**
     * @return string
     */
    protected function getSelfApsType()
    {
        return self::CLOUD_BACKUP_USERS_TYPE_CCPv2;
    }

    protected function subscribeToOnUserAvailableEvent()
    {
        $eventSubscription = new \APS\EventSubscription(\APS\EventSubscription::Available, self::ON_USER_AVAILABLE);
        $this->subscribeOnEventById($eventSubscription, $this->aps->id);
    }

    protected function unsubscribeFromOnUserAvailableEvent()
    {
        $this->unsubscribeFromEventById(\APS\EventSubscription::Available, $this->aps->id);
    }
}
